/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <coding/bcbase.hh>


BCData_Bitmaps::BCData_Bitmaps(const SearchParam& sparam)
{
  const Pixel*const* refbm;

  if (sparam.framesearch)
    refbm = sparam.bitmap->AskFrame_const();
  else
    {
      refbm = sparam.bitmap->AskField_const(sparam.top_field);
    }

  assert(refbm);

  const int w = sparam.bitmap->AskWidth();
  const int h = (sparam.framesearch ? sparam.bitmap->AskHeight() :
                                      sparam.bitmap->AskHeight() / 2);

  d_halfpel = sparam.halfpel;
  d_blk_width  = sparam.blk_width;
  d_blk_height = sparam.blk_height;
  d_pixels_per_block = sparam.blk_width * sparam.blk_height;

#ifndef MVPIX_IS_PIX
  d_reference_bm.CreateBitmap(w,h);
  d_reference = d_reference_bm.AskFrame();
#endif

  if (d_halfpel)
    {
      // Half-Pel-Bitmaps neu erzeugen
      
      d_reference_bm_h .CreateBitmap(w-1,h  );
      d_reference_bm_v .CreateBitmap(w  ,h-1);
      d_reference_bm_hv.CreateBitmap(w-1,h-1);
      
      d_reference_halfh  = d_reference_bm_h .AskFrame(); assert(d_reference_halfh);
      d_reference_halfv  = d_reference_bm_v .AskFrame(); assert(d_reference_halfv);
      d_reference_halfhv = d_reference_bm_hv.AskFrame(); assert(d_reference_halfhv);
    }


  // --- Neue Bitmaps fuellen ---

#ifndef MVPIX_IS_PIX
  {for (int y=0;y<h;y++)
    for (int x=0;x<w;x++)
      d_reference[y][x] = refbm[y][x];}
#else
  d_reference = const_cast<Pixel*const*>(refbm);
#endif

  if (d_halfpel)
    {
      assert(d_reference_halfh);
      assert(d_reference_halfv);
      assert(d_reference_halfhv);
      assert(refbm);

      {for (int y=0;y<h;y++)
        for (int x=0;x<w-1;x++)
          d_reference_halfh[y][x] = (refbm[y][x] + refbm[y][x+1] +1)/2;}

      {for (int y=0;y<h-1;y++)
        for (int x=0;x<w;x++)
          d_reference_halfv[y][x] = (refbm[y][x] + refbm[y+1][x] +1)/2;}

      {for (int y=0;y<h-1;y++)
        for (int x=0;x<w-1;x++)
          d_reference_halfhv[y][x] = (refbm[y  ][x] + refbm[y  ][x+1] +
                                      refbm[y+1][x] + refbm[y+1][x+1] +2)/4;}
    }
}


BCData_Bitmaps::~BCData_Bitmaps()
{
}


BCData* BlockComparator_Base::CreateBCData(const SearchParam& sparam)
{
  return new BCData_Bitmaps(sparam);
}
