/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CODING_BCBASE_HH
#define CODING_BCBASE_HH

#include "coding/motionest.hh"


class BlockComparator_Base : public BlockComparator
{
public:
  BlockComparator_Base() { }
  virtual ~BlockComparator_Base() { }

  virtual BCData* CreateBCData(const SearchParam&);
};



struct BCData_Bitmaps : public BCData
{
  BCData_Bitmaps(const SearchParam&);
  ~BCData_Bitmaps();

  double d_pixels_per_block;
  int  d_blk_width , d_blk_height;

  inline const MVPixel*const* AskPixels(int& x,int& y) const;

private:
      MVPixel*const * d_reference;          // Original-Bitmap (1:1)
      MVPixel*const * d_reference_halfh;    // Bitmap aus jeweils 2 horiz. gemittelten Pixeln
      MVPixel*const * d_reference_halfv;    // Bitmap aus jeweils 2 verti. gemittelten Pixeln
      MVPixel*const * d_reference_halfhv;   // Bitmap aus jeweils 4 gemittelten Pixeln
  const Pixel*const * d_image;

  bool d_halfpel;

#ifndef MVPIX_EQ_PIX
  Bitmap<MVPixel> d_reference_bm;
#endif
  Bitmap<MVPixel> d_reference_bm_h;
  Bitmap<MVPixel> d_reference_bm_v;
  Bitmap<MVPixel> d_reference_bm_hv;
};


inline const MVPixel*const* BCData_Bitmaps::AskPixels(int& x,int& y) const
{
  if (!d_halfpel)
    return d_reference;

  const MVPixel*const* ref = d_reference;

  bool halfh = (x & 1);
  bool halfv = (y & 1);

  x >>= 1;
  y >>= 1;

  if (halfv)
    {
      if (halfh)
        ref = d_reference_halfhv;
      else
        ref = d_reference_halfv;
    }
  else
    {
      if (halfh)
        ref = d_reference_halfh;
    }

  return ref;
}



#endif
