/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/bcmmx.hh"

#define MYVERSION 0
#define INTEL_ORIGINAL 0

void BlockComparator_MMX::SetImageBitmap(const Bitmap<Pixel>& bm,bool frame,bool top)
{
  if (frame)
    d_image = bm.AskFrame_const();
  else
    d_image = bm.AskField_const(top);
}


typedef unsigned long long mmx;

bool BlockComparator_MMX::Error(BCData* bcdata,
                                const ComparePos& cmppos,
                                uint32& error,
                                uint32  maxerror,
                                int subfact) const
{
  int refx = cmppos.ref_x;
  int refy = cmppos.ref_y;

  const BCData_Bitmaps* bcdat = dynamic_cast<BCData_Bitmaps*>(bcdata);
  assert(bcdat);

  const MVPixel*const* ref = bcdat->AskPixels(refx,refy);

  error=0;

  assert(bcdat->d_blk_width == 16);

  int endy = cmppos.img_y + bcdat->d_blk_height;

  mmx diff=0;
  error=0;

  __asm__ __volatile__
    (
#if MYVERSION
     "pxor      %mm3,%mm3\n\t"    // mm3 auf 0 setzen
     "pxor      %mm4,%mm4\n\t"    // Fehler auf 0 setzen (mm4)
#else
     "pxor      %mm6,%mm6\n\t"    // mm6 auf 0 setzen
     "pxor      %mm7,%mm7\n\t"    // Fehler auf 0 setzen (mm4)    (Intel-Code)
#endif
     );

  for (int iy=cmppos.img_y , ry=refy ;
       iy<endy ;
       iy+=subfact,ry+=subfact)
    {
      const mmx* ip = (const mmx*)&d_image[iy][cmppos.img_x];  // Zeiger ins Originalbild
      const mmx* rp = (const mmx*)&ref    [ry][refx];          // Zeiger ins Referenzbild

      mmx pa1,pa2,pb1,pb2;
      pa1 = *ip++;
      pa2 = *ip;
      pb1 = *rp++;
      pb2 = *rp;

#if MYVERSION
      // Meine Version

      __asm__ __volatile__
        (
         // Die ersten 8 Pixel

         "movq      %0,%%mm0\n\t"       // Daten A -> mm0
         "movq      %1,%%mm1\n\t"       // Daten B -> mm1
         "movq      %%mm0,%%mm2\n\t"    // Daten A -> mm2   (Kopie)
         "psubusb   %%mm1,%%mm2\n\t"    // A-B     -> mm2
         "psubusb   %%mm0,%%mm1\n\t"    // B-A     -> mm1
         "por       %%mm2,%%mm1\n\t"    // Zusammenfassen der Differenzen -> mm1
         "movq      %%mm1,%%mm2\n\t"    //                       und nach -> mm2
         "punpckhbw %%mm3,%%mm1\n\t"    // Oberen  4 Werte auf 16bit erweitert nach -> mm1
         "punpcklbw %%mm3,%%mm2\n\t"    // Unteren 4 Werte auf 16bit erweitert nach -> mm2
         "paddw     %%mm1,%%mm4\n\t"    // mm1 auf alten Fehler addieren
         "paddw     %%mm2,%%mm4\n\t"    // mm2 auf alten Fehler addieren

         // Die zweiten 8 Pixel

         "movq      %2,%%mm0\n\t"       // Daten A -> mm0
         "movq      %3,%%mm1\n\t"       // Daten B -> mm1
         "movq      %%mm0,%%mm2\n\t"    // Daten A -> mm2   (Kopie)
         "psubusb   %%mm1,%%mm2\n\t"    // A-B     -> mm2
         "psubusb   %%mm0,%%mm1\n\t"    // B-A     -> mm1
         "por       %%mm2,%%mm1\n\t"    // Zusammenfassen der Differenzen -> mm1
         "movq      %%mm1,%%mm2\n\t"    //                       und nach -> mm2
         "punpckhbw %%mm3,%%mm1\n\t"    // Oberen  4 Werte auf 16bit erweitert nach -> mm1
         "punpcklbw %%mm3,%%mm2\n\t"    // Unteren 4 Werte auf 16bit erweitert nach -> mm2
         "paddw     %%mm1,%%mm4\n\t"    // mm1 auf alten Fehler addieren
         "paddw     %%mm2,%%mm4\n\t"    // mm2 auf alten Fehler addieren
         : : "m" (pa1), "m" (pb1), "m" (pa2), "m" (pb2)
         );
#else

#if INTEL_ORIGINAL
      // Intel original

      __asm__ __volatile__
        (
         "movq      %0,%%mm0\n\t"
         "movq      %1,%%mm1\n\t"
         "movq      %%mm0,%%mm2\n\t"
         "movq      %2,%%mm3\n\t"
         "psubusb   %%mm1,%%mm0\n\t"
         "psubusb   %%mm2,%%mm1\n\t"
         "movq      %3,%%mm4\n\t"
         "por       %%mm1,%%mm0\n\t"
         "movq      %%mm3,%%mm5\n\t"
         "movq      %%mm0,%%mm1\n\t"
         "psubusb   %%mm4,%%mm3\n\t"
         "punpcklbw %%mm6,%%mm0\n\t"
         "psubusb   %%mm5,%%mm4\n\t"
         "psrlq     $32,%%mm1\n\t"
         "por       %%mm4,%%mm3\n\t"
         "punpcklbw %%mm6,%%mm1\n\t"
         "movq      %%mm3,%%mm4\n\t"
         "punpcklbw %%mm6,%%mm3\n\t"
         "paddw     %%mm0,%%mm7\n\t"
         "psrlq     $32,%%mm4\n\t"
         "paddw     %%mm1,%%mm7\n\t"
         "punpcklbw %%mm6,%%mm4\n\t"
         "paddw     %%mm3,%%mm7\n\t"
         "paddw     %%mm4,%%mm7\n\t"
         : : "m" (pa1), "m" (pb1), "m" (pa2), "m" (pb2)
         );
#else
      // Intel modifiziert

      __asm__ __volatile__
        (
         "movq      %0,%%mm0\n\t"
         "movq      %1,%%mm1\n\t"
         "movq      %%mm0,%%mm2\n\t"
         "movq      %2,%%mm3\n\t"
         "psubusb   %%mm1,%%mm0\n\t"
         "psubusb   %%mm2,%%mm1\n\t"
         "movq      %3,%%mm4\n\t"
         "por       %%mm1,%%mm0\n\t"
         "movq      %%mm3,%%mm5\n\t"
         "movq      %%mm0,%%mm1\n\t"
         "psubusb   %%mm4,%%mm3\n\t"
         "punpcklbw %%mm6,%%mm0\n\t"
         "psubusb   %%mm5,%%mm4\n\t"
         "por       %%mm4,%%mm3\n\t"
         "punpckhbw %%mm6,%%mm1\n\t"
         "movq      %%mm3,%%mm4\n\t"
         "punpcklbw %%mm6,%%mm3\n\t"
         "paddw     %%mm0,%%mm7\n\t"
         "paddw     %%mm1,%%mm7\n\t"
         "punpckhbw %%mm6,%%mm4\n\t"
         "paddw     %%mm3,%%mm7\n\t"
         "paddw     %%mm4,%%mm7\n\t"
         : : "m" (pa1), "m" (pb1), "m" (pa2), "m" (pb2)
         );
#endif

#endif

#if 0
      if (diff & 0xFFFF > maxerror/2)
        {
          error  =  diff      & 0xFFFF;
          error += (diff>>16) & 0xFFFF;
          error += (diff>>32) & 0xFFFF;
          error += (diff>>48) & 0xFFFF;
          if (error>maxerror)
            return false;
        }
#endif

      /*
      printf("Data: %016llx / %016llx\n", pa1,pa2);
      printf("Data: %016llx / %016llx\n", pb1,pb2);
      printf("Diff: %016llx\n", diff);

      exit(0);
      */
    }

  __asm__ __volatile__
    (
#if MYVERSION
     "movq    %%mm4,%0\n\t"
#else
     "movq    %%mm7,%0\n\t"
#endif
     "emms\n\t"
     : "=m" (diff)
    );

  error  =  diff      & 0xFFFF;
  error += (diff>>16) & 0xFFFF;
  error += (diff>>32) & 0xFFFF;
  error += (diff>>48) & 0xFFFF;

  if (subfact!=1)
    error *= subfact*subfact;

  if (error>maxerror)
    return false;
  else
    return true;
}

