/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/bcstd.hh"

void BlockComparator_Standard::SetImageBitmap(const Bitmap<Pixel>& bm,bool frame,bool top)
{
  if (frame)
    d_image = bm.AskFrame_const();
  else
    d_image = bm.AskField_const(top);
}


bool BlockComparator_Standard::Error(BCData* bcdata,
                                     const ComparePos& cmppos,
                                     uint32& error,
                                     uint32  maxerror,
                                     int subfact) const
{
  int refx = cmppos.ref_x;
  int refy = cmppos.ref_y;

  const BCData_Bitmaps* bcdat = dynamic_cast<BCData_Bitmaps*>(bcdata);
  assert(bcdat);

  const MVPixel*const* ref = bcdat->AskPixels(refx,refy);

  error=0;

  assert(bcdat->d_blk_width == 16);

  int endy = cmppos.img_y + bcdat->d_blk_height;

  for (int iy=cmppos.img_y , ry=refy ;
       iy<endy ;
       iy+=subfact,ry+=subfact)
    {
#if INTERLEAVED_MVSEARCH
      const   Pixel* ip1 = &d_image[iy][cmppos.img_x];  // Zeiger ins Originalbild
      const MVPixel* rp1 = &ref    [ry][refx];          // Zeiger ins Referenzbild
      const MVPixel* ip2 = ip1 + 8*subfact;
      const MVPixel* rp2 = rp1 + 8*subfact;

      int diff1,diff2;
      int error2=0;

      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1; ip1+=subfact; rp1+=subfact;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2; ip2+=subfact; rp2+=subfact;
      diff1 = *ip1-*rp1; if (diff1>=0) error  += diff1; else error  -= diff1;
      diff2 = *ip2-*rp2; if (diff2>=0) error2 += diff2; else error2 -= diff2;

      error += error2;
#else
      const   Pixel* ip = &d_image    [iy][cmppos.img_x];  // Zeiger ins Originalbild
      const MVPixel* rp = &ref        [ry][refx];  // Zeiger ins Referenzbild

      int diff;

      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;    ip+=subfact; rp+=subfact;
      diff = *ip - *rp; if (diff>=0) error += diff; else error -= diff;
#endif

#if 1
      if (error>maxerror)
	{ return false; }
#endif
    }

  if (subfact!=1)
    error *= subfact*subfact;

  return true;
}
