/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "types.hh"
#include "coding/dct.hh"
#include "coding/utility.hh"
#include "mpeg/constants.hh"
#include "options.hh"
#include <iomanip.h>

template <class Pel> void DCTCoder<Pel>::CodeDCTBlocks(const Image<Pel>& srcimg,
                                                       ImageData& codedimg)
{
  SetQScaleType (codedimg.pichdr.m_q_scale_type);
  SetDCPrecision(codedimg.pichdr.m_intra_dc_precision);

  for (int y=0;y<codedimg.mblks.AskHeight();y++)
    for (int x=0;x<codedimg.mblks.AskWidth();x++)
      {
	// if (codedimg.mblks.Ask(y,x).m_SkipCoding) continue;

	MBlk_FDCTQuant(srcimg,x,y,codedimg.mblks.Ask(y,x));
      }
};

template class DCTCoder<Pixel>;
template class DCTCoder<DiffPixel>;






Task_DCTQuantize::Task_DCTQuantize(int framenr)
  : m_framenr(framenr)
{
}

void Task_DCTQuantize::Execute(ProcessData& procdata)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("FDCT & quantize",procdata.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  procdata.dctcoder_pix->SetChromaFormat(CHROMA_420);
  procdata.dctcoder_pix->SetQScaleType(img->pichdr.m_q_scale_type);
  procdata.dctcoder_pix->SetDCPrecision(img->pichdr.m_intra_dc_precision);

  procdata.dctcoder_diffpix->SetChromaFormat(CHROMA_420);
  procdata.dctcoder_diffpix->SetQScaleType(img->pichdr.m_q_scale_type);
  procdata.dctcoder_diffpix->SetDCPrecision(img->pichdr.m_intra_dc_precision);

  for (int y=0;y<img->mblks.AskHeight();y++ /*,cout<<endl*/ )
    for (int x=0;x<img->mblks.AskWidth();x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);

        // Quantizer-Scale koennte noch mit Dithering versehen werden...
        int qscale = (int)mb.m_quantiser_scale;
        if (qscale<  1) qscale=1;
        if (qscale>127) qscale=127;

        mb.m_quantiser_scale_code = qscale2qcode[options.mpeg2.ExponentialQScale ? 1 : 0][qscale];


        if (mb.m_HasMotionForward && mb.m_HasMotionBackward)
          {
            //cout<<'B';
            procdata.dctcoder_diffpix->MBlk_FDCTQuant(img->pred_bidir,x,y,img->mblks.Ask(y,x));
          }
        else if (mb.m_HasMotionForward)
          {
            //cout<<'>';
            procdata.dctcoder_diffpix->MBlk_FDCTQuant(img->pred_forw ,x,y,img->mblks.Ask(y,x));
          }
        else if (mb.m_HasMotionBackward)
          {
            //cout<<'<';
            procdata.dctcoder_diffpix->MBlk_FDCTQuant(img->pred_backw,x,y,img->mblks.Ask(y,x));
          }
        else
          {
            //cout<<'I';
            procdata.dctcoder_pix    ->MBlk_FDCTQuant(*img->source   ,x,y,img->mblks.Ask(y,x));
          }
      }
}
