/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "types.hh"
#include "coding/dct_coupled.hh"

#include "coding/utility.hh"
#include "mpeg/defaultmatrices.hh"
#include "mpeg/constants.hh"
#include <iostream.h>


#define easysign(x) ((x>=0) ? 1 : -1)

#include <iostream.h>

template <class Pel> void DCTCoder_Coupled<Pel>::SetChromaFormat(uint2 chroma_format)
{
  switch (chroma_format)
    {
    case CHROMA_420: d_chromaxfact= 8; d_chromayfact= 8; break;
    case CHROMA_422: d_chromaxfact= 8; d_chromayfact=16; break;
    case CHROMA_444: d_chromaxfact=16; d_chromayfact=16; break;
    }

  DCTCoder_Separated<Pel>::SetChromaFormat(chroma_format);
}

template <class Pel> void DCTCoder_Coupled<Pel>::SetQScaleType(uint1 qscale_type)
{
  d_recalc_scalefacts |= (qscale_type != d_q_scale_type);
  d_q_scale_type       = qscale_type;

  DCTCoder_Separated<Pel>::SetQScaleType(qscale_type);
}

template <class Pel> void DCTCoder_Coupled<Pel>::SetDCPrecision(uint8 dc_precision)
{
  d_recalc_scalefacts |= (dc_precision != d_dc_precision);
  d_dc_precision       = dc_precision;

  DCTCoder_Separated<Pel>::SetDCPrecision(dc_precision);
}


template <class Pel> void DCTCoder_Coupled<Pel>::MBlk_FDCTQuant(const Image<Pel>& srcimg,
                                                                int x,int y,Macroblock& mb)
{
  if (d_recalc_scalefacts)
    {
      PrecalcQMatrix();
    }


  /*  BILD CODIEREN */

  const Pel*const * y_pix  = srcimg.AskBitmap_const(Bitmap_Y ).AskFrame_const();
  const Pel*const * cr_pix = srcimg.AskBitmap_const(Bitmap_Cr).AskFrame_const();
  const Pel*const * cb_pix = srcimg.AskBitmap_const(Bitmap_Cb).AskFrame_const();


  const Pel* in [8];

  int lx0 = x*16,  lx1 = x*16+8;
  int ly0 = y*16,  ly1 = y*16+8;

  int cx0 = x*d_chromaxfact , cx1 = x*d_chromaxfact+8;
  int cy0 = y*d_chromayfact , cy1 = y*d_chromayfact+8;

  uint1 DCTselect = (mb.m_IsIntra ? 0 : 1);
  DCT2_AraiAguiNakajima_Int& dct_lum   = d_dct[DCTselect][0][mb.m_quantiser_scale_code];
  DCT2_AraiAguiNakajima_Int& dct_chrom = d_dct[DCTselect][1][mb.m_quantiser_scale_code];


  if (mb.m_CodeY[0])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly0+i][lx0]; }}
      dct_lum.PerformFDCT2(in,mb.m_y[0]);
    }

  if (mb.m_CodeY[1])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly0+i][lx1]; }}
      dct_lum.PerformFDCT2(in,mb.m_y[1]);
    }

  if (mb.m_CodeY[2])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly1+i][lx0]; }}
      dct_lum.PerformFDCT2(in,mb.m_y[2]);
    }

  if (mb.m_CodeY[3])
    {
      {for (int i=0;i<8;i++) { in[i] = &y_pix[ly1+i][lx1]; }}
      dct_lum.PerformFDCT2(in,mb.m_y[3]);
    }

  if (mb.m_CodeCr[0])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy0+i][cx0]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cr[0]);
    }

  if (mb.m_CodeCr[1])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy1+i][cx0]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cr[1]);
    }

  if (mb.m_CodeCr[2])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy0+i][cx1]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cr[2]);
    }

  if (mb.m_CodeCr[3])
    {
      {for (int i=0;i<8;i++) { in[i] = &cr_pix[cy1+i][cx1]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cr[3]);
    }

  if (mb.m_CodeCb[0])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy0+i][cx0]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cb[0]);
    }

  if (mb.m_CodeCb[1])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy1+i][cx0]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cb[1]);
    }

  if (mb.m_CodeCb[2])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy0+i][cx1]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cb[2]);
    }

  if (mb.m_CodeCb[3])
    {
      {for (int i=0;i<8;i++) { in[i] = &cb_pix[cy1+i][cx1]; }}
      dct_chrom.PerformFDCT2(in,mb.m_cb[3]);
    }
}


template <class Pel> void DCTCoder_Coupled<Pel>::PrecalcQMatrix()
{
  if (d_recalc_scalefacts)
    {
      // DCT-Ausgangsskalierung aendern, um die
      // Quantisierung gleichzeitig mit der DCT-Transformation durchfuehren.
      
      DCTFloat  scalefact[8][8];
      DCTFloat* scalefactptr[8];
      
      for (int v=0;v<8;v++)
        scalefactptr[v]=scalefact[v];
      
      // --- fuer Intra-Bloecke ---
      
      {
        for (int q=1;q<32;q++)
          {
            for (int v=0;v<8;v++)
              for (int u=0;u<8;u++)
                {
                  scalefact[v][u] = 16.0/(double)(qcode2qscale[d_q_scale_type][q] * 
                                                  QuantMatrixIntra[v][u]);
                }
            
            scalefact[0][0] = 1.0/(1<<(11-d_dc_precision));
            d_dct[0][0][q].SetOutputScalefactors((const DCTFloat**)scalefactptr);
            d_dct[0][1][q].SetOutputScalefactors((const DCTFloat**)scalefactptr);
          }
      }
      
      // --- fuer Inter-Bloecke ---
      
      {
        for (int q=1;q<32;q++)
          {
            for (int v=0;v<8;v++)
              for (int u=0;u<8;u++)
                {
                  scalefact[v][u] = 1.0/(double)(qcode2qscale[d_q_scale_type][q]);
                }
            
            d_dct[1][0][q].SetOutputScalefactors((const DCTFloat**)scalefactptr);
            d_dct[1][1][q].SetOutputScalefactors((const DCTFloat**)scalefactptr);
          }
      }
    }
  
  d_recalc_scalefacts=false;
}


template class DCTCoder_Coupled<Pixel>;
template class DCTCoder_Coupled<DiffPixel>;
