/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CODING_DECODER_HH
#define CODING_DECODER_HH

#include "types.hh"
#include "image/image.hh"
// #include "image/codedimg.hh"
// #include "math/dct2_theoretic.hh"
// #include "math/dct2_arai.hh"
#include "math/dct2_arai_int.hh"
#include "control/datapool.hh"
#include "control/taskpool.hh"

/* Bewegungskompensation fuer einen einzelnen Bewegungsvektor durchfuehren. */

void MotionCompensateSingleSet(Pixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
                               const Pixel*const* reference,
			       int hdist,int vdist);
void MotionCompensateSingleSet(DiffPixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
                               const Pixel*const* reference,
			       int hdist,int vdist);

#if 0
inline void MotionCompensateSingleSet(Pixel*const* destination,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference,
			       int hdist,int vdist)
{ MotionCompensateSingleSet(destination,x0,y0,x0,y0,w,h,reference,hdist,vdist); }
#endif


void MotionCompensateDoubleSet(Pixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference1,
			       int hdist1,int vdist1,
			       const Pixel*const* reference2,
			       int hdist2,int vdist2);
void MotionCompensateDoubleSet(DiffPixel*const* destination,
			       int x1,int y1,
			       int x0,int y0,int w,int h,
			       const Pixel*const* reference1,
			       int hdist1,int vdist1,
			       const Pixel*const* reference2,
			       int hdist2,int vdist2);

#if 0
inline void MotionCompensateDoubleSet(Pixel*const* destination,
				      int x0,int y0,int w,int h,
				      const Pixel*const* reference1,
				      int hdist1,int vdist1,
				      const Pixel*const* reference2,
				      int hdist2,int vdist2)
{ MotionCompensateDoubleSet(destination,x0,y0,x0,y0,w,h,
			    reference1,hdist1,vdist1,reference2,hdist2,vdist2); }
#endif


class Task_Decode : public Task
{
public:
  Task_Decode(int framenr)
    : m_framenr(framenr) { }

  void Execute(ProcessData&);

  const int   m_framenr;

private:
  static IDCT2_AraiAguiNakajima_Int decoder_dct;

#if 0
  static int qW[2 /* 0 Intra / 1 Inter */]
               [2 /* 0 Lum / 1 Chrom */]
               [32] /* qcode */
               [8][8];
  uint1 d_q_scale_type;
  uint8 d_intra_dc_precision;

  int d_chromaxfact,d_chromayfact;
  int d_chromaxdiv ,d_chromaydiv;
#endif

  // void ClearBlock(short* out[8]);
  void IDCT(const DCTBlock&,uint8* out[8], uint1 inter,uint1 chrom,int qcodem,bool addvals,
            PictureHeaderData& pichdr);
};

#endif
