/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/
/*
  BUG: GOP-Anfaenge sind FALSCH !!!
       Das erste I-Frame ist noch vor der GOP!
       Siehe (*)

  Gefixt: 23.2.99

  Zwar faengt die GOP wirklich an der markierten Stelle an, aber da das folgende I-Frame frueher
  geschrieben wird als die B-Frames, muss GOP-Beginning im I-Frames gesetzt werden.
 */




#include "coding/frameclass_scenechange.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"
#include <stdlib.h>

#define THRES_0  4
#define THRES_1  7


Task_ClassifyFrames_SceneChange::Task_ClassifyFrames_SceneChange(int firstframe,int nFrames,int pdist,
                                                                 int minidist)
{
  m_firstframe=firstframe;
  m_nFrames=nFrames;
  m_LastFrames=false;
  m_PDistance=pdist;
  m_MinIDistance=minidist;
}


void Task_ClassifyFrames_SceneChange::Execute(ProcessData& data)
{
  int last_frame = m_firstframe+m_nFrames-1;

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("frametype (scene.c.)",data.threadnr);
      cerr << "frame " << m_firstframe << " to " << last_frame << endl;
      EndShowTaskInformation();
    }


  // Erstes Frame ist Anfang einer neuen GOP

  ImageData* firstframe = ImageData::GetNr(m_firstframe);
  // firstframe->pichdr.m_picture_coding_type = PICTYPE_I;

#if 0
  firstframe->SetGOPBeginning();
  firstframe->IsClosedGOP();
#endif

  if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
    cout << "Frame " << m_firstframe << " is beginning of GOP\n";

  m_last_frame_type_assigned = m_firstframe;


  // Wenn das die letzten Frames in der Sequenz sind, dann keine Szenenwechsel mehr suchen.

  if (m_LastFrames)  // noch ergaenzen durch m_nFrames<20 ?
    {
      bool i_frame_inserted=false;

      {for (int i=m_firstframe;i<=last_frame;i++)
        {
          ImageData* img = ImageData::GetNr(i);
          assert(img);
          
          img->pichdr.m_temporal_reference = i-m_firstframe;

          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
            cout << "Frame " << i << " (temp.ref. " << img->pichdr.m_temporal_reference
                 << ") [LAST] is a : ";

          if (i-m_firstframe+1 == m_PDistance ||
              (i==last_frame && !i_frame_inserted))
            {
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
                cout << "I [after GOP]\n";
              img->pichdr.m_picture_coding_type = PICTYPE_I;
              i_frame_inserted=true;

              img->SetGOPBeginning();     // BUGFIX 23.2.99
              //img->IsClosedGOP();       // BUGFIX 23.2.99   // Es war einmal ...
            }
          else if (((i-m_firstframe+1)%m_PDistance)==0 || i==last_frame)
            {
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
                cout << "P\n";
              img->pichdr.m_picture_coding_type = PICTYPE_P;
            }
          else 
            {
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
                cout << "B\n";
              img->pichdr.m_picture_coding_type = PICTYPE_B;
            }
        }}

      m_last_frame_type_assigned = last_frame;
      return;
    }

  
  // Frame mit maximalem Mittelwertsprung suchen.

  long long maxmeandiff=0;
  int maxidx=m_firstframe+m_MinIDistance-1;
  if (maxidx>last_frame)
    maxidx=last_frame;

  long long lastmean=0;

  int mitte = (m_firstframe + last_frame+1)/2;
  bool surechange = false;

  {for (int i=m_firstframe+m_MinIDistance-1;i<=last_frame;i++)
    {
      ImageData* img = ImageData::GetNr(i);
      assert(img);

      if (i==m_firstframe+m_MinIDistance-1)
        lastmean = img->total_mean;

      long long meandiff = abs(img->total_mean - lastmean);

      if (meandiff > THRES_0*(firstframe->source->AskWidth() * firstframe->source->AskHeight()))
        {
          maxmeandiff=meandiff;
          maxidx=i;
          surechange=true;

          if (options.ProgressDisplay >= Options::ProgressDisplay_Minimal)
            cout << "Scenechange detected at frame " << maxidx << endl;

          break;
        }


      /* Gewichtungsfunktion: f(n) =  /\  */

      if (i<=mitte)
        meandiff = (meandiff + meandiff*(i-m_firstframe)/(mitte-m_firstframe))/2;
      else
        meandiff = (meandiff + meandiff*(last_frame-i)/(last_frame-mitte))/2;

      if (meandiff > maxmeandiff)
        {
          maxmeandiff=meandiff;
          maxidx=i;

          maxidx = mitte; // PROBE !
        }

      lastmean = img->total_mean;
    }}


  // Messen, wie stark der Szenensprung ist

  int strength=0;
  
  if (m_firstframe>0)
    {
      ImageData* lastimg = ImageData::GetNr(m_firstframe-1);
      assert(lastimg);

      strength = abs(firstframe->total_mean - lastimg->total_mean) /
                 (firstframe->source->AskWidth() * firstframe->source->AskHeight());

      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        cout << "Scene-change strength: " << strength << endl;
    }


  // Frames bis (exclusive) 'maxidx' mit regelmaessigem B-P-Muster ausfuellen.

  bool forbid_forward_pred=true;
  bool introductory_B = true;

  {for (int i=m_firstframe;i<maxidx;i++)
    {
      ImageData* img = ImageData::GetNr(i);
      assert(img);

      img->pichdr.m_temporal_reference = i-m_firstframe;

      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        cout << "Frame " << i << " (temp.ref. " << img->pichdr.m_temporal_reference << ") is a : ";

      if (((i-m_firstframe+1)==m_PDistance && m_firstframe!=0) || i==0)
        {
          img->pichdr.m_picture_coding_type = PICTYPE_I;
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "I [after GOP]\n";
          img->SetGOPBeginning();   // BUGFIX 23.2.99
          introductory_B=false;
        }
      else if (((i-m_firstframe+1)%m_PDistance)==0 || i==maxidx-1)
        {
          if (i==maxidx-2 || i==maxidx-3)
            {
              img->pichdr.m_picture_coding_type = PICTYPE_B;
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "(B)\n";
            }
          else
            {
              img->pichdr.m_picture_coding_type = PICTYPE_P;
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "P\n";
              introductory_B=false;
            }
        }
      else 
        {
          img->pichdr.m_picture_coding_type = PICTYPE_B;

          if (strength >= THRES_1 && introductory_B)
            {
              img->no_forward_prediction     = forbid_forward_pred;
              img->scenechange_qscale_factor = 4.0;
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "B (no forw.)\n";
            }
          else
            {
              if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal) cout << "B\n";
            }
        }
    }}

  if (maxmeandiff > THRES_1*(firstframe->source->AskWidth() * firstframe->source->AskHeight()))
    {
      bool isfirstp=true;
      for (int i=maxidx-1; ;i--)
        {
          ImageData* img = ImageData::GetNr(i);
          /*
          if (!img)
            break;
          */
          assert(img);

          if ((img->pichdr.m_picture_coding_type == PICTYPE_P && isfirstp) ||
              (img->pichdr.m_picture_coding_type == PICTYPE_B))
            {
              img->scenechange_qscale_factor = 3.0;
              isfirstp=false;
            }
          else
            break;
        }
    }

  m_last_frame_type_assigned = maxidx-1;
}
