/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/loading.hh"
#include "image/colorspaces.hh"
#include "mpeg/constants.hh"
#include "options.hh"

//#include "visualize/writeppm.hh"
//#include <fstream.h>


void Task_Loading::Execute(ProcessData& data)
{
  // static int cnt=1; cout << "TASKLOAD " << cnt << " +\n";

  ImageData* img = new ImageData;
  img->source = d_preproc.GetNextImage();

  if (options.ProgressDisplay >= Options::ProgressDisplay_Minimal)
    {
      ShowTaskInformation("load frame",data.threadnr); cerr.flush();
      cerr << "frame " << img->nr;
      if (options.ProgressDisplay < Options::ProgressDisplay_Maximal)
        cerr << "\n";
      else
        cerr << " ... ";
    }

  // img->mblks.Create((img->source->AskWidth()+15)/16 , (img->source->AskHeight()+15)/16);
  
  frame_read = img->nr;



  /* Hinweis: Hier wird zwar auf die Rot-Gruen-Blau-Bitmaps zugegriffen, wenn das Bild
     jedoch aus YUV-Bitmaps besteht, werden diese geliefert. */
  img->source->AskBitmap(Bitmap_Red  ).DuplicateBorder();
  img->source->AskBitmap(Bitmap_Green).DuplicateBorder();
  img->source->AskBitmap(Bitmap_Blue ).DuplicateBorder();

  switch (img->source->AskColorspace())
    {
    case RGB_Undefined:
      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        {
          cerr << "Colorspace transformation ...\n";
        }
      EndShowTaskInformation();
      TransformRGB2YCrCb(*img->source, CHROMA_420,false);//sh.m_chroma_format,d_encopts.m_GenerateMPEG1);
      break;
    case YCrCb:
      if (!img->source->IsSubsampledH() || !img->source->IsSubsampledV())
        {
          if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
            {
              cerr << "Color subsampling ...\n";
            }
          EndShowTaskInformation();
          SubsampleYCrCb(*img->source    , CHROMA_420,false);//sh.m_chroma_format,d_encopts.m_GenerateMPEG1);
        }
      else
        EndShowTaskInformation();
      break;
    default:
      assert(0);
      break;
    }

  //cout << "TASKLOAD " << cnt << " -\n";  cnt++;
}

