/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/mvutils.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"

static uint8  offs2codetab[2*4096];
static uint8* offs2code = &offs2codetab[4096];

static class DummyInitClass
{
public:
  DummyInitClass()
    {
      int i=0;
      for (;i<  16;i++) offs2code[i]=1;
      for (;i<  32;i++) offs2code[i]=2;
      for (;i<  64;i++) offs2code[i]=3;
      for (;i< 128;i++) offs2code[i]=4;
      for (;i< 256;i++) offs2code[i]=5;
      for (;i< 512;i++) offs2code[i]=6;
      for (;i<1024;i++) offs2code[i]=7;
      for (;i<2048;i++) offs2code[i]=8;
      for (;i<4096;i++) offs2code[i]=9;

      i=0;
      for (;i>=  -16;i--) offs2code[i]=1;
      for (;i>=  -32;i--) offs2code[i]=2;
      for (;i>=  -64;i--) offs2code[i]=3;
      for (;i>= -128;i--) offs2code[i]=4;
      for (;i>= -256;i--) offs2code[i]=5;
      for (;i>= -512;i--) offs2code[i]=6;
      for (;i>=-1024;i--) offs2code[i]=7;
      for (;i>=-2048;i--) offs2code[i]=8;
      for (;i>=-4096;i--) offs2code[i]=9;
    }
} dummy;


int  MVOffsetToFCode(int offset)
{
  return offs2code[offset];
}


int  PixelRangeToFCode(int mini,int maxi)
{
  return max(offs2code[mini],offs2code[maxi]);
}


void ReduceMVRange(ImageData& imgdata)
{
  if (imgdata.pichdr.m_picture_coding_type == PICTYPE_I)
    return;

  int fwd_hrange[10],bkw_hrange[10];
  int fwd_vrange[10],bkw_vrange[10];
  int fwd_total=0,   bkw_total=0;

  {for (int fcode=1;fcode<=9;fcode++)
    fwd_hrange[fcode]=fwd_vrange[fcode]=bkw_hrange[fcode]=bkw_vrange[fcode]=0;}


  /* Anzahl der MVs, die einen bestimmten FCode benoetigen, nach H- bzw. V-Komponente und
     Forward- bzw. Backward-Praediktion getrennt zaehlen. */

  for (int y=0;y<imgdata.mblks.AskHeight();y++)
    for (int x=0;x<imgdata.mblks.AskWidth();x++)
      {
        Macroblock& mb = imgdata.mblks.Ask(y,x);
        if (mb.m_IsIntra)
          continue;

        if (mb.m_HasMotionForward)
          {
            fwd_total++;
            fwd_hrange[offs2code[mb.m_forward1.m_habs]]++;
            fwd_vrange[offs2code[mb.m_forward1.m_vabs]]++;
          }
        if (mb.m_HasMotionBackward)
          {
            bkw_total++;
            bkw_hrange[offs2code[mb.m_backward1.m_habs]]++;
            bkw_vrange[offs2code[mb.m_backward1.m_vabs]]++;
          }
      }

  /* Jeweils maximal benoetigten FCode suchen. */

  if (fwd_total || bkw_total)
    {
      int maxfwdh=1,maxbkwh=1;
      int maxfwdv=1,maxbkwv=1;
      for (int i=9;i>=1;i--)
        {
#define TH 0

          if (fwd_hrange[i]>TH && maxfwdh<i) maxfwdh=i;
          if (fwd_vrange[i]>TH && maxfwdv<i) maxfwdv=i;
          if (bkw_hrange[i]>TH && maxbkwh<i) maxbkwh=i;
          if (bkw_vrange[i]>TH && maxbkwv<i) maxbkwv=i;
        }

      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        {
          if (fwd_total || bkw_total) cerr << "Maximum MV-fcodes used:";
          if (fwd_total) cerr << "   fwd " << maxfwdh << ";" << maxfwdv;
          if (bkw_total) cerr << "   bkw " << maxbkwh << ";" << maxbkwv;
          if (fwd_total || bkw_total) cerr << endl;
        }


      // Wenn wir MPEG-1 Streams erzeugen, muss der horizontale und vertikale FCode identische sein.

      if (!options.CreateMPEG2)
        {
          maxfwdh = maxfwdv = max(maxfwdh,maxfwdv);
          maxbkwh = maxbkwv = max(maxbkwh,maxbkwv);
        }
      
      imgdata.pichdr.m_f_code[0][0] = maxfwdh;
      imgdata.pichdr.m_f_code[0][1] = maxfwdv;
      imgdata.pichdr.m_f_code[1][0] = maxbkwh;
      imgdata.pichdr.m_f_code[1][1] = maxbkwv;

      /* MPEG-1
      assert(imgdata.pichdr.m_f_code[0][0] == imgdata.pichdr.m_f_code[0][1]);
      assert(imgdata.pichdr.m_f_code[1][0] == imgdata.pichdr.m_f_code[1][1]);
      */


      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        {
          int maxfcode = 9; // max(max(maxfwdh,maxfwdv),max(maxbkwh,maxbkwv));
          for (int i=1;i<=maxfcode;i++)
            {
              cerr << fwd_hrange[i]<< ':' << fwd_vrange[i] << '/'
                   << bkw_hrange[i]<< ':' << bkw_vrange[i] << ' ';
            }
          cerr << endl;
        }

#if TH != 0
      // MVs ggf. entfernen, falls sie ausserhalb des zulaessigen Bereichs sind.

      for (int y=0;y<imgdata.mblks.AskHeight();y++)
        for (int x=0;x<imgdata.mblks.AskWidth();x++)
          {
            Macroblock& mb = imgdata.mblks.Ask(y,x);
            if (mb.m_IsIntra)
              continue;
            
            if (mb.m_HasMotionForward)
              {
                if (offs2code[mb.m_forward1.m_habs] > maxfwdh ||
                    offs2code[mb.m_forward1.m_vabs] > maxfwdv)
                  {
                    mb.m_IsIntra=true;
                    mb.m_HasMotionForward =false;
                    mb.m_HasMotionBackward=false;
                  }
              }
            if (mb.m_HasMotionBackward)
              {
                if (offs2code[mb.m_backward1.m_habs] > maxbkwh ||
                    offs2code[mb.m_backward1.m_vabs] > maxbkwv)
                  {
                    mb.m_IsIntra=true;
                    mb.m_HasMotionForward =false;
                    mb.m_HasMotionBackward=false;
                  }
              }

      assert(imgdata.pichdr.m_f_code[0][0] == imgdata.pichdr.m_f_code[0][1]);
      assert(imgdata.pichdr.m_f_code[1][0] == imgdata.pichdr.m_f_code[1][1]);
          }
#endif
    }
}

int  FCodeToMaxOffset(int fcode) { return (8<<fcode)-1; }
int  FCodeToMinOffset(int fcode) { return -(8<<fcode); }
