/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/quantscale.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"


void Task_QuantizationScale_Constant::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("qscale (constant)",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  img->pichdr.m_q_scale_type = (options.mpeg2.ExponentialQScale ? 1 : 0);

  int w = img->mblks.AskWidth();
  int h = img->mblks.AskHeight();

  int qscale=0;

  switch (img->pichdr.m_picture_coding_type)
    {
    case PICTYPE_I: qscale=options.qscale_const.qscale_I; break;
    case PICTYPE_P: qscale=options.qscale_const.qscale_P; break;
    case PICTYPE_B: qscale=options.qscale_const.qscale_B; break;
    }

  assert(qscale);

  for (int y=0;y<h;y++)
    for (int x=0;x<w;x++)
      {
        img->mblks.Ask(y,x).m_quantiser_scale = qscale;
      }
}
