/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "fuss0.typs"

#include "coding/taskcode.hh"
//#include "coding/decoder.hh"
#include "coding/dct.hh"
#include "coding/taskmvsearch.hh"
#include "coding/taskdiffpics.hh"
#include "coding/quantscale.hh"
#include "coding/taskmbtype.hh"
#include "coding/mvutils.hh"
#include "coding/ratecontrol.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"
#include "statistics/bitsestim.hh"
#include "statistics/statistics.hh"
#include "mpeg/syntax.hh"
#include "visualize/writeppm.hh"
#include "global.hh"

#include <math.h>
#include <fstream.h>


void Task_CodeFrame::Execute(ProcessData& procdata)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Minimal)
    {
      ShowTaskInformation("encode frame",procdata.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  globalstats.SetCurrentFrameNr(m_framenr);



  /* Picture-Header ausfuellen */

  img->pichdr.m_vbv_delay = 0xFFFF;
  
  img->pichdr.m_full_pel_forward_vector  = false;
  img->pichdr.m_full_pel_backward_vector = false;

  img->pichdr.m_f_code[0][0]=1;//MVOffsetToFCode(-options.HSearchRange *2); // *2 wegen Half-Pel
  img->pichdr.m_f_code[0][1]=1;//MVOffsetToFCode(-options.VSearchRange *2);
  img->pichdr.m_f_code[1][0]=1;//MVOffsetToFCode(-options.HSearchRange *2);
  img->pichdr.m_f_code[1][1]=1;//MVOffsetToFCode(-options.VSearchRange *2);

  img->pichdr.m_intra_dc_precision= options.mpeg2.DC_Precision;
  img->pichdr.m_picture_structure = PICSTRUCT_FramePicture;

  img->pichdr.m_top_field_first = true;
  img->pichdr.m_frame_pred_frame_dct = true;
  img->pichdr.m_concealment_motion_vectors = false;
  img->pichdr.m_intra_vlc_format = options.mpeg2.IntraVLCFormat;
  img->pichdr.m_alternate_scan   = (options.mpeg2.AlternateScan ? 1 : 0);
  img->pichdr.m_repeat_first_field = false;
  img->pichdr.m_progressive_frame  = true;
  
  img->pichdr.m_composite_display_flag = false;


  /* Makrobloecke initisialisieren und in Slices einteilen */

  int MB_Width =(img->source->AskWidth() +15)/16;
  int MB_Height=(img->source->AskHeight()+15)/16;

  img->mblks.Create(MB_Width , MB_Height);

  {for (int y=0;y<MB_Height;y++)
    for (int x=0;x<MB_Width;x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);

        mb.m_IsIntra=true;
        mb.m_HasMotionForward =false;
        mb.m_HasMotionBackward=false;
        mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);


        // Sliceeinteilung fuer MPEG2   (auch fuer MPEG1 benutzen)

        if (x==0)          mb.m_BeginSlice=true;
        if (x==MB_Width-1) mb.m_EndSlice  =true;
      }}


  // QScales setzen (Initialisierung)

  Task_QuantizationScale_Constant task_qscale(img->nr);
  task_qscale.Execute(procdata);


  // MEData erstellen, wenn benoetigt

  if (img->pichdr.m_picture_coding_type != PICTYPE_I)
    {
      SearchParam sparam;
      sparam.minx=-options.HSearchRange*2;
      sparam.maxx= options.HSearchRange*2-1;
      sparam.miny=-options.VSearchRange*2;
      sparam.maxy= options.VSearchRange*2-1;

      sparam.halfpel     = true; // !codedimg_PB.m_pichdr.m_full_pel_forward_vector;
      sparam.framesearch = true;
      sparam.top_field   = true;
      sparam.blk_width   = 16;
      sparam.blk_height  = 16;

      MotionEstimator* mest = img->GetMotionEstimator();

      if (img->prev_ip && img->prev_ip->medata==NULL)
        {
          sparam.bitmap = &(img->prev_ip->decoded.AskBitmap_const(Bitmap_Y));
          assert(sparam.bitmap);
          img->prev_ip->medata = mest->CreateMEData(sparam,img->prev_ip->medata);
        }

      if (img->next_ip && img->next_ip->medata==NULL)
        {
          sparam.bitmap = &(img->next_ip->decoded.AskBitmap_const(Bitmap_Y));
          assert(sparam.bitmap);
          img->next_ip->medata = mest->CreateMEData(sparam,img->next_ip->medata);
        }

      img->GiveBackMotionEstimator(mest);
    }


  // MVs suchen

  switch (img->pichdr.m_picture_coding_type)
    {
    case PICTYPE_I:
      break;
    case PICTYPE_B:
      {
        Task_MVSearch task_mvsearch_bkw(img->nr,false);
        task_mvsearch_bkw.Execute(procdata);
      }
      // fall-through
    case PICTYPE_P:
      {
        Task_MVSearch task_mvsearch_fwd(img->nr,true);
        task_mvsearch_fwd.Execute(procdata);
      }
      break;
    }


  // Differenzbilder erstellen

  CreateDifferencePictures(img);


  // MB-Typ bestimmen

  Task_MBType_Variance task_mbtype(img->nr);
  task_mbtype.Execute(procdata);



  /* Statistiken */


  const Pixel*const* pix = img->source->AskBitmap_const(Bitmap_Y).AskFrame_const();

#if 0
  Array2<int32> deriv;
  deriv.Create(img->mblks.AskWidth(),
               img->mblks.AskHeight());

  // spatial derivatives berechnen

  {for (int32 y=0;y<MB_Height;y++)
    for (int32 x=0;x<MB_Width;x++)
      {
        uint32 this_dh=0;
        uint32 this_dv=0;

        for (int yy=0;yy<16;yy++)
          for (int xx=0;xx<15;xx++)
            this_dh += abs(pix[y*16+yy][x*16+xx] - pix[y*16+yy][x*16+xx+1]);
            
        for (int yy=0;yy<15;yy++)
          for (int xx=0;xx<16;xx++)
            this_dv += abs(pix[y*16+yy][x*16+xx] - pix[y*16+yy+1][x*16+xx]);
        
        deriv.Ask(y,x) = this_dh + this_dv;
      }}

  {for (int y=0;y<MB_Height;y++)
    for (int x=0;x<MB_Width;x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);
        
        int32 val = deriv.Ask(y,x)/(15*16*2);
        // cout << tanh(val-12) << endl;

        val -= 10;
        if (val<-10) val=-10;
        if (val> 10) val= 10;

        mb.m_quantiser_scale += val; // 10.0*tanh(val-10); // 15
        /*
          if (val>= 8) { mb.m_quantiser_scale=50; }
          else         { mb.m_quantiser_scale= 2; }
        */
      }}
#endif

#if 0
  {
    {for (int y=0;y<MB_Height;y++,cout << endl)
      for (int x=0;x<MB_Width;x++)
        {
          Macroblock& mb = img->mblks.Ask(y,x);
          
          cout << setw(3) << (int)mb.m_quantiser_scale+9;
        }}
  }
#endif

#if 0
  {
  Task_DCTQuantize task_dct(img->nr);
  task_dct.Execute(procdata);

    Array2<uint8> typs;
    typs.Create(MB_Width,MB_Height);

    Array2<uint8> btyps;
    btyps.Create(2*MB_Width,2*MB_Height);

    Array2<double> variance;
    variance.Create(2*MB_Width,2*MB_Height);

    const Pixel*const* pix = img->source->AskBitmap_const(Bitmap_Y).AskFrame_const();
    {for (int y=0;y<2*MB_Height;y++)
      for (int x=0;x<2*MB_Width;x++)
        {
          // ----------- 8x8 Mittelwert -----------

          double mean=0.0;
          
          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<8;xx++)
              mean += pix[8*y+yy][8*x+xx];
          }
          mean /= 8*8;


          // -------- 8x8 Varianz ---------

          double var=0.0;

          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<8;xx++)
              var += (pix[8*y+yy][8*x+xx]-mean) * (pix[8*y+yy][8*x+xx]-mean);
          }
          var /= 8*8;

          variance.Ask(y,x) = var;


          // ------- Varianz auf Mittelwerte der 4x4 Subbloecke ------

          double var4=0.0;
          double mean4[4]; for (int i=0;i<4;i++) { mean4[i]=0; }
          {for (int yy=0;yy<4;yy++)
            for (int xx=0;xx<4;xx++)
              {
                mean4[0] += pix[8*y+yy  ][8*x+xx  ];
                mean4[1] += pix[8*y+yy+4][8*x+xx  ];
                mean4[2] += pix[8*y+yy+4][8*x+xx+4];
                mean4[3] += pix[8*y+yy  ][8*x+xx+4];
              }
          }
          {for (int i=0;i<4;i++) mean4[i] /= 4*4;}

          double totalmean4 = (mean4[0]+mean4[1]+mean4[2]+mean4[3])/4.0;
          {for (int i=0;i<4;i++)
            var4 += (mean4[i]-totalmean4)*(mean4[i]-totalmean4);}
          var4 /= 4;

          // ------- 8x8 Betrag der horiz. und vertik. Richtungsableitung ----

          double dh=0.0 , dv=0.0;

          int nHl=0,nHh=0;
          int nVl=0,nVh=0;

          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<6;xx++)
              {
                double d = abs(pix[8*y+yy][8*x+xx+1] - pix[8*y+yy][8*x+xx]);
                dh += d;
                if (d>60) nHh++;
                if (d<40) nHl++;
              }}
          dh /= 7*8;

          {for (int yy=0;yy<6;yy++)
            for (int xx=0;xx<8;xx++)
              {
                double d = abs(pix[8*y+yy+1][8*x+xx] - pix[8*y+yy][8*x+xx]);
                dv += d;
                if (d>60) nVh++;
                if (d<40) nVl++;
              }}
          dv /= 7*8;

          double deriv=(dh+dv)/2.0;


          // ------------ 8x8 Maximum, Minimum ----------

          uint8 maxi = 0,mini = 255;
          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<8;xx++)
              {
                maxi = max(maxi,pix[8*y+yy][8*x+xx]);
                mini = min(mini,pix[8*y+yy][8*x+xx]);
              }}

          double range = maxi-mini+1;


          // --------------------------------

          uint8 nGreater=0,nLess=0;
          {for (int yy=0;yy<8;yy++/*,cout << endl*/)
            for (int xx=0;xx<8;xx++)
              {
                     if (pix[8*y+yy][8*x+xx]>(maxi-mini)*2/3+mini) {/*cout<<"x";*/ nGreater++;}
                else if (pix[8*y+yy][8*x+xx]>(maxi-mini)*1/3+mini) {/*cout<<"-";*/ }
                else { /*cout <<".";*/ nLess++;}
              }}

          uint8 e[4];
          e[0] = pix[8*y  ][8*x  ];
          e[1] = pix[8*y  ][8*x+7];
          e[2] = pix[8*y+7][8*x  ];
          e[3] = pix[8*y+7][8*x+7];

          uint8 cnt[4]; cnt[0]=cnt[1]=cnt[2]=cnt[3]=0;
          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<8;xx++)
              for (int i=0;i<4;i++)
                {
                  if (abs(pix[8*y+yy][8*x+xx] - e[i]) < 20)
                    cnt[i]++;
                }}

          double spatderiv[7*7];
          double fact = 255.0/range;
          int idx=0;
          {for (int yy=0;yy<7;yy++)
            for (int xx=0;xx<7;xx++)
              {
                int d = (int)fact*(2*pix[8*y+yy  ][8*x+xx  ]
                                   - pix[8*y+yy+1][8*x+xx  ]
                                   - pix[8*y+yy  ][8*x+xx+1]);
                spatderiv[idx++] = d;
                //cout << setw(4) << d << " ";
              }}
          double spatderivmean=0;
          double spatderivvar=0;
          for (int i=0;i<7*7;i++) spatderivmean+=spatderiv[i]; spatderivmean /= 7*7;
          for (int i=0;i<7*7;i++) spatderivvar +=(spatderiv[i]-spatderivmean)*(spatderiv[i]-spatderivmean);
          spatderivvar /= 7*7;


          double hmean[8],hvar[8];
          double vmean[8],vvar[8];
          {for (int i=0;i<8;i++) { hmean[i]=hvar[i]=0.0; vmean[i]=vvar[i]=0.0; }}
          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<8;xx++)
              {
                hmean[yy] += pix[8*y+yy][8*x+xx];
                vmean[xx] += pix[8*y+yy][8*x+xx];
              }}
          {for (int i=0;i<8;i++) { hmean[i] /= 8; vmean[i] /= 8; }}
          {for (int yy=0;yy<8;yy++)
            for (int xx=0;xx<8;xx++)
              {
                hvar[yy] += (pix[8*y+yy][8*x+xx]-hmean[yy])*(pix[8*y+yy][8*x+xx]-hmean[yy]);
                vvar[xx] += (pix[8*y+yy][8*x+xx]-vmean[xx])*(pix[8*y+yy][8*x+xx]-vmean[xx]);
              }}

          double vvarmax=0.0,hvarmax=0.0;
          {for (int i=0;i<8;i++)
            {
              vvarmax += vvar[i];
              hvarmax += hvar[i];
            }}

          vvarmax *= fact*fact;
          hvarmax *= fact*fact;

          if (img->nr==0)
            {
              cout.setf(ios::showpoint);
              cout.setf(ios::fixed);
              //cout << fussb0[x+y*384/8] << " ";
              cout << setw(2) << werb24_0[2*(x+y*368/8)  ] << " ";
              cout << setw(2) << werb24_0[2*(x+y*368/8)+1] << " ";
              cout << setw(3) << x << " " << setw(3) << y << " ";
              cout << setprecision(1);
              cout << setw(5) << range << " ";
              cout << setprecision(3);
              cout << setw(8) << var   << " " << setw(8) << var4 << " " << setw(8) << var/var4 << " ";
              cout << setw(8) << deriv << " ";
              cout << setw(4) << cnt[0]+cnt[1]+cnt[2]+cnt[3] << " ";
              cout << setw(10) << spatderivmean << " " << setw(10) << spatderivvar << " ";
              cout << setw(8) << vvarmax << " " << hvarmax << " " << min(vvarmax,hvarmax) << " ";
              int nr = 0;
              if (x&1) nr++;
              if (y&1) nr+=2;
              cout << setw(3) << img->mblks.Ask(y/2,x/2).m_blks[nr].m_coeff[1][0] << " ";
              cout << setw(3) << img->mblks.Ask(y/2,x/2).m_blks[nr].m_coeff[0][1] << " ";
              cout << setw(3) << img->mblks.Ask(y/2,x/2).m_blks[nr].m_coeff[2][2] << " ";
              cout << endl;
            }
#endif

#if 0
               if (var < 100)                      btyps.Ask(y,x) = BLK_BLUE|BLK_RECT;
          else if (deriv*7*8 < 8*range && var>700) btyps.Ask(y,x) = BLK_RED |BLK_RECT;
          else
            btyps.Ask(y,x) = 0;

          cout << nHh << "/" << nHl << "  " << nVh << "/" << nVl << endl;
#endif

#if 0
          btyps.Ask(y,x) = 0;
#endif

#if 0
          //switch (fussb0[x+y*384/8])
          switch (werb24_0[2*(x+y*368/8)])
            {
            case 'f': btyps.Ask(y,x) = BLK_BLUE |BLK_RECT;  break;
            case 'e': btyps.Ask(y,x) = BLK_GREEN|BLK_RECT;  break;
            case 't': btyps.Ask(y,x) = BLK_RED  |BLK_CROSS; break;
            }
#endif

#if 0
          if (1)
            btyps.Ask(y,x) = BLK_BLUE|BLK_DOTS;
          else
#endif

#if 0
          if (range<50) // FLAT
            btyps.Ask(y,x) = BLK_BLUE|BLK_RECT;
          //          else if (range < 110) // EDGE
          else if (range>80) // spatderivvar<3000) // if (deriv > 300-var*300/1000)
            btyps.Ask(y,x) = BLK_GREEN|BLK_RECT;
          else // TEXTURE
            btyps.Ask(y,x) = BLK_RED|BLK_CROSS;
#endif

#if 0
          if (range<50) // FLAT
            btyps.Ask(y,x) = BLK_BLUE|BLK_DOTS;
          else if (min(vvarmax,hvarmax)<70000 && maxi>120)
            btyps.Ask(y,x) = BLK_GREEN|BLK_DOTS;
          else // TEXTURE
            btyps.Ask(y,x) = BLK_RED|BLK_DOTS;
#endif

#if 0
          if (1) { }
          else
            {
              if (var >= 750)
                btyps.Ask(y,x) |= BLK_RED|BLK_LINE;

              if (cnt[0]+cnt[1]+cnt[2]+cnt[3] >   64 &&
                  cnt[0]+cnt[1]+cnt[2]+cnt[3] < 3*64 && range>60)
                btyps.Ask(y,x) |= BLK_GREEN|BLK_LINE2;
            }
#endif
          /*
          else if (var > 500 && var < 750)
            btyps.Ask(y,x) = BLK_BLUE|BLK_CROSS;
          else if (var >= 750)
            btyps.Ask(y,x) = BLK_GREEN|BLK_CROSS;
          else if (maxi-mini < 25)
            btyps.Ask(y,x) = BLK_RED|BLK_GREEN|BLK_RECT;
          else
            btyps.Ask(y,x) = BLK_GREEN|BLK_DOTS;
          */
#if 0
        }}
#endif

#if 0
    {for (int y=0;y<MB_Height;y++)
      for (int x=0;x<MB_Width;x++)
        {
          int nEdge=0;
          int nText=0;
          int nFlat=0;

               if ((btyps.Ask(2*y  ,2*x  ) & 0x07) == BLK_BLUE)  nFlat++;
          else if ((btyps.Ask(2*y  ,2*x  ) & 0x07) == BLK_GREEN) nEdge++;
          else if ((btyps.Ask(2*y  ,2*x  ) & 0x07) == BLK_RED)   nText++;
          else { assert(0); }

               if ((btyps.Ask(2*y  ,2*x+1) & 0x07) == BLK_BLUE)  nFlat++;
          else if ((btyps.Ask(2*y  ,2*x+1) & 0x07) == BLK_GREEN) nEdge++;
          else if ((btyps.Ask(2*y  ,2*x+1) & 0x07) == BLK_RED)   nText++;
          else { assert(0); }

               if ((btyps.Ask(2*y+1,2*x  ) & 0x07) == BLK_BLUE)  nFlat++;
          else if ((btyps.Ask(2*y+1,2*x  ) & 0x07) == BLK_GREEN) nEdge++;
          else if ((btyps.Ask(2*y+1,2*x  ) & 0x07) == BLK_RED)   nText++;
          else { assert(0); }

               if ((btyps.Ask(2*y+1,2*x+1) & 0x07) == BLK_BLUE)  nFlat++;
          else if ((btyps.Ask(2*y+1,2*x+1) & 0x07) == BLK_GREEN) nEdge++;
          else if ((btyps.Ask(2*y+1,2*x+1) & 0x07) == BLK_RED)   nText++;
          else { assert(0); }

          assert(nText+nEdge+nFlat==4);

               if (nFlat >= 2)     typs.Ask(y,x)=BLK_BLUE  | BLK_RECT;
          else if (nEdge<=2 && nEdge+nText>=3) typs.Ask(y,x)=BLK_RED   | BLK_RECT;
          else if (nEdge >= 2)     typs.Ask(y,x)=BLK_GREEN | BLK_RECT;
          else                     typs.Ask(y,x) = BLK_RED | BLK_RECT;
        }}
#endif

#if 0
    const double T1 =  90.0;
    const double T2 = 120.0;
    const double T3 = 2;
    const double T4 = 4;
    const double T5 = 3;

    {for (int y=0;y<MB_Height;y++)
      for (int x=0;x<MB_Width;x++)
        {
          typs.Ask(y,x) = 0;

          double v1,v2,v3,v4;
          v1 = variance.Ask(y*2  ,x*2  );
          v2 = variance.Ask(y*2+1,x*2  );
          v3 = variance.Ask(y*2  ,x*2+1);
          v4 = variance.Ask(y*2+1,x*2+1);

          double v_mean = (v1+v2+v3+v4)/4.0;
          double v_max  = max(max(v1,v2),max(v3,v4));
          double v_min  = min(min(v1,v2),min(v3,v4));

          // v_i der Groesse nach sortieren

               if (v2==v_min) Swap(v2,v1);
          else if (v3==v_min) Swap(v3,v1);
          else if (v4==v_min) Swap(v4,v1);

               if (v2==v_max) Swap(v2,v4);
          else if (v3==v_max) Swap(v3,v4);

          if (v2>v3) Swap(v2,v3);

          assert(v1<=v2);
          assert(v2<=v3);
          assert(v3<=v4);

          if (v_mean < T1 && v_max < T2)
            typs.Ask(y,x) = BLK_BLUE|BLK_RECT;

          int nT3=0,nT4=0,nT5=0;

          if (v_max/v1 < T3) { nT3++; } else if (v_max/v1 < T4) nT4++;
          if (v_max/v2 < T3) { nT3++; } else if (v_max/v2 < T4) nT4++;
          if (v_max/v3 < T3) { nT3++; } else if (v_max/v3 < T4) nT4++;

          if (v_max/v1 < T5) { nT5++; }
          if (v_max/v2 < T5) { nT5++; }
          if (v_max/v3 < T5) { nT5++; }

          if (v_min > T1 && ((nT3==2 && nT4==1) || nT5==3))
            typs.Ask(y,x) = BLK_RED|BLK_RECT;
        }}
#endif

#if 0
    {for (int y=0;y<MB_Height;y++)
      for (int x=0;x<MB_Width;x++)
        {
          Macroblock& mb = img->mblks.Ask(y,x);
          uint8& typ = typs.Ask(y,x);

          if (img->variance.Ask(y,x)/(16*16) < 40)
            {
              typ = BLK_RED|BLK_GREEN|BLK_RECT;
            }
          else
          switch ((int)(mb.m_quantiser_scale+9))
            {
            case  0:
            case  1:
            case  2:
            case  3:
              typ = 0;
              break;
            case  4:
            case  5:
            case  6:
            case  7:
              typ = BLK_RED|BLK_LINE;
              break;
            case  8:
            case  9:
            case 10:
            case 11:
              typ = BLK_RED|BLK_CROSS;
              break;
            case 12:
            case 13:
            case 14:
            case 15:
              typ = BLK_RED|BLK_RECT;
              break;
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
              typ = BLK_GREEN|BLK_CROSS|BLK_DOTS;
              break;
            }
        }}
#endif

#if 0
    ofstream vis;
    OpenStream(vis,"out/mbt%03d.ppm",img->nr);

    Image<Pixel>* visimg = CreateVisualizationImage(*img->source,false);
    FillBlockClass(*visimg,btyps,false);
    FillBlockClass(*visimg,typs,true);
    WriteImageAsPPM(*visimg,vis);
    delete visimg;
#endif


#if 0
    if (options.Enable_JND)
      {
        // ----------- JND - Map aufbauen ------------------

        static const double jnd_pattern[8][8] =
        {
          { 0.0, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.0 },
          { 0.1, 0.2, 0.3, 0.3, 0.3, 0.3, 0.2, 0.1 },
          { 0.1, 0.3, 0.6, 0.7, 0.7, 0.6, 0.3, 0.1 },
          { 0.1, 0.3, 0.7, 1.0, 1.0, 0.7, 0.3, 0.1 },
          { 0.1, 0.3, 0.7, 1.0, 1.0, 0.7, 0.3, 0.1 },
          { 0.1, 0.3, 0.6, 0.7, 0.7, 0.6, 0.3, 0.1 },
          { 0.1, 0.2, 0.3, 0.3, 0.3, 0.3, 0.2, 0.1 },
          { 0.0, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.0 }
        };

        int PixHeight=MB_Height*16;
        int PixWidth =MB_Width *16;
        Array2<double> jnd; jnd.Create(PixWidth , PixHeight);
        {for (int y=0;y<PixHeight;y++)
          for (int x=0;x<PixWidth;x++)
            jnd.Ask(y,x)=0.0;}
        
        {for (int y=4;y<PixHeight-5;y++)
          for (int x=4;x<PixWidth-5;x++)
            {
              double mag = abs(pix[y][x]-pix[y][x+1]) + abs(pix[y][x]-pix[y+1][x]);
              mag /= 200.0;
              if (mag>1.0) mag=1.0;
              mag /= 2;

              for (int yy=0;yy<8;yy++)
                for (int xx=0;xx<8;xx++)
                  {
                    double& j = jnd.Ask(y-3+yy,x-3+xx);
                    //j = 1 - (1 - j)*(1 - jnd_pattern[xx][yy]*mag);
                    double pat = jnd_pattern[xx][yy]*mag;
                    j += pat-j*pat;
                  }
            }}

#if 0
        ofstream vis2;
        OpenStream(vis2,"out/jnd%03d.ppm",img->nr);
        Image<Pixel> vis2img; vis2img.CreateImage(RGB_Undefined,PixWidth,PixHeight);
        {for (int y=0;y<PixHeight;y++)
          for (int x=0;x<PixWidth;x++)
            {
              vis2img.AskBitmap(Bitmap_Red)  .AskFrame()[y][x] = (int)(255*jnd.Ask(y,x));
              vis2img.AskBitmap(Bitmap_Green).AskFrame()[y][x] = (int)(255*jnd.Ask(y,x));
              vis2img.AskBitmap(Bitmap_Blue) .AskFrame()[y][x] = (int)(255*jnd.Ask(y,x));
            }}
        WriteImageAsPPM(vis2img,vis2);
#endif


        // QScale entsprechend JND setzen

        //cout << "JND -> QScale\n";

        {for (int y=0;y<MB_Height;y++)
          for (int x=0;x<MB_Width;x++)
            {
              double jndmean=0.0;
              for (int xx=0;xx<16;xx++)
                for (int yy=0;yy<16;yy++)
                  jndmean += jnd.Ask(y*16+yy,x*16+xx);
              jndmean /= 16*16;

              double ampl = 8;
              double qscale_offset = jndmean*(2*ampl+1)-ampl;
              //cout << setw(4) << qscale_offset << " ";
              img->mblks.Ask(y,x).m_quantiser_scale += qscale_offset;
            }}
      }
#else
    if (options.Enable_JND)
      {
        // ----------- JND - Map aufbauen ------------------

        static const int jnd_pattern[8][8] =
        {
          {   0,  10,  10,  10,  10,  10,  10,   0 },
          {  10,  20,  30,  30,  30,  30,  20,  10 },
          {  10,  30,  60,  70,  70,  60,  30,  10 },
          {  10,  30,  70, 100, 100,  70,  30,  10 },
          {  10,  30,  70, 100, 100,  70,  30,  10 },
          {  10,  30,  60,  70,  70,  60,  30,  10 },
          {  10,  20,  30,  30,  30,  30,  20,  10 },
          {   0,  10,  10,  10,  10,  10,  10,   0 }
        };

        int PixHeight=MB_Height*16;
        int PixWidth =MB_Width *16;
        Array2<int> jnd; jnd.Create(PixWidth , PixHeight);
        {for (int y=0;y<PixHeight;y++)
          for (int x=0;x<PixWidth;x++)
            jnd.Ask(y,x)=0;}
        
        {for (int y=4;y<PixHeight-5;y++)
          for (int x=4;x<PixWidth-5;x++)
            {
              int mag = abs(pix[y][x]-pix[y][x+1]) + abs(pix[y][x]-pix[y+1][x]);
              if (mag>200) mag=200;
              mag /= 2;

              for (int yy=0;yy<8;yy++)
                for (int xx=0;xx<8;xx++)
                  {
                    int&  j = jnd.Ask(y-3+yy,x-3+xx);
                    int pat = jnd_pattern[xx][yy]*mag/256;
                    j += pat-j*pat/128;
                  }
            }}

#if 0
        ofstream vis2;
        OpenStream(vis2,"out/jnd%03d.ppm",img->nr);
        Image<Pixel> vis2img; vis2img.CreateImage(RGB_Undefined,PixWidth,PixHeight);
        {for (int y=0;y<PixHeight;y++)
          for (int x=0;x<PixWidth;x++)
            {
              vis2img.AskBitmap(Bitmap_Red)  .AskFrame()[y][x] = (int)(255*jnd.Ask(y,x)/128);
              vis2img.AskBitmap(Bitmap_Green).AskFrame()[y][x] = (int)(255*jnd.Ask(y,x)/128);
              vis2img.AskBitmap(Bitmap_Blue) .AskFrame()[y][x] = (int)(255*jnd.Ask(y,x)/128);
            }}
        WriteImageAsPPM(vis2img,vis2);
#endif


        // QScale entsprechend JND setzen

        //cout << "JND(INT) -> QScale\n";

        {for (int y=0;y<MB_Height;y++)
          for (int x=0;x<MB_Width;x++)
            {
              double jndmean=0.0;
              for (int xx=0;xx<16;xx++)
                for (int yy=0;yy<16;yy++)
                  jndmean += jnd.Ask(y*16+yy,x*16+xx);

              double ampl = 8;
              double qscale_offset = jndmean*(2*ampl+1)/(128*16*16)-ampl;
              //cout << setw(4) << qscale_offset << " ";
              img->mblks.Ask(y,x).m_quantiser_scale += qscale_offset;
            }}
      }
#endif

    if (options.Enable_ActivityMasking)
      {
        Array2<int32> mbval;
        mbval.Create(MB_Width,MB_Height);
        Array2<int32> mbvalneg;
        mbvalneg.Create(MB_Width,MB_Height);
        Array2<int32> mbbusy;
        mbbusy.Create(MB_Width,MB_Height);
        Array2<int32> mbring;
        mbring.Create(MB_Width,MB_Height);
        Array2<int32> actarr;
        actarr.Create(MB_Width*2,MB_Height*2);
        Array2<int32> subactarr;
        subactarr.Create(MB_Width*4,MB_Height*4);

        {for (int y=0;y<MB_Height;y++)
          for (int x=0;x<MB_Width;x++)
            {
              double fact[4];

              for (int yy=0;yy<2;yy++)
                for (int xx=0;xx<2;xx++)
                  {
                    int dh[8], dv[8];

                    for (int i=0;i<8;i++) { dh[i]=dv[i]=0; }

                    // spat. deriv. der 4 Eckbloecke berechnen

                    {for (int yyy=0;yyy<2;yyy++)
                      for (int xxx=0;xxx<2;xxx++)
                        {
                          {for (int yyyy=0;yyyy<4;yyyy++)
                            for (int xxxx=0;xxxx<3;xxxx++)
                              dh[2*yyy+xxx] += abs(pix[y*16+yy*8+yyy*4+yyyy][x*16+xx*8+xxx*4+xxxx]   -
                                                   pix[y*16+yy*8+yyy*4+yyyy][x*16+xx*8+xxx*4+xxxx+1]);
                          }
                          {for (int yyyy=0;yyyy<3;yyyy++)
                            for (int xxxx=0;xxxx<4;xxxx++)
                              dv[2*yyy+xxx] += abs(pix[y*16+yy*8+yyy*4+yyyy  ][x*16+xx*8+xxx*4+xxxx] -
                                                   pix[y*16+yy*8+yyy*4+yyyy+1][x*16+xx*8+xxx*4+xxxx]);
                          }
                        }}
                    

                    // v-spat. deriv. von zwei Subbloecken

                    {for (int xxx=0;xxx<2;xxx++)
                      {
                        {for (int yyyy=0;yyyy<3;yyyy++)
                          for (int xxxx=0;xxxx<4;xxxx++)
                            dv[4+xxx] += abs(pix[y*16+yy*8+yyyy+4][x*16+xx*8+xxx*4+xxxx] -
                                             pix[y*16+yy*8+yyyy+5][x*16+xx*8+xxx*4+xxxx]);
                        }
                      }}

                    // h-spat. deriv. von zwei Subbloecken

                    {for (int yyy=0;yyy<2;yyy++)
                      {
                        {for (int yyyy=0;yyyy<4;yyyy++)
                          for (int xxxx=0;xxxx<3;xxxx++)
                            dh[6+yyy] += abs(pix[y*16+yy*8+yyy*4+yyyy][x*16+xx*8+xxxx+4] -
                                             pix[y*16+yy*8+yyy*4+yyyy][x*16+xx*8+xxxx+5]);
                        }
                      }}


//#define tlog(x) log(1+(x)/10)
#define tlog(x) log((x)+1)
                    double l[8];
                    for (int i=0;i<4;i++)
                      l[i] = (tlog(((double)dh[i])/12) + tlog(((double)dv[i])/12))/2;
                    l[4] = tlog(((double)dv[4])/12);
                    l[5] = tlog(((double)dv[5])/12);
                    l[6] = tlog(((double)dh[6])/12);
                    l[7] = tlog(((double)dh[7])/12);
#if 0
for (int i=0;i<8;i++)
  cout << l[i] << endl;
 cout << endl;
#endif
                    {for (int i=1;i<8;i++) if (l[i]<0) l[i]=0;}

                    double lmin = l[0];
                    {for (int i=1;i<8;i++) lmin = min(lmin,l[i]);}
                    double lsum = 0.0;
                    {for (int i=0;i<8;i++) lsum += l[i];}
                    double lmean = lsum/8;

#define C1 0.01
#define C2 0
                    fact[2*yy+xx] = C1*(lmin - C2*(lmean-lmin));

                    subactarr.Ask(4*y+2*yy  ,4*x+2*xx  ) = l[0]*220;
                    subactarr.Ask(4*y+2*yy  ,4*x+2*xx+1) = l[1]*220;
                    subactarr.Ask(4*y+2*yy+1,4*x+2*xx  ) = l[2]*220;
                    subactarr.Ask(4*y+2*yy+1,4*x+2*xx+1) = l[3]*220;

                    actarr.Ask(2*y+yy,2*x+xx) = (int)(50* fact[2*yy+xx]);

#if 0
                    cout << setprecision(3);
                    cout << lmin << "\t" << lsum/8 << "\t" << fact[2*yy+xx] << endl;
#endif
                  }

              double meanfact = min(min(fact[0],fact[1]),min(fact[2],fact[3]))/4;
              //cout << meanfact << endl;
              //img->mblks.Ask(y,x).m_quantiser_scale += meanfact;
#if 0
              actarr.Ask(2*y+0,2*x+0) = (meanfact>0 ? 1000 : 0);
              actarr.Ask(2*y+0,2*x+1) = (meanfact>0 ? 1000 : 0);
              actarr.Ask(2*y+1,2*x+0) = (meanfact>0 ? 1000 : 0);
              actarr.Ask(2*y+1,2*x+1) = (meanfact>0 ? 1000 : 0);
#endif

              int ddh=0,ddv=0;
              for (int a=0;a<3;a++)
                for (int b=0;b<4;b++)
                  ddh += abs(subactarr.Ask(y*4+b,x*4+a  ) -
                             subactarr.Ask(y*4+b,x*4+a+1));
              for (int a=0;a<4;a++)
                for (int b=0;b<3;b++)
                  ddv += abs(subactarr.Ask(y*4+b  ,x*4+a) -
                             subactarr.Ask(y*4+b+1,x*4+a));
              double mbmean=0.0;
              for (int a=0;a<4;a++)
                for (int b=0;b<4;b++)
                  mbmean += subactarr.Ask(y*4+b,x*4+a);
              mbmean/=16;


#define BUSY_AMPL  15
#define RING_AMPL -20
              double ringing  = ((double)(ddh+ddv))/(2*12);
              double busyness = mbmean;

              busyness /= 250.0;
              ringing  /= 175.0;

              busyness = 0.2*pow(busyness,1.6);
              ringing  = pow(ringing,0.9);

#     if 0
              cout << "Busyness: " << busyness << endl;
              if (busyness>1.0) cout << "----------\n";
              cout << "Ringing:  " << ringing  << endl;
              if (ringing >1.0) cout << "----------\n";
#     endif
              busyness *= BUSY_AMPL;
              ringing  *= RING_AMPL;
              double masking = busyness + ringing;

              if (masking>0) masking=busyness  +ringing/2;
              else           masking=busyness/2+ringing;

              //              cout << "Masking: " << masking << endl;

              img->mblks.Ask(y,x).m_quantiser_scale += masking;

#define SHIFT 0
              mbbusy.Ask(y,x) = (int)busyness;
              mbring.Ask(y,x) = (int)-ringing;
              mbval   .Ask(y,x) = (int) (masking+SHIFT);
              mbvalneg.Ask(y,x) = (int)-(masking+SHIFT);
            }}

#if 0
#if 1
        {
          Array2<int32> lum; lum.Create(MB_Width*16,MB_Height*16);

          {for (int y=0;y<MB_Height*16;y++)
            for (int x=0;x<MB_Width*16;x++)
              {
                lum.Ask(y,x) = pix[y][x];
              }}
          
          ofstream vis;
          OpenStream(vis,"out/orig%03d.ppm",img->nr);
          WriteArrayAsPPM(lum,0,255,vis,1);
        }
#endif
#if 1
        {
          ofstream vis;
          OpenStream(vis,"out/mblp%03d.ppm",img->nr);
          WriteArrayAsPPM(mbval,0,BUSY_AMPL,vis,16);
        }
        {
          ofstream vis;
          OpenStream(vis,"out/mbln%03d.ppm",img->nr);
          WriteArrayAsPPM(mbvalneg,0,-RING_AMPL,vis,16);
        }
#endif
#if 1
        {
          ofstream vis;
          OpenStream(vis,"out/busy%03d.ppm",img->nr);
          WriteArrayAsPPM(mbbusy,0,BUSY_AMPL,vis,16);
        }
#endif
#if 1
        {
          ofstream vis;
          OpenStream(vis,"out/ring%03d.ppm",img->nr);
          WriteArrayAsPPM(mbring,0,-RING_AMPL,vis,16);
        }
#endif
#if 1
        {
          ofstream vis;
          OpenStream(vis,"out/act%03d.ppm",img->nr);
          WriteArrayAsPPM(actarr,-300,1000,vis,8);
        }
#endif
#if 1
        {
          ofstream vis;
          OpenStream(vis,"out/subact%03d.ppm",img->nr);
          WriteArrayAsPPM(subactarr,0,1000,vis,4);
        }
#endif
#endif
      }

    if (options.Enable_LuminanceMasking)
      {
        const double DarkAmpl   =14;
        const double BrightAmpl =10;
        const double DarkThres  =70;
        const double BrightThres=210;
        
        {for (int y=0;y<MB_Height;y++)
          for (int x=0;x<MB_Width;x++)
            {
              double mean = img->mean.Ask(y,x)/(16*16);
              
              if (mean<DarkThres)
                {
                  double qoffset = DarkAmpl*(DarkThres-mean)/DarkThres;
                  assert(qoffset>=0);
                  img->mblks.Ask(y,x).m_quantiser_scale += qoffset;
                }
              else if (mean>BrightThres)
                {
                  double qoffset = BrightAmpl*(mean-BrightThres)/(255-BrightThres);
                  assert(qoffset>=0);
                  img->mblks.Ask(y,x).m_quantiser_scale += qoffset;
                }
            }}
      }


  if (options.Enable_CBR_RateControl)
    {
      CalcSpatialDerivative(img,img->derivatives);
      SetQOffsetToMeetFrameSize(img);

#if 0
      {
        ofstream vis;
        OpenStream(vis,"out/vis%03d.ppm",img->nr);
        WriteArrayAsPPM(img->derivatives,0,25 *15*16*2,vis);
      }
#endif

      if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
        {
          cerr << ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> QSCALE (";
          switch (img->pichdr.m_picture_coding_type)
            {
            case PICTYPE_I: cerr << "I"; break;
            case PICTYPE_P: cerr << "P"; break;
            case PICTYPE_B: cerr << "B"; break;
            }
          cerr << ") = " << img->mblks.Ask(0,0).m_quantiser_scale << endl;
        }
    }


  // QScale veraendern, da Scenechange erkannt wurde

  if (img->scenechange_qscale_factor != 1.0)
    {
      {for (int y=0;y<MB_Height;y++)
        for (int x=0;x<MB_Width;x++)
          {
            img->mblks.Ask(y,x).m_quantiser_scale *= img->scenechange_qscale_factor;
          }}
    }


  // DCT / Quantisieren

  Task_DCTQuantize task_dct(img->nr);
  task_dct.Execute(procdata);



#if 0
      uint32 correct=0;
      {for (int y=0;y<MB_Height;y++)
        for (int x=0;x<MB_Width;x++)
          {
            Macroblock& mb = img->mblks.Ask(y,x);
            
            correct += BitsToCodeCoeff(mb,img->pichdr);
          }}
      correct /= 8;

      uint32 bytes = EstimateBitsToCode(img) / 8;

      cout << "GESCHAETZTE GROESSE in Bytes:  " << bytes << endl;
      cout << "MB-GEMESSENE GROESSE in Bytes: " << correct << endl;
#endif


#if 0
      if (1) //(options.Show_DerivVsBits)
    {
      CalcSpatialDerivative(img,img->derivatives);

#if 1
      {for (int y=0;y<MB_Height;y++)
        for (int x=0;x<MB_Width;x++)
          {
            Macroblock& mb = img->mblks.Ask(y,x);
            
            uint32 bits = BitsToCodeCoeff(mb,img->pichdr);
        
            cout << img->derivatives.Ask(y,x) << '\t' << img->variance.Ask(y,x) << '\t' << bits << '\t';
            
            if (mb.m_HasMotionForward && mb.m_HasMotionBackward)
              {
                cout << "D\n";
              }
            else if (mb.m_HasMotionForward)
              {
                cout << "F\n";
              }
            else if (mb.m_HasMotionBackward)
              {
                cout << "B\n";
              }
            else
              {
                cout << "I\n";
              }
          }}
#else
      uint32 correct=0;
      {for (int y=0;y<MB_Height;y++)
        for (int x=0;x<MB_Width;x++)
          {
            Macroblock& mb = img->mblks.Ask(y,x);
            
            correct += BitsToCodeCoeff(mb,img->pichdr);
          }}
      correct /= 8;

      uint32 bytes = EstimateBitsToCode(img) / 8;

      static double lastfact_I = 1.0;
      static double lastfact_P = 1.0;
      static double lastfact_B = 1.0;

      double fact = ((double)correct)/bytes;

      double lastfact = 0; /* Compiler beruhigen */
      switch (img->pichdr.m_picture_coding_type)
        {
        case PICTYPE_I: lastfact=lastfact_I; cout << "I "; break;
        case PICTYPE_P: lastfact=lastfact_P; cout << "P "; break;
        case PICTYPE_B: lastfact=lastfact_B; cout << "B "; break;
        }

      cout.precision(6);
      cout.setf(ios::showpoint);
      cout.setf(ios::fixed);

      cout << " Est: " << setw(6) << bytes;
      cout << " Cor: " << setw(6) << correct;
      cout << " Fact: " << setw(8) << fact;
      cout << "   (Est: " << setw(6) << ((int)(bytes*lastfact)) << " "
           << setw(8) << (correct/(bytes*lastfact)) << ")\n";
      cout.flush();

      fact = (lastfact+fact)/2.0;
      if (fact>2.0) fact=2.0;
      if (fact<0.5) fact=0.5;

      switch (img->pichdr.m_picture_coding_type)
        {
        case PICTYPE_I: lastfact_I=fact; break;
        case PICTYPE_P: lastfact_P=fact; break;
        case PICTYPE_B: lastfact_B=fact; break;
        }
#endif
    }
#endif

}

template class Array2<int32>;
#include "image/array2.cc"
