/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/taskdiffpics.hh"
#include "coding/decoder.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"

#if USE_MMX
#include "arch/MMX/mmx.hh"
#include "arch/MMX/mmxlib.icc"
#endif

extern bool DisableSomething;


void CreateDifferencePictures(ImageData* img)
{
  // Referenzbild(er) besorgen

  ImageData* lastIP = img->prev_ip;
  ImageData* nextIP = img->next_ip;


  // Bildgroesse zwischenspeichern

  assert(img->source);

  uint32 width  = img->source->AskWidth();
  uint32 height = img->source->AskHeight();
  uint32 mbWidth  = img->mblks.AskWidth();
  uint32 mbHeight = img->mblks.AskHeight();


  // Forward- und Backward-Praediktion erstellen

  switch (img->pichdr.m_picture_coding_type)
    {
    case PICTYPE_B:

      // ------ Backward Praediktion --------

      img->pred_backw.CreateImage(YCrCb,width,height,   true,true /*TODO: anderes Chromaformat*/);

      {
        Pixel*const* this_yy = img->source->AskBitmap(Bitmap_Y ).AskFrame();
        Pixel*const* this_cb = img->source->AskBitmap(Bitmap_Cb).AskFrame();
        Pixel*const* this_cr = img->source->AskBitmap(Bitmap_Cr).AskFrame();

        DiffPixel*const* yy = img->pred_backw.AskBitmap(Bitmap_Y ).AskFrame();
        DiffPixel*const* cb = img->pred_backw.AskBitmap(Bitmap_Cb).AskFrame();
        DiffPixel*const* cr = img->pred_backw.AskBitmap(Bitmap_Cr).AskFrame();

        assert(nextIP);

        Pixel*const* future_yy = nextIP->decoded.AskBitmap(Bitmap_Y ).AskFrame();
        Pixel*const* future_cb = nextIP->decoded.AskBitmap(Bitmap_Cb).AskFrame();
        Pixel*const* future_cr = nextIP->decoded.AskBitmap(Bitmap_Cr).AskFrame();

        {for (uint32 y=0;y<mbHeight;y++)
          for (uint32 x=0;x<mbWidth;x++)
            {
              Macroblock& mb = img->mblks.Ask(y,x);

              MotionCompensateSingleSet(yy,x*16,y*16,
                                        x*16,y*16,16,16,future_yy,
                                        mb.m_backward1.m_habs,
                                        mb.m_backward1.m_vabs);
              MotionCompensateSingleSet(cb,x*8,y*8,
                                        x*8,y*8,8,8,future_cb,
                                        mb.m_backward1.m_habs/2,
                                        mb.m_backward1.m_vabs/2);
              MotionCompensateSingleSet(cr,x*8,y*8,
                                        x*8,y*8,8,8,future_cr,
                                        mb.m_backward1.m_habs/2,
                                        mb.m_backward1.m_vabs/2);
            }}


#if USE_MMX
        if (MMX_Supported && !DisableSomething)
          {
            SetMM0_0();

            {for (uint32 y=0;y<height;y++)
              for (uint32 x=0;x<width;x+=16)
                {
                  Differences16_8_16(&this_yy[y][x],&yy[y][x]);
                }}
            {for (uint32 y=0;y<height/2;y++)
              for (uint32 x=0;x<width/2;x+=8)
                {
                  Differences8_8_16(&this_cb[y][x],&cb[y][x]);
                  Differences8_8_16(&this_cr[y][x],&cr[y][x]);
                }}
          }
        else
#endif
          {
            {for (uint32 y=0;y<height;y++)
              for (uint32 x=0;x<width;x++)
                {
                  yy[y][x] = ((DiffPixel)this_yy[y][x] - (DiffPixel)yy[y][x]);
                }}
            {for (uint32 y=0;y<height/2;y++)
              for (uint32 x=0;x<width/2;x++)
                {
                  cb[y][x] = ((DiffPixel)this_cb[y][x] - (DiffPixel)cb[y][x]);
                  cr[y][x] = ((DiffPixel)this_cr[y][x] - (DiffPixel)cr[y][x]);
                }}
          }
      }

      // fall-through

    case PICTYPE_P:

      // ------ Forward Praediktion --------

      img->pred_forw.CreateImage(YCrCb,width,height,   true,true /*TODO: anderes Chromaformat*/);

      {
        DiffPixel*const* yy = img->pred_forw.AskBitmap(Bitmap_Y ).AskFrame();
        DiffPixel*const* cb = img->pred_forw.AskBitmap(Bitmap_Cb).AskFrame();
        DiffPixel*const* cr = img->pred_forw.AskBitmap(Bitmap_Cr).AskFrame();

        assert(lastIP);

        Pixel*const* past_yy = lastIP->decoded.AskBitmap(Bitmap_Y ).AskFrame();
        Pixel*const* past_cb = lastIP->decoded.AskBitmap(Bitmap_Cb).AskFrame();
        Pixel*const* past_cr = lastIP->decoded.AskBitmap(Bitmap_Cr).AskFrame();

        Pixel*const* this_yy = img->source->AskBitmap(Bitmap_Y ).AskFrame();
        Pixel*const* this_cb = img->source->AskBitmap(Bitmap_Cb).AskFrame();
        Pixel*const* this_cr = img->source->AskBitmap(Bitmap_Cr).AskFrame();

        {for (uint32 y=0;y<mbHeight;y++)
          for (uint32 x=0;x<mbWidth;x++)
            {
              Macroblock& mb = img->mblks.Ask(y,x);

              MotionCompensateSingleSet(yy,x*16,y*16,
                                        x*16,y*16,16,16,past_yy,
                                        mb.m_forward1.m_habs,
                                        mb.m_forward1.m_vabs);
              MotionCompensateSingleSet(cb,x*8,y*8,
                                        x*8,y*8,8,8,past_cb,
                                        mb.m_forward1.m_habs/2,
                                        mb.m_forward1.m_vabs/2);
              MotionCompensateSingleSet(cr,x*8,y*8,
                                        x*8,y*8,8,8,past_cr,
                                        mb.m_forward1.m_habs/2,
                                        mb.m_forward1.m_vabs/2);
            }}

#if USE_MMX
        if (MMX_Supported && !DisableSomething)
          {
            SetMM0_0();

            {for (uint32 y=0;y<height;y++)
              for (uint32 x=0;x<width;x+=16)
                {
                  Differences16_8_16(&this_yy[y][x],&yy[y][x]);
                }}
            {for (uint32 y=0;y<height/2;y++)
              for (uint32 x=0;x<width/2;x+=8)
                {
                  Differences8_8_16(&this_cb[y][x],&cb[y][x]);
                  Differences8_8_16(&this_cr[y][x],&cr[y][x]);
                }}
          }
        else
#endif
          {
            {for (uint32 y=0;y<height;y++)
              for (uint32 x=0;x<width;x++)
                {
                  yy[y][x] = ((DiffPixel)this_yy[y][x] - (DiffPixel)yy[y][x]);
                }}
            {for (uint32 y=0;y<height/2;y++)
              for (uint32 x=0;x<width/2;x++)
                {
                  cb[y][x] = ((DiffPixel)this_cb[y][x] - (DiffPixel)cb[y][x]);
                  cr[y][x] = ((DiffPixel)this_cr[y][x] - (DiffPixel)cr[y][x]);
                }}
          }
      }
    }


 
  // ------ Bidirektionale Praediktion erstellen ----------
 
  if (img->pichdr.m_picture_coding_type==PICTYPE_B)
    {
      img->pred_bidir.CreateImage(YCrCb,width,height,   true,true /*TODO: anderes Chromaformat*/);

      DiffPixel*const* fyy = img->pred_forw.AskBitmap(Bitmap_Y ).AskFrame();
      DiffPixel*const* fcb = img->pred_forw.AskBitmap(Bitmap_Cb).AskFrame();
      DiffPixel*const* fcr = img->pred_forw.AskBitmap(Bitmap_Cr).AskFrame();
      
      DiffPixel*const* byy = img->pred_backw.AskBitmap(Bitmap_Y ).AskFrame();
      DiffPixel*const* bcb = img->pred_backw.AskBitmap(Bitmap_Cb).AskFrame();
      DiffPixel*const* bcr = img->pred_backw.AskBitmap(Bitmap_Cr).AskFrame();

      DiffPixel*const* dyy = img->pred_bidir.AskBitmap(Bitmap_Y ).AskFrame();
      DiffPixel*const* dcb = img->pred_bidir.AskBitmap(Bitmap_Cb).AskFrame();
      DiffPixel*const* dcr = img->pred_bidir.AskBitmap(Bitmap_Cr).AskFrame();

#if USE_MMX
      if (MMX_Supported && !DisableSomething)
        {
          Prepare_Mean8_S16();

          {for (uint32 y=0;y<height;y++)
            for (uint32 x=0;x<width;x+=8)
              {
                Mean8_S16(&fyy[y][x],&byy[y][x],&dyy[y][x]);
              }}

          {for (uint32 y=0;y<height/2;y++)
            for (uint32 x=0;x<width/2;x+=8)
              {
                Mean8_S16(&fcb[y][x],&bcb[y][x],&dcb[y][x]);
                Mean8_S16(&fcr[y][x],&bcr[y][x],&dcr[y][x]);
              }}
        }
      else
#endif
        {
          {for (uint32 y=0;y<height;y++)
            for (uint32 x=0;x<width;x+=2)
              {
                dyy[y][x  ] = (fyy[y][x  ] + byy[y][x  ] +1)/2;
                dyy[y][x+1] = (fyy[y][x+1] + byy[y][x+1] +1)/2;
              }}

          {for (uint32 y=0;y<height/2;y++)
            for (uint32 x=0;x<width/2;x++)
              {
                dcb[y][x] = (fcb[y][x] + bcb[y][x] +1)/2;
                dcr[y][x] = (fcr[y][x] + bcr[y][x] +1)/2;
              }}
        }
    }

#if USE_MMX
  if (MMX_Supported && !DisableSomething)
    EMMS();
#endif
}







void CreateZeroDifferencePictures(ImageData* img)
{
  // Bildgroesse zwischenspeichern

  assert(img->source);

  uint32 width  = img->source->AskWidth();
  uint32 height = img->source->AskHeight();


  // Forward- und Backward-Praediktion erstellen

  switch (img->pichdr.m_picture_coding_type)
    {
    case PICTYPE_B:

      // ------ Backward Praediktion --------

      img->pred_backw.CreateImage(YCrCb,width,height,   true,true /*TODO: anderes Chromaformat*/);

      {
        DiffPixel*const* yy = img->pred_backw.AskBitmap(Bitmap_Y ).AskFrame();
        DiffPixel*const* cb = img->pred_backw.AskBitmap(Bitmap_Cb).AskFrame();
        DiffPixel*const* cr = img->pred_backw.AskBitmap(Bitmap_Cr).AskFrame();

        {for (uint32 y=0;y<height;y++)
          for (uint32 x=0;x<width;x+=2)
            {
              yy[y][x  ] = 0;
              yy[y][x+1] = 0;
            }}
        {for (uint32 y=0;y<height/2;y++)
          for (uint32 x=0;x<width/2;x+=2)
            {
              cb[y][x]   = 0;
              cb[y][x+1] = 0;
              cr[y][x]   = 0;
              cr[y][x+1] = 0;
            }}
      }

      // fall-through

    case PICTYPE_P:

      // ------ Forward Praediktion --------

      img->pred_forw.CreateImage(YCrCb,width,height,   true,true /*TODO: anderes Chromaformat*/);

      {
        DiffPixel*const* yy = img->pred_forw.AskBitmap(Bitmap_Y ).AskFrame();
        DiffPixel*const* cb = img->pred_forw.AskBitmap(Bitmap_Cb).AskFrame();
        DiffPixel*const* cr = img->pred_forw.AskBitmap(Bitmap_Cr).AskFrame();

        {for (uint32 y=0;y<height;y++)
          for (uint32 x=0;x<width;x++)
            {
              yy[y][x] = 0;
            }}
        {for (uint32 y=0;y<height/2;y++)
          for (uint32 x=0;x<width/2;x++)
            {
              cb[y][x] = 0;
              cr[y][x] = 0;
            }}
      }
    }


 
  // ------ Bidirektionale Praediktion erstellen ----------
 
  if (img->pichdr.m_picture_coding_type==PICTYPE_B)
    {
      img->pred_bidir.CreateImage(YCrCb,width,height,   true,true /*TODO: anderes Chromaformat*/);

      DiffPixel*const* dyy = img->pred_bidir.AskBitmap(Bitmap_Y ).AskFrame();
      DiffPixel*const* dcb = img->pred_bidir.AskBitmap(Bitmap_Cb).AskFrame();
      DiffPixel*const* dcr = img->pred_bidir.AskBitmap(Bitmap_Cr).AskFrame();
        
      {for (uint32 y=0;y<height;y++)
        for (uint32 x=0;x<width;x+=2)
          {
            dyy[y][x]   = 0;
            dyy[y][x+1] = 0;
          }}
      {for (uint32 y=0;y<height/2;y++)
        for (uint32 x=0;x<width/2;x+=2)
          {
            dcb[y][x]   = 0;
            dcb[y][x+1] = 0;
            dcr[y][x]   = 0;
            dcr[y][x+1] = 0;
          }}
    }
}







void Task_CreateDifferencePictures::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("create diff-pictures",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  CreateDifferencePictures(img);
}


void Task_CreateDifferencePictures_0::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("create diff-pics 0",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  CreateZeroDifferencePictures(img);
}

