/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "control/codingloop.hh"
#include "control/process.hh"
#include "utility/thread.hh"
#include <unistd.h>
#include <time.h>
#include <errno.h>

#include <sys/time.h>
#include <unistd.h>


static long long msec_work = 0;
static long long msec_wait = 0;

static int TimeDifference(struct timeval& t1,struct timeval& t2)
{
  int diff = (t2.tv_sec - t1.tv_sec)*1000;
  if (t2.tv_usec >= t1.tv_usec)
    diff += (t2.tv_usec-t1.tv_usec)/1000;
  else
    diff += 1000-(t1.tv_usec-t2.tv_usec)/1000;

  return diff;
}

static void MainCodeLoop(TaskPool& tpool,int threadnr)
{
  struct timeval t1,t2;

  ProcessData procdata(threadnr);

  while (!tpool.Finished())
    {
      Task* t = tpool.GetTask();

      // cout << "Execute ------------ " << t << endl;

      if (t!=NULL)
        {
	  gettimeofday(&t1,NULL);

          t->Execute(procdata);
          tpool.TaskCompleted(t);
          delete t;

	  gettimeofday(&t2,NULL);

	  msec_work += TimeDifference(t1,t2);
        }
      else
        {
	  // cout << "BLOCKED\n";

	  gettimeofday(&t1,NULL);

//          tpool.ShowBlocked();
          // Blockiert, da benoetigte Ergebnisse noch nicht vorliegen.   ---> etwas warten

          timespec ts;
          ts.tv_sec  = 0;
          ts.tv_nsec = 100000000;  // 0.1 s

          for (;;)
            {
              timespec remaining;
              int result = nanosleep(&ts,&remaining);
              if (result != -1) break;
              if (errno != EINTR) return;
	      
              ts=remaining;
	      cout << ".\n";
            }

	  gettimeofday(&t2,NULL);

	  msec_wait += TimeDifference(t1,t2);
        }
    }
}

struct ThreadStartData
{
  TaskPool* taskpool;
  int       threadnr;
};

static int StartLoop(struct ThreadStartData* data)
{
  MainCodeLoop(*data->taskpool, data->threadnr);
  return 0;
}


void DoCode(TaskPool& tpool,int nThreads)
{
#define MAXTHREADS 50
  Thread threads[MAXTHREADS];

  for (int i=0;i<nThreads-1;i++)
    {
      ThreadStartData data;
      data.taskpool = &tpool;
      data.threadnr = i+1;
      threads[i].Start((void* (*)(void*))StartLoop,(void*)&data);
    }

  if (nThreads>1)
    MainCodeLoop(tpool,nThreads);
  else
    MainCodeLoop(tpool,0);

  cout << "Time for coding (ms): " << msec_work << "  Time for waiting (ms): " << msec_wait << endl;
  cout << "Total time: " << msec_work+msec_wait << endl;
}
