/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CONTROL_DATAPOOL_HH
#define CONTROL_DATAPOOL_HH

#include <types.hh>
#include <utility/queue.hh>
#include <image/image.hh>
#include <image/array2.hh>
#include <image/codedimg.hh>
#include <mpeg/syntaxelems.hh>
#include <output/bitbuffer.hh>
#include <utility/thread.hh>

class ImageData
{
public:
  ImageData();
  ~ImageData();

  const uint32 nr;

  static ImageData* GetNr(uint32 nr);
  ImageData* GetNext() const { return GetNr(nr+1); }
  ImageData* GetPrev() const { return GetNr(nr-1); }

  static void   DeleteFrame(uint32 nr);

  static uint32 AskFirstFrame() { if (s_imgdata.IsEmpty()) return 0; else return s_imgdata.AskHead()->nr; }

  // === GOP-Data ===

  void SetGOPBeginning() { d_start_gop=true; }
  bool IsGOPBeginning() const { return d_start_gop; }
  void SetClosedGOP() { d_is_closed_gop=true; }         // Nur bei GOP-Beginning
  bool IsClosedGOP() const { return d_is_closed_gop; }  // Nur gueltig bei GOP-Beginning
  void GetGOPHdr(GOPHeaderData&) const;

  // ======================== Hier Beginnen die Daten zur Codierung ========================

  class MotionEstimator* GetMotionEstimator();
  void GiveBackMotionEstimator(class MotionEstimator*);

  // bool finished;

  Image<Pixel>* source;

  PictureHeaderData  pichdr;
  Array2<Macroblock> mblks;
  Array2<uint16>     mean;
  Array2<uint32>     variance;
  Array2<uint32>     derivatives;
  long long total_mean, total_variance;

  uint32 estimated_bits;

  bool   no_forward_prediction;     // Im B-Frame keine forward-Praediktion benutzen.
  double scenechange_qscale_factor;
  const class MEData* medata;

  ImageData* prev_ip; // gesetzt vom TaskPool beim Erzeugen der Post-Classify-Tasks
  ImageData* next_ip; //  " " "

  Image<DiffPixel>   pred_forw;  // Differenzbild fuer Forward-Prediction (P-Frames)
  Image<DiffPixel>   pred_backw;
  Image<DiffPixel>   pred_bidir;

  uint32             desired_size;  // Gewuenschte Bildgroesse in Bytes.
  uint32             expected_size; // Vorhergesagte Bildgroesse in Bytes.
  BitBuffer output;

  Image<Pixel> decoded;

private:
  bool  d_start_seqhdr;

  bool  d_start_gop;
  bool  d_is_closed_gop;

  class MotionEstimator* d_motionestim;
  Mutex d_motionest_access_mutex;

  static uint32 s_nextnr;
  static Queue<ImageData*> s_imgdata;
  static Mutex s_poolaccess_mutex;

  static int GetIndexOfFrame(uint32 framenr);
};

#if 0
struct PictureHeaderData
{
  /*       Variable                         setzender Task
         ===================================================  */
  uint10 m_temporal_reference;              // FrameType
  uint3  m_picture_coding_type;             // FrameType
  uint16 m_vbv_delay;                       //   ???
  bool   m_full_pel_forward_vector;         // MVSearch
#define  m_forward_f_code  m_f_code[0][0]   //   ""
  bool   m_full_pel_backward_vector;        //   ""
#define  m_backward_f_code m_f_code[1][0]   //   ""
  
  // MPEG2
  
  uint4  m_f_code[2][2];                    // MVSearch
  uint8  m_intra_dc_precision;  /* 8-11 */  // PicHdrInit
  uint2  m_picture_structure;               // PicHdrInit
  bool   m_top_field_first;                 // PicHdrInit
  bool   m_frame_pred_frame_dct;            // PicHdrInit
  bool   m_concealment_motion_vectors;      // PicHdrInit
  uint1  m_q_scale_type;                    // QScale
  uint1  m_intra_vlc_format;                //   ???
  bool   m_alternate_scan;                  // PicHdrInit
  bool   m_repeat_first_field;              // PicHdrInit
  bool   m_progressive_frame;               // PicHdrInit
  bool   m_composite_display_flag;          // PicHdrInit
  uint1  m_v_axis;                          //   ...
  uint3  m_field_sequence;
  bool   m_sub_carrier;
  uint7  m_burst_amplitude;
  uint8  m_sub_carrier_phase;
};
#endif



class SeqHdrPool
{
public:
  static void SetSeqHdr(int framenr,const SequenceHeaderData&);
  static bool IsSeqHdrBeforeFrame(int framenr);
  static void ForgetSeqHdrBefore(int framenr);
  static void AskSeqHdr(int framenr,SequenceHeaderData&);

private:
  static Queue<SequenceHeaderData*> seqhdr_queue;
  static Queue<int> seqhdrpos_queue;
};

#endif
