/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CONTROL_ENCODER_SIMPLE_HH
#define CONTROL_ENCODER_SIMPLE_HH

#include "control/encoder.hh"
#include "utility/queue.hh"
//#include "coding/dct_separated.hh"
#include "coding/dct_coupled.hh"
#include "coding/motionest.hh"
#include "coding/decoder.hh"

class EncoderControl_Simple : public EncoderControl
{
public:
  EncoderControl_Simple(InputPreprocessor& preproc,
                        StreamSink& sink,
                        const EncodingOptions& opts)
    : EncoderControl(preproc,sink,opts),
      d_motionalgo(LogSearch),
      d_gop_len(12), d_p_distance(1),
      d_intra_dc_precision(8),
      d_qscale(8), d_qscale_type(0),
      d_intra_vlc_format(0),
      d_alternate_scan(false),
      d_visualize(false),
      d_DisablePredictionCorrection(false),
      d_pic_nr(0),
      forward_motionest(NULL),
      backward_motionest(NULL)
    {
      SetFCodes(1,1,1,1);
    }

  ~EncoderControl_Simple();

  void EnableVisualization(const char* filenametemplate,bool flag=true)
  { d_visualizetemplate=filenametemplate; d_visualize=flag; }

  void SetGOPLengths(int goplength,int Pdistance);
  void SetQScale(int intra_dc_precision, int qscalecode,uint1 qscale_type);
  void SetIntraVLCFormat(uint1 f) { d_intra_vlc_format=f; }
  void SetAlternateScan(bool flag=true) { d_alternate_scan=flag; }
  void SetMotionEstAlgo(enum MotionEstimationAlgo algo) { d_motionalgo=algo; }
  void SetFCodes(int fwdh,int fwdv,int bkwh,int bkwv)
    {
      d_fcode[0][0] = fwdh;      d_fcode[0][1] = fwdv;
      d_fcode[1][0] = bkwh;      d_fcode[1][1] = bkwv;
    }
  
  void Encode();

private:
  //int d_framerate100;

  void FillSequenceHeader(SequenceHeaderData&);
  void FillPictureHeader (PictureHeaderData&);
  void FinishPicture(CodedImage&,int nr,bool decode);

  void FillQScale(const Image<Pixel>&,CodedImage&);
  void AdjustMotionRange(CodedImage& codedimg);

  //DCTCoder_Coupled<Pixel> d_qscale_dctcoder;
  DCTCoder_Separated<Pixel> d_qscale_dctcoder;
  uint8 d_qscale_field[100][200];  // TODO: dynamisch an Groesse anpassen

  // Optionen fuer die Codierung

  enum MotionEstimationAlgo d_motionalgo;

  int d_gop_len;
  int d_p_distance;

  int   d_fcode[2][2];
  int   d_intra_dc_precision;
  int   d_qscale;
  uint1 d_qscale_type;
  uint1 d_intra_vlc_format;
  bool  d_alternate_scan;


  // Debugging-Optionen

  bool d_visualize;
  const char* d_visualizetemplate;

public:
  bool d_DisablePredictionCorrection;
private:


  // Variablen fuer die Codierung.

  int d_pic_nr;
  int d_width,d_height; // Bildgroessen

  Queue<Image<Pixel>*> b_frame_queue;     // Die B-Frames, deren Codierung aufgeschoben wurde.

  CodedImage codedimg_I;     // Picture-Header fuer die I-Frames
  CodedImage codedimg_PB;    // Picture-Header fuer die P- und B-Frames

  SequenceHeaderData sh;     // Sequence Header
  GOPHeaderData d_gop_data;

  //DCTCoder_Coupled<Pixel>     dctcoder_pix;
  DCTCoder_Separated<Pixel>     dctcoder_pix;
  //DCTCoder_Coupled<DiffPixel> dctcoder_dpix;
  DCTCoder_Separated<DiffPixel> dctcoder_dpix;
  Decoder decoder;
  MotionEstimator* forward_motionest;   // Fuer die P-Frames und B-Frames.
  MotionEstimator* backward_motionest;  // Fuer B-Frames.
};

#endif
