/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CONTROL_MEHELPER_HH
#define CONTROL_MEHELPER_HH

#include "coding/dct_coupled.hh"
#include "coding/motionest.hh"

class MotionEstimationHelper
{
public:
  MotionEstimationHelper();
  ~MotionEstimationHelper();

  void SetMotionEstAlgo(enum MotionEstimationAlgo algo);

  /* Bitmap muss in SearchParam nicht gesetzt werden. */
  void SetSearchParam(const SearchParam* fwd,const SearchParam* bkw);
  void SetReferenceBitmaps(const Bitmap<Pixel>* fwd,
                           const Bitmap<Pixel>* bkw,
                           const Bitmap<Pixel>* fwd2 = NULL,
                           const Bitmap<Pixel>* bkw2 = NULL);
  void SetImageBitmap(const Bitmap<Pixel>* img);

  void MVSearch(int mb_x,int mb_y,Macroblock& mb);
  void DoMVSearch(CodedImage&);
  void AdjustMotionRange(CodedImage& codedimg);

  void SetSearchRange(int fwdh,int fwdv,int bkwh,int bkwv);

private:
  // Optionen fuer die Codierung

  enum MotionEstimationAlgo d_motionalgo;
  SearchParam d_sparam_fwd,d_sparam_bkw;

  int d_fcode[2][2];
  int d_width,d_height; // Bildgroessen

  MotionEstimator* forward_motionest;   // Fuer die P-Frames und B-Frames.
  MotionEstimator* backward_motionest;  // Fuer B-Frames.

  Image<DiffPixel> mb_pred;
};

#endif
