/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "control/taskpichdr.hh"
#include "control/datapool.hh"
#include "coding/mvutils.hh"
#include "mpeg/constants.hh"
#include "input/preproc.hh"
#include "options.hh"

void Task_FillPicHdr::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("fill PicHdr",data.threadnr); cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  img->pichdr.m_vbv_delay = 0xFFFF;
  
  img->pichdr.m_full_pel_forward_vector  = false;
  img->pichdr.m_full_pel_backward_vector = false;

  img->pichdr.m_f_code[0][0]=1; //MVOffsetToFCode(-options.HSearchRange *2); // *2 wegen Half-Pel
  img->pichdr.m_f_code[0][1]=1; //MVOffsetToFCode(-options.VSearchRange *2);
  img->pichdr.m_f_code[1][0]=1; //MVOffsetToFCode(-options.HSearchRange *2);
  img->pichdr.m_f_code[1][1]=1; //MVOffsetToFCode(-options.VSearchRange *2);

  img->pichdr.m_intra_dc_precision= options.mpeg2.DC_Precision;

  img->pichdr.m_picture_structure = PICSTRUCT_FramePicture;

  img->pichdr.m_top_field_first = true;
  img->pichdr.m_frame_pred_frame_dct = true;
  img->pichdr.m_concealment_motion_vectors = false;
  img->pichdr.m_intra_vlc_format = options.mpeg2.IntraVLCFormat;
  img->pichdr.m_alternate_scan   = options.mpeg2.AlternateScan ? 1 : 0;
  img->pichdr.m_repeat_first_field = false;
  img->pichdr.m_progressive_frame = true;
  
  img->pichdr.m_composite_display_flag = false;
}


void Task_InitMBs::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("init MBs",data.threadnr); cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  int w=(img->source->AskWidth() +15)/16;
  int h=(img->source->AskHeight()+15)/16;

  if (!img->mblks.IsInitialized())
    {
      img->mblks.Create(w,h);
    }

  for (int y=0;y<h;y++)
    for (int x=0;x<w;x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);

        /* Makroblock muss hier passend fuer I-Frames ausgefuellt werden, da fuer I-Frames
           kein MBType-Task erzeugt wird. */

        mb.m_IsIntra=true;
        mb.m_HasMotionForward=false;
        mb.m_HasMotionBackward=false;
        mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);

        if (options.CreateMPEG2)
          {
            if (x==0)   mb.m_BeginSlice=true;
            if (x==w-1) mb.m_EndSlice  =true;
          }
        else
          {
            if (x==0   && y==0  ) mb.m_BeginSlice=true;
            if (x==w-1 && y==h-1) mb.m_EndSlice  =true;
          }
      }
}
