/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CONTROL_TASKPOOL_HH
#define CONTROL_TASKPOOL_HH

#include <types.hh>
#include <utility/queue.hh>
#include <control/process.hh>

enum FrameTypeProc { IOnly, Pattern, SceneChange };      // TODO: das muss hier weg


#define MAXDEPS 200

struct Task
{
  Task() : nDeps(0) { }
  virtual ~Task() { }

  //  virtual void Execute() { assert(0); }
  virtual void Execute(ProcessData& data) = 0;
  virtual bool ReadyToExecute() const { return true; }

  void AddDependency(Task* t)
    {
      // cout << "AddDependency  -------- " << t << " -> " << this << endl;

      if (t==NULL) return;
      assert(nDeps<MAXDEPS);
      finished[nDeps]=false;
      dependencies[nDeps++]=t;
    }
  bool DependenciesFulfilled() const
    {
      for (int i=0;i<nDeps;i++)
        if (!finished[i])
          return false;
      return true;
    }
  void TaskFinished(const Task* t)
    {
      for (int i=0;i<nDeps;i++)
        if (dependencies[i]==t)
          finished[i]=true;
    }

  void ShowBlocked() const
    {
      for (int i=0;i<nDeps;i++)
        if (!finished[i])
          cout << dependencies[i] << ' ';
    }

  Task* dependencies[MAXDEPS];
  bool  finished[MAXDEPS];
  int   nDeps;
};

class Task_AllInOne : public Task
{
public:
  // Task_AllInOne() { }
};


class TaskPool
{
public:
  virtual ~TaskPool() { }

  virtual bool  Finished() = 0;
  virtual Task* GetTask() = 0;
  virtual void  TaskCompleted(Task*) = 0;

  virtual void ShowBlocked() = 0;
};


class AutoLock_Output
{
public:
  AutoLock_Output();
  ~AutoLock_Output();

  void NOP() { }
};

#define OutputLock AutoLock_Output Dummy_AutoLock_Output_Variable; Dummy_AutoLock_Output_Variable.NOP()

/* threadnr = 0 --> SingleThread-Lauf
   sonst: threadnr>=1
*/
void ShowTaskInformation(const char* task,int threadnr=0);
void EndShowTaskInformation();

#endif
