/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <control/taskpool_allinone.hh>
#include <coding/loading.hh>
#include <coding/varianz.hh>
#include <coding/frameclass.hh>
#include <coding/frameclass_scenechange.hh>
#include <coding/taskcode.hh>
#include <coding/decoder.hh>
#include <control/taskdel.hh>
#include <control/taskseqhdr.hh>
#include <control/taskwrite.hh>
#include <input/preproc.hh>
#include <output/streamsink.hh>
#include <mpeg/constants.hh>
#include "coding/ratecontrol.hh"

#include <iostream.h>

#define enable 0    // Ausgabe aktivieren ?


#define MAX_P_DISTANCE 100
const uint32 nFrameDataStructs = 2*MAX_P_DISTANCE+1;

struct framedatastruct2
{
  framedatastruct2() { Clear(); }

  void Clear()
    {
      allinone =NULL;
      decode   =NULL;
      bitstream=NULL;
      write    =NULL;
      delframe =NULL;
    }
  
  void TaskFinished(Task* task)
    {
      if (allinone ) allinone ->TaskFinished(task);
      if (decode   ) decode   ->TaskFinished(task);
      if (bitstream) bitstream->TaskFinished(task);
      if (write    ) write    ->TaskFinished(task);
      if (delframe ) delframe ->TaskFinished(task);
    }

  void AppendToQueue(Queue<Task*>& queue)
    {
      queue.Append(allinone);
      if (decode) queue.Append(decode);

      queue.AppendAtHead(delframe);
      queue.AppendAtHead(write);
      queue.AppendAtHead(bitstream);
    }

  uint32     framenr;
  ImageData* imagedata;
  
  uint3      pictype;     // I,P,B

  Task* allinone;
  Task* decode;
  Task* bitstream;
  Task* write;
  Task* delframe;
};



TaskPool_AllInOne::TaskPool_AllInOne(InputPreprocessor& preproc,StreamSink& sink)
  : d_preproc(preproc),
    d_sink(sink),
    d_nextframenr(0),
    d_load_pending(false),
    d_classify_pending(false),
    d_startframe_of_next_classify(0),
    d_task_classify(NULL),
    d_last_write_task(NULL)
{
  SetFrameTypeProc(SceneChange,50);
  SetIPDistances(20,5);

  Task* t = new Task_InsertSeqHdr(0,d_preproc);
  d_set.Append(t);
  if (enable) cout << "NEW SeqHdr " << t << endl;

  d_framedata = new framedatastruct2[nFrameDataStructs];
  d_nFrameData = 0;
}


TaskPool_AllInOne::~TaskPool_AllInOne()
{
  if (d_framedata) delete[] d_framedata;
}


void  TaskPool_AllInOne::ShowBlocked()
{
  for (uint32 i=0;i<d_set.Length();i++)
    {
      if (enable) cout << "Entry " << d_set.AskFromHead(i) << " is blocked, because of: ";
      d_set.AskFromHead(i)->ShowBlocked();
      if (enable) cout << endl;
    }
}

bool  TaskPool_AllInOne::Finished()
{
  return d_set.IsEmpty() && !d_preproc.HasMoreInput() && d_startframe_of_next_classify == d_nextframenr;
}


Task* TaskPool_AllInOne::GetTask()
{
  d_mutex.Lock();

  // Wenn noch ein Frame im Stream ist und der davorgehende schon geladen ist, naechsten Frame laden.
  // Auch dann nicht laden, wenn schon so viele Frames im voraus geladen wurden, wie in einem
  // Classify-Task bearbeitet werden, da sonst bei Single-Thread Verarbeitung der ganze Stream auf
  // einmal geladen wuerde.
  // Auch dann nicht laden, wenn der Bereich [letztes noch im Speicher befindliche Bild ; neues Bild]
  // groesser als ein (noch genau festzulegender 'TODO') Wert ist. Dadurch soll verhindert werden,
  // dass einige unguenstig plazierte, extrem langsame Threads das Weitercodieren blockieren und
  // somit Bilder weit in die Zukunft gelesen werden, ohne codiert werden zu koennen.

#if 0
  cout << "g1\n";
  cout << ((!d_load_pending)?1:0) << endl;
  cout << d_nextframenr << " " << d_startframe_of_next_classify << " " << d_lookahead << endl;
  cout << ((d_nextframenr < d_startframe_of_next_classify + d_lookahead)?1:0) << endl;
  cout << ((d_nextframenr < ImageData::AskFirstFrame()+200)?1:0) << endl;
#endif

#if 0
  cout << (d_load_pending ? "LOAD PENDING" : "load pending")
       << " Next-Frame: " << d_nextframenr
       << " StartOfNextClassify: " << d_startframe_of_next_classify
       << " Lookahead: " << d_lookahead
       << " ImageData::AskFirstFrame(): " << ImageData::AskFirstFrame() << endl;
#endif

  if (!d_load_pending &&                                              // Kein Lese-Thread in Betrieb.
      d_nextframenr < d_startframe_of_next_classify + d_lookahead &&  // Max. 'd_lookahead' vorauslesen.
      d_nextframenr < ImageData::AskFirstFrame()+200) /* TODO: Threshold abhaengig von nThreads,Lookahead */
    {
      // Falls noch mehr Frames im Stream -> Lesen
      if (d_preproc.HasMoreInput())
        {
          // Naechsten Frame laden und Varianz/Mean bestimmen.

          Task* loading = new Task_Loading(d_preproc);
          if (enable) cout << "NEW loading " << loading << endl;
          Task* var     = new Task_ComputeVarianzMean(d_nextframenr);
          if (enable) cout << "NEW var " << var << endl;

          var->AddDependency(loading);
          d_set.Append(var);
          
          // Frame spaeter auch klassifizieren.
          
          if (d_task_classify==NULL)
            d_task_classify = CreateTaskClassifyFrames();
          
          d_task_classify->AddDependency(var);
          d_task_classify->m_nFrames++;
          
          d_nextframenr++;
          d_load_pending=true;
          
          d_mutex.Unlock();
          return loading;
        }
      else
        {
        }
    }
      
  
  /* Wenn genug Frames fuer einen Classify-Task gelesen wurden (d_lookahead) oder der Stream zu Ende ist,
     einen neuen Classify-Task erzeugen. */

  if (!d_classify_pending &&                                         // Letzter Classify-Task wurde beendet
      !d_load_pending &&           // (sonst wissen wir nicht, ob noch ein Frame folgt)
      (d_nextframenr >= d_startframe_of_next_classify+d_lookahead || // Genug Bilder geladen oder
       (!d_preproc.HasMoreInput() &&                         // [keine Bilder mehr im Stream und
        d_startframe_of_next_classify != d_nextframenr)))    //  noch nicht alle klassifiziert].
    {
      if (d_task_classify==NULL)
        d_task_classify = CreateTaskClassifyFrames();

      if (!d_preproc.HasMoreInput())
        d_task_classify->m_LastFrames=true;

      d_set.Append(d_task_classify);
      d_classify_pending=true;
      d_task_classify=NULL;
    }


  // Warteschlange abarbeiten, falls moeglich

  for (uint32 i=0;i<d_set.Length();i++)
    if (d_set.AskFromHead(i)->DependenciesFulfilled())
      {
        Task* task =d_set.AskFromHead(i);
        d_set.RemoveFromHead(i);
        d_mutex.Unlock();
        return task;
      }


#if 0
  {
    OutputLock;
    cerr << "No new task ready because loading or classification is too slow\n";
  }
#endif

  d_mutex.Unlock();
  return NULL;
}

void  TaskPool_AllInOne::TaskCompleted(Task* task)
{
  d_mutex.Lock();

  if (enable) cout << "Completed ------------------ " << task << endl;

  // Dependencies abarbeiten

  {for (uint32 i=0;i<d_set.Length();i++)
    {
      assert(d_set.AskFromHead(i));
      if (enable) cout << "AFHead " << d_set.AskFromHead(i) << endl;
      d_set.AskFromHead(i)->TaskFinished(task);
    }}

  if (enable) cout << "DTASKClass " << d_task_classify << endl;
  if (d_task_classify) d_task_classify->TaskFinished(task);
  if (enable) cout << "DLWrite " << d_last_write_task << endl;
  if (d_last_write_task == task) d_last_write_task=NULL;

 d_framedata[0].TaskFinished(task);


  // Task-Spezifische Routinen

  if (dynamic_cast<Task_Loading*>(task))
    {
      d_load_pending=false;
    }

  if (dynamic_cast<Task_ClassifyFrames*>(task))
    {
      Task_ClassifyFrames* task_class = dynamic_cast<Task_ClassifyFrames*>(task);
      d_startframe_of_next_classify = task_class->m_last_frame_type_assigned+1;
      d_classify_pending = false;

      uint32 FirstGOPFrame = task_class->m_firstframe;
      uint32 LastGOPFrame  = task_class->m_last_frame_type_assigned;
      uint32 nClassFrames  = LastGOPFrame - FirstGOPFrame + 1;
      bool   AreLastFrames = ((FirstGOPFrame + task_class->m_nFrames-1 == LastGOPFrame) &&
                              task_class->m_LastFrames);

      if (options.Enable_CBR_RateControl)
        AssignFrameSizes(FirstGOPFrame,LastGOPFrame);


      // --- Neue FrameData-Structs ausfuellen. ---

      assert(d_nFrameData + nClassFrames <= nFrameDataStructs);  // Es sollte genug Platz vorhanden sein.

      uint32 srcframeidx = FirstGOPFrame;
      uint32 destidx     = d_nFrameData;
      for ( ; srcframeidx <= LastGOPFrame ; srcframeidx++, destidx++ )
        {
          d_framedata[destidx].framenr   = srcframeidx;
          d_framedata[destidx].imagedata = ImageData::GetNr(srcframeidx);
          d_framedata[destidx].pictype   = d_framedata[destidx].imagedata->pichdr.m_picture_coding_type;

          assert(d_framedata[destidx].framenr == d_framedata[destidx].imagedata->nr);
        }
      d_nFrameData += nClassFrames;

      while (GeneratePartOfGraph(AreLastFrames))
        {
          /* NOP */
        }
    }

  d_mutex.Unlock();
}






bool TaskPool_AllInOne::GeneratePartOfGraph(bool AreLastFrames)
{
  // Ersten und zweiten I/P-Frame suchen

  const uint32 FirstIP = 0;  // Index in d_framedata.
  assert(d_framedata[FirstIP].pictype != PICTYPE_B); // Erster Frame in d_framedata[] ist immer I/P-Frame

  uint32 SecondIP=0;

  {for (uint32 i=1;i<d_nFrameData;i++)
    {
      if (d_framedata[i].pictype != PICTYPE_B)
        {
          SecondIP=i;
          break;
        }
    }}


  // Wenn noch keine zwei I/P-Frames im Buffer sind, dann kann nichts gemacht werden.
  if (SecondIP==0)
    return false;


  // Flag fuer die 'letzten Frames' wieder loeschen, falls das noch nicht die letzten
  // zu codierenden Bilder sind.

  if (AreLastFrames && SecondIP != d_nFrameData-1)
    AreLastFrames = false;


//  cout << FirstIP << " ---> " << SecondIP << endl;

  // ----- prev_ip / next_ip - Felder in ImageData (FirstIP;SecondIP] setzen -----

  {for (uint32 i=FirstIP+1 ; i<SecondIP ; i++)
    {
      d_framedata[i].imagedata->prev_ip = d_framedata[FirstIP ].imagedata;
      d_framedata[i].imagedata->next_ip = d_framedata[SecondIP].imagedata;
    }}

  if (d_framedata[SecondIP].imagedata->pichdr.m_picture_coding_type == PICTYPE_P)
    d_framedata[SecondIP].imagedata->prev_ip = d_framedata[FirstIP].imagedata;
    

  /* ------------------- Graph fuer den Bereich (FirstIP;SecondIP] aufbauen ----------------- */
  // Ausnahme: beim ersten Mal auch fuer FirstIP.

  {for (uint32 i=FirstIP ; i<=SecondIP ; i++)
    {
      // Ausnahmebehandlung fuer FirstIP bei ersten Mal.
      if (i!=FirstIP || d_framedata[FirstIP].framenr==0)
        {
          uint32 framenr = d_framedata[i].framenr;

          Task* allinone   = d_framedata[i].allinone  = CreateMainTask(framenr);
          Task* bitstream  = d_framedata[i].bitstream = new Task_WriteFrame(framenr);
          Task* write      = d_framedata[i].write     = new Task_WriteToStream(framenr,d_sink);
          Task* deletetask = d_framedata[i].delframe  = new Task_DeleteFrame(framenr);

          bitstream ->AddDependency(allinone);
          write     ->AddDependency(bitstream);
          deletetask->AddDependency(write);
        }
    }}


  /* Zeitliche Quer-Abhaengigkeiten der Write-Tasks setzen. */

  d_framedata[SecondIP].write->AddDependency(d_last_write_task);
  d_last_write_task = d_framedata[SecondIP].write;
  {for (uint32 i=FirstIP+1;i<SecondIP;i++)
    {
      d_framedata[i].write->AddDependency(d_last_write_task);
      d_last_write_task = d_framedata[i].write;
    }}


  /* Wenn noetig, das Frame decodieren. */

  if (1) // SecondIP>1)
    /* Dekodierung fuer alle I/P-Frames erzwingen, da es schwierig zu entscheiden ist,
       ob bei der Eingabe "IP" das P-Frame decodiert werden muss. Das ist naemlich eigentlich
       nur der Fall, wenn noch ein "P" folgt, das kann aber zu dem Zeitpunkt noch nicht
       festgestellt werden. Wir brauchen aber schon den Decode-Task, damit wir die Abhaengigkeit
       fuer den Delete-Task des I-Frames setzen koennen.
    */
    {
      if (d_framedata[FirstIP].decode == NULL)
        {
          d_framedata[FirstIP].decode = new Task_Decode(d_framedata[FirstIP].framenr);
          d_framedata[FirstIP].decode  ->AddDependency(d_framedata[FirstIP].allinone);
          d_framedata[FirstIP].delframe->AddDependency(d_framedata[FirstIP].decode);
        }

      d_framedata[SecondIP].decode = new Task_Decode(d_framedata[SecondIP].framenr);
      d_framedata[SecondIP].decode  ->AddDependency(d_framedata[SecondIP].allinone);
      d_framedata[SecondIP].delframe->AddDependency(d_framedata[SecondIP].decode);
    }
  else
    {
      // Keine B-Frames -> FirstIP nur decodieren, wenn SecondIP ein P-Frame ist.

      if (d_framedata[SecondIP].pictype == PICTYPE_P &&
          d_framedata[FirstIP].decode == NULL)
        {
          d_framedata[FirstIP].decode = new Task_Decode(d_framedata[FirstIP].framenr);
          d_framedata[FirstIP].decode  ->AddDependency(d_framedata[FirstIP].allinone);
          d_framedata[FirstIP].delframe->AddDependency(d_framedata[FirstIP].decode);
        }
    }


  /* Quer-Abhaengigkeiten fuer die Decode<->CodeFrame Tasks setzen */
  
  {for (uint32 i=1;i<SecondIP;i++)
    {
      d_framedata[i].allinone->AddDependency(d_framedata[FirstIP] .decode);
      d_framedata[i].allinone->AddDependency(d_framedata[SecondIP].decode);

      d_framedata[FirstIP] .delframe->AddDependency(d_framedata[i].allinone);
      d_framedata[SecondIP].delframe->AddDependency(d_framedata[i].allinone);
    }}

  // Und noch fuer die I---P Abhaengigkeit

  if (d_framedata[SecondIP].pictype == PICTYPE_P)
    {
      d_framedata[SecondIP].allinone->AddDependency(d_framedata[FirstIP] .decode);
      d_framedata[FirstIP] .delframe->AddDependency(d_framedata[SecondIP].decode);
    }
  
  /* Tasks in Queue einfuegen */

  {for (uint32 i=FirstIP;i<SecondIP;i++)
    {
      if (enable) cout << "INSERT INTO QUEUE " << i << endl;
      d_framedata[i].AppendToQueue(d_set);
    }}

  if (AreLastFrames)
    {
      if (enable) cout << "INSERT INTO QUEUE    SecondIP" << endl;
      d_framedata[SecondIP].AppendToQueue(d_set);
    }

  // ShowBlocked();


  /* Jetzt alle Frames aus d_framedata[] nach vorne schieben, so dass SecondIP an die erste
     Stelle kommt. */

  {for (uint32 i=0 ; i < d_nFrameData-SecondIP ; i++)
    d_framedata[i]=d_framedata[i+SecondIP];}

  // Die hinteren (jetzt leeren) Eintraege loeschen.
  {for (uint32 i=d_nFrameData-SecondIP ; i<d_nFrameData ; i++)
    d_framedata[i].Clear();}

  d_nFrameData -= SecondIP;

  if (AreLastFrames)
    {
      d_nFrameData--;
      assert(d_nFrameData==0);
      d_framedata[0].Clear();
    }

  return !AreLastFrames;
}










Task_ClassifyFrames* TaskPool_AllInOne::CreateTaskClassifyFrames() const
{
  int firstframe = d_startframe_of_next_classify;
  int nFrames    = d_nextframenr-d_startframe_of_next_classify;

  switch (d_frametypeproc)
    {
    case IOnly:
      return new Task_ClassifyFrames_IOnly      (firstframe,nFrames);
    case Pattern:
      {
        Task_ClassifyFrames_Regular* t = new Task_ClassifyFrames_Regular    (firstframe,nFrames);
        t->m_IDistance = d_IDistance;
        t->m_PDistance = d_PDistance;
        return t;
      }
    case SceneChange:
      return new Task_ClassifyFrames_SceneChange(firstframe,nFrames);
    default:
      assert(0);
      return 0;
    }
}


class Task_AllInOne* TaskPool_AllInOne::CreateMainTask(int framenr) const
{
  return new Task_CodeFrame(framenr);
}

// template class Queue<Task*>;

#include "utility/queue.cc"
