/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <control/taskpool_picstat.hh>
#include <coding/loading.hh>
#include <coding/varianz.hh>
#include <input/preproc.hh>

#include <iostream.h>
#include <stdlib.h>


TaskPool_PicStat::TaskPool_PicStat(InputPreprocessor& preproc)
  : d_preproc(preproc)
{
}


TaskPool_PicStat::~TaskPool_PicStat()
{
}


bool  TaskPool_PicStat::Finished()
{
  return d_set.IsEmpty() && !d_preproc.HasMoreInput();
}


Task* TaskPool_PicStat::GetTask()
{
  if (!d_set.IsEmpty())
    if (d_set.AskHead()->ReadyToExecute())
      {
        Task* task =d_set.AskHead();
        d_set.RemoveHead();
        return task;
      }

  return new Task_Loading(d_preproc);
}

void  TaskPool_PicStat::TaskCompleted(Task* task)
{
  if (dynamic_cast<Task_Loading*>(task))
    {
      Task* t = new Task_ComputeVarianzMean(dynamic_cast<Task_Loading*>(task)->frame_read);
      d_set.Append(t);
    }

  if (dynamic_cast<Task_ComputeVarianzMean*>(task))
    {
      static bool first=true;
      static long long lastmean,lastvar;

      Task_ComputeVarianzMean* t = dynamic_cast<Task_ComputeVarianzMean*>(task);

      long long mean,var;

      mean = ImageData::GetNr(t->m_framenr)->total_mean;
      var  = ImageData::GetNr(t->m_framenr)->total_variance;

      cout << t->m_framenr << " " << mean << " " << var << " ";

      if (first)
        { cout << "0 0\n"; first=false; }
      else
        {
          cout << abs(lastmean-mean) << " " << abs(lastvar-var) << endl;
        }
      lastmean=mean; lastvar=var;

      delete ImageData::GetNr(t->m_framenr);
    }
}

