/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include <control/taskpool_regular.hh>
#include <coding/loading.hh>
#include <coding/varianz.hh>
#include <coding/frameclass.hh>
#include <coding/frameclass_scenechange.hh>
#include <coding/quantscale.hh>
#include <coding/dct.hh>
#include <coding/taskmotion.hh>
#include <coding/taskmvsearch.hh>
#include <coding/taskdiffpics.hh>
#include <coding/taskmbtype.hh>
#include <coding/decoder.hh>
#include <control/taskdel.hh>
#include <control/taskseqhdr.hh>
#include <control/taskpichdr.hh>
#include <control/taskwrite.hh>
#include <input/preproc.hh>
#include <output/streamsink.hh>
#include <mpeg/constants.hh>
#include <options.hh>

#include <iostream.h>

#define enable 0    // Ausgabe aktivieren ?

/*
  Es gibt einen Fall, der nicht einfach behandelt werden kann, wenn man die P-Frames nur Decodiert,
  wenn es wirklich benoetigt wird:
  Bei einem I/P-Stream (ohne B-Frames) z.B. I1 P2 P3, wird zuerst der Graph fuer I1 P2 erzeugt, wobei kein
  Decodier-Task fuer das P2 erzeugt wird, da noch nicht klar ist, ob dieses benoetigt wird.
  Wird dann im zweiten Schritt P2 P3 codiert, so muss P2 decodiert werden, was jetzt aber gar nicht
  mehr moeglich ist, da I1 schon geloescht wurde.
  -> Die Strategie ist, alle P-Frames zu decodieren, auch wenn man das letzte nicht unbedingt
     decodieren muesste. Diese kleine Ineffizient ist zu verantworten, da das sowieso nur bei Streams
     ohne B-Frames auftritt, was eher die Ausnahme sein wird.
 */

#define MAX_P_DISTANCE 100
const uint32 nFrameDataStructs = 2*MAX_P_DISTANCE+1;

struct framedatastruct
{
  framedatastruct() { Clear(); }

  void Clear()
    {
      InitMBs=NULL;
      QScale=NULL;
      MEstInit1=NULL;
      MEstInit2=NULL;
      MVSearch1=NULL;
      MVSearch2=NULL;
      diff_task=NULL;
      MBType_task=NULL;
      pichdr_task=NULL;
      dct_task=NULL;
      decode_task=NULL;
      bitstream_task=NULL;
      write_task=NULL;
      delete_task=NULL;
    }
  
  void TaskFinished(Task* task)
    {
      if (enable) cout << "1 " << InitMBs << endl;
      if (InitMBs)   InitMBs  ->TaskFinished(task);
      if (enable) cout << "2 " << QScale << endl;
      if (QScale)    QScale   ->TaskFinished(task);
      if (enable) cout << "3 " << MEstInit1 << endl;
      if (MEstInit1) MEstInit1->TaskFinished(task);
      if (enable) cout << "4 " << MEstInit2 << endl;
      if (MEstInit2) MEstInit2->TaskFinished(task);
      if (enable) cout << "5 " << MVSearch1 << endl;
      if (MVSearch1) MVSearch1->TaskFinished(task);
      if (enable) cout << "6 " << MVSearch2 << endl;
      if (MVSearch2) MVSearch2->TaskFinished(task);
      if (enable) cout << "7 " << diff_task << endl;
      if (diff_task) diff_task->TaskFinished(task);
      if (enable) cout << "8 " << MBType_task << endl;
      if (MBType_task)    MBType_task->TaskFinished(task);
      if (enable) cout << "9 " << pichdr_task << endl;
      if (pichdr_task)    pichdr_task->TaskFinished(task);
      if (enable) cout << "10 " << dct_task << endl;
      if (dct_task)       dct_task   ->TaskFinished(task);
      if (enable) cout << "11 " << decode_task << endl;
      if (decode_task)    decode_task->TaskFinished(task);
      if (enable) cout << "12 " << bitstream_task << endl;
      if (bitstream_task) bitstream_task->TaskFinished(task);
      if (enable) cout << "13 " << write_task << endl;
      if (write_task)     write_task    ->TaskFinished(task);
      if (enable) cout << "14 " << delete_task << endl;
      if (delete_task)    delete_task   ->TaskFinished(task);
      if (enable) cout << "xxx" << endl;
    }

  void AppendToQueue(Queue<Task*>& queue)
    {
      if (InitMBs)        queue.Append(InitMBs);
      if (QScale)         queue.Append(QScale);
      if (MEstInit1)      queue.Append(MEstInit1);
      if (MEstInit2)      queue.Append(MEstInit2);
      if (MVSearch1)      queue.Append(MVSearch1);
      if (MVSearch2)      queue.Append(MVSearch2);
      if (diff_task)      queue.Append(diff_task);
      if (MBType_task)    queue.Append(MBType_task);
      if (pichdr_task)    queue.Append(pichdr_task);
      if (dct_task)       queue.Append(dct_task);
      if (decode_task)    queue.Append(decode_task);

      if (delete_task)    queue.AppendAtHead(delete_task);
      if (write_task)     queue.AppendAtHead(write_task);
      if (bitstream_task) queue.AppendAtHead(bitstream_task);
    }

  uint32     framenr;
  ImageData* imagedata;
  
  uint3      pictype;     // I,P,B

  Task* InitMBs;
  Task* QScale;
  Task* MEstInit1, *MEstInit2;      // nicht bei B-Frames
  Task* MVSearch1, *MVSearch2;      // nicht bei I-Frames
  Task* diff_task;                  //  """
  Task* MBType_task;                //  """
  Task* pichdr_task;
  Task* dct_task;
  Task* decode_task;                // nicht bei B-Frames
  Task* bitstream_task;
  Task* write_task;
  Task* delete_task;
};



TaskPool_Regular::TaskPool_Regular(InputPreprocessor& preproc,StreamSink& sink)
  : d_preproc(preproc),
    d_sink(sink),
    d_nextframenr(0),
    d_load_pending(false),
    d_classify_pending(false),
    d_startframe_of_next_classify(0),
    d_task_classify(NULL),
    d_last_write_task(NULL)
{
  SetFrameTypeProc(SceneChange,50);
  SetIPDistances(20,5);

  Task* t = new Task_InsertSeqHdr(0,d_preproc);
  d_set.Append(t);
  if (enable) cout << "NEW SeqHdr " << t << endl;

  d_framedata = new framedatastruct[nFrameDataStructs];
  d_nFrameData = 0;
}


TaskPool_Regular::~TaskPool_Regular()
{
  if (d_framedata) delete[] d_framedata;
}


void  TaskPool_Regular::ShowBlocked()
{
  for (uint32 i=0;i<d_set.Length();i++)
    {
      if (enable) cout << "Entry " << d_set.AskFromHead(i) << " is blocked, because of: ";
      d_set.AskFromHead(i)->ShowBlocked();
      if (enable) cout << endl;
    }
}

bool  TaskPool_Regular::Finished()
{
  return d_set.IsEmpty() && !d_preproc.HasMoreInput() && d_startframe_of_next_classify == d_nextframenr;
}


Task* TaskPool_Regular::GetTask()
{
  d_mutex.Lock();

  // Wenn noch ein Frame im Stream ist und der davorgehende schon geladen ist, naechsten Frame laden.
  // Auch dann nicht laden, wenn schon so viele Frames im voraus geladen wurden, wie in einem
  // Classify-Task bearbeitet werden, da sonst bei Single-Thread Verarbeitung der ganze Stream auf
  // einmal geladen wuerde.
  // Auch dann nicht laden, wenn der Bereich [letztes noch im Speicher befindliche Bild ; neues Bild]
  // groesser als ein (noch genau festzulegender 'TODO') Wert ist. Dadurch soll verhindert werden,
  // dass einige unguenstig plazierte, extrem langsame Threads das Weitercodieren blockieren und
  // somit Bilder weit in die Zukunft gelesen werden, ohne codiert werden zu koennen.

  if (!d_load_pending &&                                              // Kein Lese-Thread in Betrieb.
      d_nextframenr < d_startframe_of_next_classify + d_lookahead &&  // Max. 'd_lookahead' vorauslesen.
      d_nextframenr < ImageData::AskFirstFrame()+200) /* TODO: Threshold abhaengig von nThreads,Lookahead */
    {
      // Falls noch mehr Frames im Stream -> Lesen
      if (d_preproc.HasMoreInput())
        {
          // Naechsten Frame laden und Varianz/Mean bestimmen.
          
          Task* loading = new Task_Loading(d_preproc);
          if (enable) cout << "NEW loading " << loading << endl;
          Task* var     = new Task_ComputeVarianzMean(d_nextframenr);
          if (enable) cout << "NEW var " << var << endl;

          var->AddDependency(loading);
          d_set.Append(var);
          
          // Frame spaeter auch klassifizieren.
          
          if (d_task_classify==NULL)
            d_task_classify = CreateTaskClassifyFrames();
          
          d_task_classify->AddDependency(var);
          d_task_classify->m_nFrames++;
          
          d_nextframenr++;
          d_load_pending=true;
          
          d_mutex.Unlock();
          return loading;
        }
      else
        {
        }
    }
      
  
  /* Wenn genug Frames fuer einen Classify-Task gelesen wurden (d_lookahead) oder der Stream zu Ende ist,
     einen neuen Classify-Task erzeugen. */

  if (!d_classify_pending &&                                         // Letzter Classify-Task wurde beendet
      !d_load_pending &&           // (sonst wissen wir nicht, ob noch ein Frame folgt)
      (d_nextframenr >= d_startframe_of_next_classify+d_lookahead || // Genug Bilder geladen oder
       (!d_preproc.HasMoreInput() &&                         // [keine Bilder mehr im Stream und
        d_startframe_of_next_classify != d_nextframenr)))    //  noch nicht alle klassifiziert].
    {
      if (d_task_classify==NULL)
        d_task_classify = CreateTaskClassifyFrames();

      if (!d_preproc.HasMoreInput())
        d_task_classify->m_LastFrames=true;

      d_set.Append(d_task_classify);
      d_classify_pending=true;
      d_task_classify=NULL;
    }


  // Warteschlange abarbeiten, falls moeglich

  for (uint32 i=0;i<d_set.Length();i++)
    if (d_set.AskFromHead(i)->DependenciesFulfilled())
      {
        Task* task =d_set.AskFromHead(i);
        d_set.RemoveFromHead(i);
        d_mutex.Unlock();
        return task;
      }


#if 0
  {
    OutputLock;
    cerr << "No new task ready because loading or classification is too slow\n";
  }
#endif

  d_mutex.Unlock();
  return NULL;
}

void  TaskPool_Regular::TaskCompleted(Task* task)
{
  d_mutex.Lock();

  if (enable) cout << "Completed ------------------ " << task << endl;

  // Dependencies abarbeiten

  {for (uint32 i=0;i<d_set.Length();i++)
    {
      assert(d_set.AskFromHead(i));
      if (enable) cout << "AFHead " << d_set.AskFromHead(i) << endl;
      d_set.AskFromHead(i)->TaskFinished(task);
    }}

  if (enable) cout << "DTASKClass " << d_task_classify << endl;
  if (d_task_classify) d_task_classify->TaskFinished(task);
  if (enable) cout << "DLWrite " << d_last_write_task << endl;
  if (d_last_write_task == task) d_last_write_task=NULL;

  if (enable) cout << "x1\n";
  d_framedata[0].TaskFinished(task);
  if (enable) cout << "x2\n";


  // Task-Spezifische Routinen

  if (dynamic_cast<Task_Loading*>(task))
    {
      d_load_pending=false;
    }

  if (dynamic_cast<Task_ClassifyFrames*>(task))
    {
      Task_ClassifyFrames* task_class = dynamic_cast<Task_ClassifyFrames*>(task);
      d_startframe_of_next_classify = task_class->m_last_frame_type_assigned+1;
      d_classify_pending = false;

      uint32 FirstGOPFrame = task_class->m_firstframe;
      uint32 LastGOPFrame  = task_class->m_last_frame_type_assigned;
      uint32 nClassFrames  = LastGOPFrame - FirstGOPFrame + 1;
      bool   AreLastFrames = ((FirstGOPFrame + task_class->m_nFrames-1 == LastGOPFrame) &&
                              task_class->m_LastFrames);

      // --- Neue FrameData-Structs ausfuellen. ---

      assert(d_nFrameData + nClassFrames <= nFrameDataStructs);  // Es sollte genug Platz vorhanden sein.

      uint32 srcframeidx = FirstGOPFrame;
      uint32 destidx     = d_nFrameData;
      for ( ; srcframeidx <= LastGOPFrame ; srcframeidx++, destidx++ )
        {
          d_framedata[destidx].framenr   = srcframeidx;
          d_framedata[destidx].imagedata = ImageData::GetNr(srcframeidx);
          d_framedata[destidx].pictype   = d_framedata[destidx].imagedata->pichdr.m_picture_coding_type;

          assert(d_framedata[destidx].framenr == d_framedata[destidx].imagedata->nr);
        }
      d_nFrameData += nClassFrames;

      while (GeneratePartOfGraph(AreLastFrames))
        {
          /* NOP */
        }
    }

  d_mutex.Unlock();
}






bool TaskPool_Regular::GeneratePartOfGraph(bool AreLastFrames)
{
  // Ersten und zweiten I/P-Frame suchen

  const uint32 FirstIP = 0;  // Index in d_framedata.
  assert(d_framedata[FirstIP].pictype != PICTYPE_B); // Erster Frame in d_framedata[] ist immer I/P-Frame

  uint32 SecondIP=0;

  {for (uint32 i=1;i<d_nFrameData;i++)
    {
      if (d_framedata[i].pictype != PICTYPE_B)
        {
          SecondIP=i;
          break;
        }
    }}


  // Wenn noch keine zwei I/P-Frames im Buffer sind, dann kann nichts gemacht werden.
  if (SecondIP==0)
    return false;


  // Flag fuer die 'letzten Frames' wieder loeschen, falls das noch nicht die letzten
  // zu codierenden Bilder sind.

  if (AreLastFrames && SecondIP != d_nFrameData-1)
    AreLastFrames = false;


//  cout << FirstIP << " ---> " << SecondIP << endl;

  // ----- prev_ip / next_ip - Felder in ImageData (FirstIP;SecondIP] setzen -----

  {for (uint32 i=FirstIP+1 ; i<SecondIP ; i++)
    {
      d_framedata[i].imagedata->prev_ip = d_framedata[FirstIP ].imagedata;
      d_framedata[i].imagedata->next_ip = d_framedata[SecondIP].imagedata;
    }}

  if (d_framedata[SecondIP].imagedata->pichdr.m_picture_coding_type == PICTYPE_P)
    d_framedata[SecondIP].imagedata->prev_ip = d_framedata[FirstIP].imagedata;
    

  /* ------------------- Graph fuer den Bereich (FirstIP;SecondIP] aufbauen ----------------- */
  // Ausnahme: beim ersten Mal auch fuer FirstIP.
  // SecondIP aber noch nicht einfuegen

  // Task erzeugen, die fuer jeden Frame benoetigt werden:
  // InitMB, FillPicHdr, QScale, DCT, Bitstream, Write

  {for (uint32 i=FirstIP ; i<=SecondIP ; i++)
    {
      // Ausnahmebehandlung fuer FirstIP bei ersten Mal.
      if (i!=FirstIP || d_framedata[FirstIP].framenr==0)
        {
          uint32 framenr = d_framedata[i].framenr;

          Task* initmbs    = d_framedata[i].InitMBs        = new Task_InitMBs(framenr);
          Task* pichdr     = d_framedata[i].pichdr_task    = new Task_FillPicHdr(framenr);
          Task* quanttask  = d_framedata[i].QScale         = CreateTaskQuantizationScale(framenr);
          Task* dctquant   = d_framedata[i].dct_task       = new Task_DCTQuantize(framenr);
          Task* bitstream  = d_framedata[i].bitstream_task = new Task_WriteFrame(framenr);
          Task* write      = d_framedata[i].write_task     = new Task_WriteToStream(framenr,d_sink);
          Task* deletetask = d_framedata[i].delete_task    = new Task_DeleteFrame(framenr);

          if (enable)
            {
              cout << "NEW initmb (" << framenr << ") " << initmbs << endl;
              cout << "NEW pichdr (" << framenr << ") " << pichdr << endl;
              cout << "NEW qscale (" << framenr << ") " << quanttask << endl;
              cout << "NEW dct    (" << framenr << ") " << dctquant << endl;
              cout << "NEW bitstr (" << framenr << ") " << bitstream << endl;
              cout << "NEW write  (" << framenr << ") " << write << endl;
              cout << "NEW delete (" << framenr << ") " << deletetask << endl;
            }

          quanttask->AddDependency(initmbs);
          dctquant->AddDependency(pichdr);
          dctquant->AddDependency(quanttask);
          bitstream->AddDependency(dctquant);
          write->AddDependency(bitstream);
          deletetask->AddDependency(write);
        }
    }}


  /* CreateMEData-Tasks erzeugen.
   */

  // Bei FirstIP fuer forward-Prediction
  if (SecondIP>1          ||                                     // Es existieren B-Frames   oder
      d_framedata[SecondIP].imagedata->pichdr.m_picture_coding_type == PICTYPE_P) // Es folgt P-Frame
    {
      if (d_framedata[FirstIP].decode_task == NULL)
        {
          d_framedata[FirstIP].decode_task = new Task_Decode(d_framedata[FirstIP].framenr);
          if (enable)
            cout << "NEW decode1 (" << d_framedata[FirstIP].framenr << ") "
                 << d_framedata[FirstIP].decode_task << endl;

          d_framedata[FirstIP].decode_task->AddDependency(d_framedata[FirstIP].dct_task);
        }

      // TODO: MEData fuer jedes Frame erzeugen. Dafuer ist aber auch noch Modifikation in MV-Search
      // notwendig.
      if (d_framedata[FirstIP].MEstInit1==NULL)
        {
          Task_CreateMEData* t = new Task_CreateMEData(d_framedata[SecondIP].framenr,
                                                       d_framedata[FirstIP].framenr);
          if (enable)
            cout << "NEW MEData1 ("
                 << d_framedata[SecondIP].framenr << "/"
                 << d_framedata[FirstIP].framenr << ") " << t << endl;
          
          SetSearchParam(t->m_searchparam,FirstIP);
          d_framedata[FirstIP].MEstInit1 = t;
          t->AddDependency(d_framedata[FirstIP].decode_task);

          d_framedata[FirstIP].delete_task->AddDependency(t);
          d_framedata[SecondIP].delete_task->AddDependency(t);
        }
    }

#if 1
  // Bei SecondIP fuer backward-Prediction
  if (d_framedata[SecondIP].imagedata->pichdr.m_picture_coding_type == PICTYPE_P)
    {
      d_framedata[SecondIP].decode_task = new Task_Decode(d_framedata[SecondIP].framenr);
      if (enable)
        cout << "NEW decode2 (" << d_framedata[SecondIP].framenr << ") "
             << d_framedata[SecondIP].decode_task << endl;
      d_framedata[SecondIP].decode_task->AddDependency(d_framedata[SecondIP].dct_task);
      d_framedata[SecondIP].decode_task->AddDependency(d_framedata[FirstIP].decode_task);

      d_framedata[FirstIP].delete_task->AddDependency(d_framedata[SecondIP].decode_task);
    }

  if (SecondIP>1)             // Es existieren B-Frames
    {
      if (d_framedata[SecondIP].decode_task == NULL)
        {
          d_framedata[SecondIP].decode_task = new Task_Decode(d_framedata[SecondIP].framenr);
          if (enable)
            cout << "NEW decode3 (" << d_framedata[SecondIP].framenr << ") "
                 << d_framedata[SecondIP].decode_task << endl;
          d_framedata[SecondIP].decode_task->AddDependency(d_framedata[SecondIP].dct_task);
          d_framedata[SecondIP].decode_task->AddDependency(d_framedata[FirstIP].decode_task);

          d_framedata[FirstIP].delete_task->AddDependency(d_framedata[SecondIP].decode_task);
        }

      // TODO: MEData fuer jedes Frame erzeugen. Dafuer ist aber auch noch Modifikation in MV-Search
      // notwendig.
      if (d_framedata[SecondIP].MEstInit2==NULL)
        {
          Task_CreateMEData* t = new Task_CreateMEData(d_framedata[1].framenr,
                                                       d_framedata[SecondIP].framenr);
          if (enable)
            cout << "NEW MEData3 ("
                 << d_framedata[1].framenr << "/"
                 << d_framedata[SecondIP].framenr << ") " << t << endl;
            
          SetSearchParam(t->m_searchparam,SecondIP);
          d_framedata[SecondIP].MEstInit2 = t;
          t->AddDependency(d_framedata[SecondIP].decode_task);
          
          d_framedata[SecondIP].delete_task->AddDependency(t);
          d_framedata[1].delete_task->AddDependency(t);
        }
    }
#endif

#if 1
  /* MV-Such-, DiffPics- und MBTypes-Tasks erzeugen
   */

  {for (uint32 i=1;i<=SecondIP;i++)
    {
      switch (d_framedata[i].imagedata->pichdr.m_picture_coding_type)
        {
        case PICTYPE_I:
          break;

        case PICTYPE_B:
        case PICTYPE_P:
          {
            // Forward

            Task* mvtask1 = new Task_MVSearch (d_framedata[i].framenr,true);
            if (enable) cout << "NEW MVSearch> " << mvtask1 << endl;
            mvtask1->AddDependency(d_framedata[i].pichdr_task);
            mvtask1->AddDependency(d_framedata[FirstIP].MEstInit1);
            d_framedata[i].MVSearch1=mvtask1;

            // ggf. Backward

            Task* mvtask2=NULL;
            if (d_framedata[i].imagedata->pichdr.m_picture_coding_type == PICTYPE_B)
              {
                mvtask2 = new Task_MVSearch (d_framedata[i].framenr,false);
                if (enable) cout << "NEW MVSearch< " << mvtask2 << endl;
                mvtask2->AddDependency(d_framedata[i].pichdr_task);
                mvtask2->AddDependency(d_framedata[SecondIP].MEstInit2);
                d_framedata[i].MVSearch2=mvtask2;
              }

            Task* difftask = new Task_CreateDifferencePictures (d_framedata[i].framenr);
            if (enable) cout << "NEW diff " << difftask << endl;
            d_framedata[i].diff_task=difftask;
            difftask->AddDependency(mvtask1);
            difftask->AddDependency(mvtask2);

            //Task* mbtypes = new Task_MBType_TotalPrediction(d_framedata[i].framenr);
            Task* mbtypes = new Task_MBType_Variance(d_framedata[i].framenr);
            if (enable) cout << "NEW MBTypes " << mbtypes << endl;
            d_framedata[i].MBType_task=mbtypes;
            mbtypes->AddDependency(difftask);

            // d_framedata[i].delete_task->AddDependency(mbtypes);

            d_framedata[i].dct_task->AddDependency(mbtypes);

            if (d_framedata[i].imagedata->prev_ip!=NULL)
              d_framedata[FirstIP].delete_task->AddDependency(difftask);

            if (d_framedata[i].imagedata->next_ip!=NULL)
              d_framedata[SecondIP].delete_task->AddDependency(difftask);
          }
          break;
        }
    }}
#endif

  /* Zeitliche Quer-Abhaengigkeiten der Write-Tasks setzen. */

  d_framedata[SecondIP].write_task->AddDependency(d_last_write_task);
  d_last_write_task = d_framedata[SecondIP].write_task;
  {for (uint32 i=FirstIP+1;i<SecondIP;i++)
    {
      d_framedata[i].write_task->AddDependency(d_last_write_task);
      d_last_write_task = d_framedata[i].write_task;
    }}


  /* Tasks in Queue einfuegen */

  {for (uint32 i=FirstIP;i<SecondIP;i++)
    {
      if (enable) cout << "INSERT INTO QUEUE " << i << endl;
      d_framedata[i].AppendToQueue(d_set);
    }}

  if (AreLastFrames)
    {
      if (enable) cout << "INSERT INTO QUEUE    SecondIP" << endl;
      d_framedata[SecondIP].AppendToQueue(d_set);
    }

  // ShowBlocked();


  /* Jetzt alle Frames aus d_framedata[] nach vorne schieben, so dass SecondIP an die erste
     Stelle kommt. */

  {for (uint32 i=0 ; i < d_nFrameData-SecondIP ; i++)
    d_framedata[i]=d_framedata[i+SecondIP];}

  // Die hinteren (jetzt leeren) Eintraege loeschen.
  {for (uint32 i=d_nFrameData-SecondIP ; i<d_nFrameData ; i++)
    d_framedata[i].Clear();}

  d_nFrameData -= SecondIP;

  if (AreLastFrames)
    {
      d_nFrameData--;
      assert(d_nFrameData==0);
      d_framedata[0].Clear();
    }

  return !AreLastFrames;
}










Task_ClassifyFrames* TaskPool_Regular::CreateTaskClassifyFrames() const
{
  int firstframe = d_startframe_of_next_classify;
  int nFrames    = d_nextframenr-d_startframe_of_next_classify;

  switch (d_frametypeproc)
    {
    case IOnly:
      return new Task_ClassifyFrames_IOnly      (firstframe,nFrames);
    case Pattern:
      {
        Task_ClassifyFrames_Regular* t = new Task_ClassifyFrames_Regular    (firstframe,nFrames);
        t->m_IDistance = d_IDistance;
        t->m_PDistance = d_PDistance;
        return t;
      }
    case SceneChange:
      return new Task_ClassifyFrames_SceneChange(firstframe,nFrames);
    default:
      assert(0);
      return 0;
    }
}


Task_QuantizationScale* TaskPool_Regular::CreateTaskQuantizationScale(int framenr) const
{
  //  switch (d_frametypeproc)
    {
      //    case IOnly:
      Task_QuantizationScale* t = new Task_QuantizationScale_Constant(framenr);
      return t;
    }
}

void TaskPool_Regular::SetSearchParam(SearchParam& sparam,int idx)
{
  sparam.minx=-options.HSearchRange*2;
  sparam.maxx= options.HSearchRange*2-1;
  sparam.miny=-options.VSearchRange*2;
  sparam.maxy= options.VSearchRange*2-1;

  sparam.halfpel     = true; // !codedimg_PB.m_pichdr.m_full_pel_forward_vector;
  sparam.framesearch = true;
  sparam.top_field   = true;
  sparam.blk_width   = 16;
  sparam.blk_height  = 16;
  sparam.bitmap = &(d_framedata[idx].imagedata->decoded.AskBitmap_const(Bitmap_Y));
}

template class Queue<Task*>;

#include "utility/queue.cc"
