/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef CONTROL_TASKPOOL_REGULAR_HH
#define CONTROL_TASKPOOL_REGULAR_HH

#include <control/taskpool.hh>
#include <utility/thread.hh>



class TaskPool_Regular : public TaskPool
{
public:
  TaskPool_Regular(class InputPreprocessor&,class StreamSink&);
  ~TaskPool_Regular();

  void  SetFrameTypeProc(FrameTypeProc frtype,int frame_lookahead=1)
    { d_frametypeproc=frtype; d_lookahead=frame_lookahead; }
  void  SetIPDistances(int i,int p) { d_IDistance=i; d_PDistance=p; }

  bool  Finished();
  Task* GetTask();
  void  TaskCompleted(Task*);

  void ShowBlocked();

private:
  Mutex d_mutex;

  // Parameter

  enum FrameTypeProc d_frametypeproc;
  int d_IDistance,d_PDistance;

  // Variablen

  class InputPreprocessor& d_preproc;
  class StreamSink&        d_sink;
  Queue<Task*> d_set; // Eine (geordnete) Menge waere besser, da sie sich vielleicht schneller impl. laesst

  // fuer den Vor-Classify-Graph-Aufbau

  uint32 d_nextframenr;
  bool   d_load_pending;
  bool   d_classify_pending;

  uint32 d_lookahead;
  uint32 d_startframe_of_next_classify;
  class Task_ClassifyFrames* d_task_classify;
  

  // fuer den Nach-Classify-Graph-Aufbau

  struct framedatastruct* d_framedata;
  uint32 d_nFrameData;
  class Task* d_last_write_task;

  // private Methoden

  class Task_ClassifyFrames*    CreateTaskClassifyFrames() const;
  class Task_QuantizationScale* CreateTaskQuantizationScale(int framenr) const;
  bool GeneratePartOfGraph(bool AreLastFrames);
  void SetSearchParam(struct SearchParam&,int idx);
};

#endif
