/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "control/taskseqhdr.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "input/preproc.hh"
#include "options.hh"

void Task_InsertSeqHdr::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("insert SeqHdr",data.threadnr); cerr << "before frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  SequenceHeaderData sh;

  sh.m_width  = m_preproc.AskWidth();
  sh.m_height = m_preproc.AskHeight();
  sh.m_aspect_ratio_information = SAR_Square;
  sh.m_framerate100 = m_preproc.AskFramerate100();
  sh.m_bit_rate = MASK30;
  sh.m_vbv_buffer_size = MASK18;
  sh.m_constrained_parameters_flag = false;
  sh.m_load_intra_quantiser_matrix    = false;
  sh.m_load_nonintra_quantiser_matrix = false;

  sh.m_profile = PROFILE_Main;
  sh.m_level   = LEVEL_Main;
  sh.m_progressive_sequence = m_preproc.IsProgressiveSequence();
  sh.m_chroma_format = CHROMA_420;
  sh.m_low_delay = false;

  SeqHdrPool::SetSeqHdr(m_framenr,sh);
}


void Task_DeleteSeqHdr::Execute(ProcessData& data)
{
  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("delete SeqHdr",data.threadnr); cerr << "before frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  SeqHdrPool::ForgetSeqHdrBefore(m_framenr);
}
