/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "image/bitmap.hh"

template <class Pel> void Bitmap<Pel>::CreateBitmap(int width,int height)
{
  assert(BITMAP_ALIGN <= 16*8);

  d_width  = width;
  d_height = height; 

  d_rounded_width  = d_width;  if (d_width%16)  { d_rounded_width  += 16-(d_width%16);  }
  d_rounded_height = d_height; if (d_height%16) { d_rounded_height += 16-(d_height%16); }

  assert(d_rounded_width %16 == 0);
  assert(d_rounded_height%16 == 0);

  Pel* bitmap    = new Pel[d_rounded_width * d_rounded_height];
  d_frameptr     = new Pel* [d_rounded_height];
  d_top_field    = new Pel* [(d_rounded_height+1)/2];
  d_bottom_field = new Pel* [ d_rounded_height   /2];

  // Frame-Pointer ausfuellen
  {for (int y=0;y<d_rounded_height;y++)
    d_frameptr[y]= &bitmap[y*d_rounded_width];}

  // Field-Pointer ausfuellen
  {for (int y=0,y2=0;y2<d_rounded_height;y++)
    {
      d_top_field   [y]= &bitmap[y2*d_rounded_width]; y2++;
      d_bottom_field[y]= &bitmap[y2*d_rounded_width]; y2++;
    }}
}

template <class Pel> void Bitmap<Pel>::FreeBitmap()
{
  if (d_frameptr)
    {
      delete[] d_frameptr[0];
      delete[] d_frameptr;
      delete[] d_top_field;
      delete[] d_bottom_field;

      d_frameptr=NULL;
    }
}

template <class Pel> void Bitmap<Pel>::FillBorder(Pel val)
{
  if (d_width<d_rounded_width)
    {
      for (int x=d_width;x<d_rounded_width;x++)
        for (int y=0;y<d_rounded_height;y++)
          d_frameptr[y][x] = val;
    }

  if (d_height<d_rounded_height)
    {
      for (int y=d_height;y<d_rounded_height;y++)
        for (int x=0;x<d_rounded_width;x++)
          d_frameptr[y][x] = val;
    }
}


template <class Pel> void Bitmap<Pel>::DuplicateBorder()
{
  if (d_width<d_rounded_width)
    {
      for (int x=d_width;x<d_rounded_width;x++)
        for (int y=0;y<d_rounded_height;y++)
          d_frameptr[y][x] = d_frameptr[y][x-1];
    }

  if (d_height<d_rounded_height)
    {
      for (int y=d_height;y<d_rounded_height;y++)
        for (int x=0;x<d_rounded_width;x++)
          d_frameptr[y][x] = d_frameptr[y-1][x];
    }
}


template class Bitmap<Pixel>;
template class Bitmap<DiffPixel>;
template class Bitmap<MVPixel>;
