/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef IMAGE_CODEDIMG_HH
#define IMAGE_CODEDIMG_HH

#include "types.hh"
#include "mpeg/syntaxelems.hh"
#include "image/array2.hh"
#include "image/image.hh"

struct DCTBlock
{
  short m_coeff[8 /* v */][8 /* u */];
};

struct MotionVector
{
  uint1 m_vertical_field_select;  // 0 ist TopField.
   int2 m_dmvector[2];

  int   m_habs, m_vabs;
};

struct Macroblock
{
  Macroblock();

#if 0
  bool  m_SkipCoding;  /* Nur als Flag, dass die Codierung dieses MB uebersprungen werden soll.
			  Hat nichts mit uebersprungenen Makrobloecken zu tun, dafuer ist
			  'm_Skip' da. */
#endif

  bool  m_IsIntra;

  double m_quantiser_scale;      /* 1 bis 127 */
  uint7  m_quantiser_scale_code; /* 1 bis 31  */

  DCTBlock m_blks[12]; bool m_CodeBlk[12];
  bool m_HasData;

  DCTBlock* m_y;  bool* m_CodeY;
  DCTBlock* m_cb; bool* m_CodeCb;
  DCTBlock* m_cr; bool* m_CodeCr;

  // Fuer Frame-Pictures mit 'frame_pred_frame_dct==0' und vorhandenen DCT-Koeff.
  bool m_FieldDCT;

  enum FrameMotionType { FRMT_FieldBased=1, FRMT_FrameBased=2, FRMT_DualPrime=3 };
  enum FieldMotionType { FIMT_FieldBased=1, FIMT_16x8MC=2,     FIMT_DualPrime=3 };
  union {
    // MotionComp.Type fuer Pictures mit 'picture_structure==frame' und
    // 'frame_pred_frame_dct==0'. Ist 'frame_pred_frame_dct==1', so wird FRMT_FrameBased angenommen,
    // m_frame_motion_type aber nicht in den Bitstream codiert.
    enum FrameMotionType m_frame_motion_type;
    
    // MotionComp.Type fuer Pictures mit 'picture_structure!=frame'.
    enum FieldMotionType m_field_motion_type;
  };
  uint2 m_motion_vector_count;
  enum { MVFormat_Field, MVFormat_Frame } m_mv_format;
  bool  m_dmv;

  void SetFrameMotionType(FrameMotionType); // Bei SpatialScalability ggf. noch 'm_motion_vector_count++'
  void SetFieldMotionType(FieldMotionType);

  MotionVector m_vector[2 /*r*/][2 /*s*/];
  bool m_HasMotionForward;     // <- wird auch als Consealment-Vektor-Flag ( vector[0][0] ) benutzt
  bool m_HasMotionBackward;

#define m_forward1  m_vector[0][0]
#define m_forward2  m_vector[1][0]
#define m_backward1 m_vector[0][1]
#define m_backward2 m_vector[1][1]

  bool m_BeginSlice;   // Mit diesem MB beginnt neuer Slice.
  bool m_EndSlice;     // Mit diesem MB endet der aktuelle Slice;
  bool m_Skip;         // Diesen Makroblock nicht codieren.
};

#if 0
class ImageData
{
public:
  ImageData() : finished(false) { }
  ~ImageData() { }

  // enum ProcessOrder  process_queue_pos;
  bool finished;

  int nr;
  Image<Pixel>* source;

  PictureHeaderData  pichdr;
  Array2<uint16>     mean;
  Array2<uint32>     variance;
  Array2<Macroblock> mblks;

  const class MVData* mvdata;

  Image<DiffPixel>   pred_forw;  // Differenzbild fuer Forward-Prediction (P-Frames)
  Image<DiffPixel>   pred_backw;
  Image<DiffPixel>   pred_bidir;

  // Buffer output;

  Image<Pixel> decoded;
};
#endif

#endif
