/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef IMAGE_IMAGE_HH
#define IMAGE_IMAGE_HH

#include "types.hh"
#include "image/bitmap.hh"

enum ImageColorspace { Uninitialized, RGB_Undefined, YUV, YCrCb };
enum ImageBitmap { Bitmap_Red = 0, Bitmap_Green = 1, Bitmap_Blue = 2,
                   Bitmap_Y   = 0, Bitmap_Cr    = 1, Bitmap_Cb   = 2,
                                   Bitmap_U     = 1, Bitmap_V    = 2
};


template <class Pel> class Image
{
public:
  Image() : IsFramePicture(false), IsFieldPicture(false), Is2FieldPicture(false),
    d_colorspace(Uninitialized) { }
  ~Image() { }
  void CreateImage(ImageColorspace cs,int width,int height,bool halfwidth=false,bool halfheight=false)
  {
    if (d_colorspace==cs)
    {
      assert(AskWidth() ==width);
      assert(AskHeight()==height);
      return;
    }
    assert(d_colorspace==Uninitialized); d_colorspace=cs;
    d_bm[0].CreateBitmap(width,height);

    if (halfwidth)  width  /= 2;
    if (halfheight) height /= 2;

    d_bm[1].CreateBitmap(width,height);
    d_bm[2].CreateBitmap(width,height);

    m_subsampledH = halfwidth;
    m_subsampledV = halfheight;
  }

  ImageColorspace AskColorspace() const             { return d_colorspace; }
  void            SetColorspace(ImageColorspace cs) { d_colorspace=cs; }

  int  AskWidth()  const { return d_bm[0].AskWidth();  }
  int  AskHeight() const { return d_bm[0].AskHeight(); }

  int  AskRoundedWidth()  const { return d_bm[0].AskRoundedWidth();  }
  int  AskRoundedHeight() const { return d_bm[0].AskRoundedHeight(); }

  Bitmap<Pel>&       AskBitmap      (ImageBitmap bm_id)       { return d_bm[bm_id]; }
  const Bitmap<Pel>& AskBitmap_const(ImageBitmap bm_id) const { return d_bm[bm_id]; }

  void DuplicateBorder()
    {
      for (int i=0;i<3;i++)
        d_bm[i].DuplicateBorder();
    }

  void AlignSize()       // Quick-Fix
    {
      for (int i=0;i<3;i++)
        d_bm[i].AlignSize();
    }

  bool IsSubsampledH() const { return m_subsampledH; }
  bool IsSubsampledV() const { return m_subsampledV; }



  bool IsFramePicture;
  bool IsFieldPicture;  bool IsTopField;
  bool Is2FieldPicture; bool TopFieldFirst;   bool SkipFirstField; bool SkipLastField;

  bool m_subsampledH;
  bool m_subsampledV;

private:
  ImageColorspace d_colorspace;
  Bitmap<Pel>     d_bm[3];
};

#endif
