/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef INPUT_PREPROC_HH
#define INPUT_PREPROC_HH

#include "types.hh"
#include "image/image.hh"

/*
  NOTE

  Eingabe ist entweder: - Progressive Frames. Jeweils ein Frame pro Eingabebild.
                        - NonProgressive Frames. Jeweils zwei Fields pro Eingabebild,
                          wobei sowohl das obere, als auch das untere zuerst angezeigt
                          werden kann.
                        - Fields. Jeweils ein Eingabebild pro Field. Dabei muss bekannt
                          sein, welches Field (Top/Bottom) gemeint ist.
 */


class InputPreprocessor
{
public:
  virtual ~InputPreprocessor() { }

  virtual bool         IsProgressiveSequence() const = 0;
  virtual bool         HasMoreInput() const = 0;
  virtual int          AskNextImageNum() const = 0;
  virtual const char*  AskNextImageFilename() const = 0;
  virtual class Image<Pixel>* GetNextImage(class Image<Pixel>* reuse=NULL) = 0;
  virtual int          AskFramerate100() const = 0;

  virtual int AskWidth() const = 0;
  virtual int AskHeight() const = 0;
};

#endif
