/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef INPUT_PREPROC_AVI_HH
#define INPUT_PREPROC_AVI_HH

#ifndef DISABLE_MJPEG
#define DISABLE_MJPEG 0
#endif

#if !DISABLE_MJPEG

#include "input/preproc.hh"
#include "utility/queue.hh"
#include <fstream.h>
#include "types.hh"

extern "C" {
#include <stdio.h>
#include <jpeglib.h>
}

#define CHKID(a,b,c,d) ((((uint32)a)<<24)|(((uint32)b)<<16)|(((uint32)c)<<8)|((uint32)d))

const uint32 ID_RIFF = CHKID('R','I','F','F');
const uint32 ID_AVI  = CHKID('A','V','I',' ');
const uint32 ID_LIST = CHKID('L','I','S','T');
const uint32 ID_hdrl = CHKID('h','d','r','l');
const uint32 ID_avih = CHKID('a','v','i','h');
const uint32 ID_strl = CHKID('s','t','r','l');
const uint32 ID_strh = CHKID('s','t','r','h');
const uint32 ID_strf = CHKID('s','t','r','f');
const uint32 ID_strd = CHKID('s','t','r','d');
const uint32 ID_vids = CHKID('v','i','d','s');
const uint32 ID_auds = CHKID('a','u','d','s');
const uint32 ID_MJPG = CHKID('M','J','P','G');
const uint32 ID_0000 = CHKID( 0,  0 , 0 , 0 );
const uint32 ID_indx = CHKID('i','n','d','x');
const uint32 ID_idx1 = CHKID('i','d','x','1');
const uint32 ID___db = CHKID('0','0','d','b');
const uint32 ID___dc = CHKID('0','0','d','c');
const uint32 ID___wb = CHKID('0','0','w','b');
const uint32 ID_movi = CHKID('m','o','v','i');
const uint32 ID_rec  = CHKID('r','e','c',' ');
const uint32 ID_JUNK = CHKID('J','U','N','K');
const uint32 ID_AVI1 = CHKID('A','V','I','1');
const uint32 ID_DC30 = CHKID('D','C','3','0');

const uint32 AVIH_HAS_IDX1    = (1<<4);
const uint32 AVIH_USE_INDEX   = (1<<5);
const uint32 AVIH_INTERLEAVED = (1<<8);
const uint32 AVIH_CAPTURE     = (1<<16);
const uint32 AVIH_COPYRIGHTED = (1<<17);

const int MAX_AVI_STREAMS=10;


class InputPreprocessor_AVI : public InputPreprocessor
{
public:
  InputPreprocessor_AVI();
  ~InputPreprocessor_AVI();

  // Initialisierungsroutinen sind in dieser Reihenfolge anzuwenden !
  void SetInputFile(const char* filename,bool interlaced=false);
  void SetFrameRange(uint32 firstframe,uint32 lastframe);
  void SetSmallMode();
  void Init();

  bool         IsProgressiveSequence() const;
  bool         HasMoreInput() const;
  int          AskNextImageNum() const;
  const char*  AskNextImageFilename() const;
  class Image<Pixel>* GetNextImage(class Image<Pixel>* reuse);
  int          AskFramerate100() const { return d_framerate100; }

  int          AskWidth() const  { return d_width  -d_width %16; }
  int          AskHeight() const { return d_height -d_height%16; }

private:
  struct   jpeg_decompress_struct cinfo;
  struct   jpeg_error_mgr jerr;
  ifstream infile;

  ofstream audiostream;

  uint32 d_FirstFrame,d_LastFrame;
  bool   d_smallmode;
  uint32 d_nPicsInStream;
  uint32 d_nStreams;
  uint32 d_streamtype[MAX_AVI_STREAMS];
  uint32 d_codec     [MAX_AVI_STREAMS];
  bool   d_IsInterlaced;

  int    d_VideoStreamIdx; // Idx des Videostreams, der geladen werden soll.

  uint32 piccnt;

  int d_width,d_height;
  int d_framerate100;

  enum State { AVI_BeforeHDRL, AVI_BeforeMOVI, AVI_BeforeIDX1,
               BeforeData };

  struct Pos
  {
    uint32     current_chunk;
    uint32     current_chunk_type;
    uint32     chunk_length;
    uint64     abs_chunk_end;
    enum State state;
  };

  struct Buffer
  {
    uint8* buf;
    uint32 size;
  };

  Buffer        d_lastframe;
  Queue<Pos>    d_chkqueue;
  Queue<Buffer> d_bufqueue;
  bool   d_EndOfFileReached;
  uint32 d_padding;
  uint32 d_avih_flags;

  uint32 ReadID();
  uint32 ReadDWORD();

  void ReadChunk();  // Liest den Chunk-Header (incl. SubType bei RIFF und LIST ein)
                     // und haengt ihn an die Queue
  void EndChunk();   // Spring an das Ende des aktuellen Chunks und entfernt ihn aus der Queue
  bool AreAtEndOfChunk() const;
  void ProcessFileUntilData(bool firsttime);
  void ReadAVIHeader(bool firstheader);
  void ReadHDRL();

  Buffer GetNextBuffer();

  class Image<Pixel>* ReadImage_MJPG(class Image<Pixel>*,bool skip);
  class Image<Pixel>* ReadImage_RAW (class Image<Pixel>*,bool skip);

  friend boolean fill_input_buffer (j_decompress_ptr cinfo);
};

#endif

#endif
