/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/


#include "input/preproc_v4l_sglstep.hh"
#include "image/image.hh"

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>



/************* 1) HIER DEN PFAD GGF AENDERN !!! **********/
/************* 2) Dieses Include muss unbedingt am Schluss stehen, sonst gibt es
                  haufenweise Typ-Konflikte mit vielen Typen aus verschiedenen Includes. */
#include <linux/videodev.h>



static struct video_mmap vidmmap;
static struct video_mbuf vidmbuf;



InputPreprocessor_V4L_SglStep::V4LParameter::V4LParameter()
  : m_width(12*16), m_height(9*16),
    m_nFrames(20), m_framerate100(2500), m_crop_top(0), m_crop_bottom(0)
{
}

InputPreprocessor_V4L_SglStep::InputPreprocessor_V4L_SglStep(const V4LParameter& p)
  : d_parameter(p),
    d_next_num(0),
    //d_remote(Panasonic_FS1)
    //d_remote(Samsung_SV300W)
    //d_remote(Pioneer_CLD950)
    d_remote(Sony_VCR)
{
  sprintf(d_next_name,"V4L-frame%05d",0);



  fd = open("/dev/video",O_RDWR);
  if (fd==-1)
    { perror("open (device): "); exit(10); }

  if (-1 == ioctl(fd,VIDIOCGMBUF,&vidmbuf)) {
    perror("ioctl VIDIOCGMBUF");
  }

  map = mmap(0,vidmbuf.size, PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
  if ((unsigned char*)-1 == map) {
    perror("mmap");
  }

  vidmmap.frame=0;
  vidmmap.width =d_parameter.m_width;
  vidmmap.height=d_parameter.m_height;
  vidmmap.format = VIDEO_PALETTE_YUV422;
}


InputPreprocessor_V4L_SglStep::~InputPreprocessor_V4L_SglStep()
{
  munmap((char*)map,vidmbuf.size);
  close(fd);
}


bool InputPreprocessor_V4L_SglStep::IsProgressiveSequence() const
{
  return d_parameter.m_height < 400;
}


bool InputPreprocessor_V4L_SglStep::HasMoreInput() const
{
  if (d_remote.EndOfStream())
    return false;

  if (d_next_num >= d_parameter.m_nFrames)
    { return false; }
  else
    { return true; }
}

class Image<Pixel>* InputPreprocessor_V4L_SglStep::GetNextImage(class Image<Pixel>* reuse)
{
  d_remote.WaitForNextPicture();
  
  vidmmap.frame=0;
  if (-1 == ioctl(fd,VIDIOCMCAPTURE,&vidmmap)) {
    perror("ioctl VIDIOCMCAPTURE");
    exit(10);
  }

  vidmmap.frame=0;
  if (-1 == ioctl(fd,VIDIOCSYNC,&vidmmap)) {
    perror("ioctl VIDIOCSYNC");
    exit(10);
  }

  d_remote.NextPicture();

  Image<Pixel>* img = new Image<Pixel>;
  img->CreateImage(YCrCb ,
                   AskWidth(),
                   AskHeight(),
                   true,true);

  Bitmap<Pixel>& yy = img->AskBitmap(Bitmap_Y);
  Bitmap<Pixel>& cb = img->AskBitmap(Bitmap_Cb);
  Bitmap<Pixel>& cr = img->AskBitmap(Bitmap_Cr);

  unsigned char* m=map + 4*(d_parameter.m_width/2)*d_parameter.m_crop_top;
  for (int y=0;y<AskHeight();y+=2)
    {
      Pixel* yp  = yy.AskFrame()[y];
      Pixel* cbp = cb.AskFrame()[y/2];
      Pixel* crp = cr.AskFrame()[y/2];

      for (int x=0;x<AskWidth()/2;x++)
        {
           yp[2*x]   = m[0];
           yp[2*x+1] = m[2];
          cbp[x]     = m[1];
          crp[x]     = m[3];

          m+=4;
        }

      m += (d_parameter.m_crop_right+d_parameter.m_crop_left)*2;

      yp  = yy.AskFrame()[y+1];

      for (int x=0;x<AskWidth();x++)
        {
          yp[x] = *m;
          //cbp[x]     = m[1];  // TODO: Mittelung mit letzter Zeile
          //crp[x]     = m[3];
          m+=2;
        }

      m += (d_parameter.m_crop_right+d_parameter.m_crop_left)*2;
    }

  img->IsFramePicture = true;
  img->DuplicateBorder();

  d_next_num++;
  sprintf(d_next_name,"V4L-frame%05d",d_next_num);

  return img;
}
