/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef INPUT_PREPROC_V4L_SGLSTEP_HH
#define INPUT_PREPROC_V4L_SGLSTEP_HH

#include "types.hh"
#include "input/preproc.hh"
#include "image/image.hh"
#include "input/remote.hh"

class InputPreprocessor_V4L_SglStep : public InputPreprocessor
{
public:
  struct V4LParameter
  {
    V4LParameter();

    int  m_width,m_height;
    int  m_nFrames;
    int  m_framerate100;  // Framerate*100
    
    int  m_crop_top;
    int  m_crop_bottom;
    int  m_crop_left;      // wird ignoriert
    int  m_crop_right;
  };

   InputPreprocessor_V4L_SglStep(const V4LParameter&);
  ~InputPreprocessor_V4L_SglStep();

  bool         IsProgressiveSequence() const;
  bool         HasMoreInput() const;
  int          AskNextImageNum() const { return d_next_num; }
  const char*  AskNextImageFilename() const { return d_next_name; }
  class Image<Pixel>* GetNextImage(class Image<Pixel>*);
  int          AskFramerate100() const { return d_parameter.m_framerate100; }

  int AskWidth() const  { return d_parameter.m_width -d_parameter.m_crop_left -d_parameter.m_crop_right;}
  int AskHeight() const { return d_parameter.m_height -d_parameter.m_crop_top -d_parameter.m_crop_bottom;}

private:
  V4LParameter d_parameter;
  int          d_next_num;
#define MaxNameLen 20
  char         d_next_name[MaxNameLen];

  int fd;
  unsigned char* map;

  Remote_IR    d_remote;
};

#endif

