/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "math/dct2_arai.hh"
#include <math.h>
#include <assert.h>
#include <iostream.h>
#include <iomanip.h>

#ifndef PI
#define PI 3.1415926535
#endif

int c1,c2;

#define SC(x,u,v) { D(x,0); fbuf[u][v] = ((long long)(x*(1<<FDCTScale)) >> (FDCTScale-FDCT_STAGE1_NK)); }
#define D(x,s) { cout << setbase(16) << setw(16) << setfill('0') << ((long long)(x*(1<<s))) << " "; \
  fbuf[c1][c2] = ((long long)(x*(1<<s))); c2++; if (c2==8) { c2=0; c1++; } cout << setfill(' '); }


static DCTFloat  A1 =  0.707106781;
static DCTFloat MA2 = -0.5411961;
static DCTFloat  A3 =  0.707106781;
static DCTFloat  A4 =  1.306562965;
static DCTFloat  A5 =  0.382683432;

static DCTFloat C2;
static DCTFloat C4;
static DCTFloat C6;
static DCTFloat Q;
static DCTFloat R;

static DCTFloat OutputScaleFactor[8];  // fuer FDCT
static DCTFloat InputScaleFactor[8];   // fuer IDCT

static struct DummyInit
{
  DummyInit()
  {
    // FDCT

     A1 = sqrt(0.5);
    MA2 = -(cos(PI/8.0) - cos(3.0*PI/8.0));
     A3 = sqrt(0.5);
     A4 = cos(PI/8.0) + cos(3.0*PI/8.0);
     A5 = cos(3.0*PI/8.0);

    OutputScaleFactor[0] = sqrt(0.5)/2.0;
    {for (int i=1;i<8;i++) OutputScaleFactor[i] = 0.25/cos(PI*(double)i/16.0);}


    // IDCT

    C2 = 2.0*cos(PI/8.0);
    C4 = sqrt(2.0);
    C6 = 2.0*cos(3.0*PI/8.0);
    Q  = -C2+C6;
    R  =  C2+C6;

    InputScaleFactor[0] = 0.5*0.5*sqrt(2);
    {for (int i=1;i<8;i++) InputScaleFactor[i] = 0.5*cos(PI*((double)i)/16.0);}
  }
} dummyinit;


DCT2_AraiAguiNakajima::DCT2_AraiAguiNakajima()
{
  RemoveOutputScalefactors();
}

void DCT2_AraiAguiNakajima::RemoveOutputScalefactors()
{
  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = OutputScaleFactor[u]*OutputScaleFactor[v];
      }
}

void DCT2_AraiAguiNakajima::SetOutputScalefactors(const DCTFloat* fact[8])
{
  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = OutputScaleFactor[u]*OutputScaleFactor[v] * fact[v][u];
      }
}


void DCT2_AraiAguiNakajima::PerformFDCT2(const short* in[8] /* [y][x] */, DCTBlock& out)
{
  DCTFloat tmp[8][8];

  DCTFloat x1,x2,x3,x4,x5,x6,x7,x8;
  DCTFloat y1,y2,y3,y4,y5,y6,y7,y8;

  // Zeilentransformationen

  for (int u=0;u<8;u++)
    {
      x1 = in[0][u] + in[7][u];
      y8 = in[0][u] - in[7][u];
      x2 = in[1][u] + in[6][u];
      x7 = in[1][u] - in[6][u];
      x3 = in[2][u] + in[5][u];
      x6 = in[2][u] - in[5][u];
      x4 = in[3][u] + in[4][u];
      x5 = in[3][u] - in[4][u];


      y1 =   x1 + x4;
      y2 =   x2 + x3;
      y3 =   x2 - x3;
      y4 =   x1 - x4;
      y5 = - x5 - x6;
      y6 =   x6 + x7;
      y7 =   x7 + y8;

      x1 = y1 + y2;
      x2 = y1 - y2;
      x3 = y3 + y4;
      
      DCTFloat a5 = (y5+y7) * A5;

      y3 =  A1 * x3;
      x5 = MA2 * y5 - a5;
      y6 =  A3 * y6;
      x7 =  A4 * y7 - a5;

#if SHOW_M1
      cout << "FM1: ";
      D(x1,FDCT_PostM1_NK); D(x2,FDCT_PostM1_NK); D(y3,FDCT_PostM1_NK); D(y4,FDCT_PostM1_NK);
      cout << "\n     ";
      D(x5,FDCT_PostM1_NK); D(y6,FDCT_PostM1_NK); D(x7,FDCT_PostM1_NK); D(y8,FDCT_PostM1_NK);
      cout << endl;
#endif

      x3 = y4 + y3;
      x4 = y4 - y3;
      x6 = y8 + y6;
      x8 = y8 - y6;

      
      tmp[0][u] = x1;
      tmp[4][u] = x2;
      tmp[2][u] = x3;
      tmp[6][u] = x4;
      tmp[5][u] = x5 + x8;
      tmp[1][u] = x6 + x7;
      tmp[7][u] = x6 - x7;
      tmp[3][u] = x8 - x5;
    }

  // Spaltentransformationen

  for (int v=0;v<8;v++)
    {
      x1 = tmp[v][0] + tmp[v][7];
      x2 = tmp[v][1] + tmp[v][6];
      x3 = tmp[v][2] + tmp[v][5];
      x4 = tmp[v][3] + tmp[v][4];
      x5 = tmp[v][3] - tmp[v][4];
      x6 = tmp[v][2] - tmp[v][5];
      x7 = tmp[v][1] - tmp[v][6];
      y8 = tmp[v][0] - tmp[v][7];

      y1 =   x1 + x4;
      y2 =   x2 + x3;
      y3 =   x2 - x3;
      y4 =   x1 - x4;
      y5 = - x5 - x6;
      y6 =   x6 + x7;
      y7 =   x7 + y8;

      x1 = y1 + y2;
      x2 = y1 - y2;
      x3 = y3 + y4;
      
      DCTFloat a5 = (y5+y7) * A5;

      y3 =  A1 * x3;
      x5 = MA2 * y5 - a5;
      y6 =  A3 * y6;
      x7 =  A4 * y7 - a5;

#if SHOW_M2
      cout << "FM2: ";
      D(x1,FDCT_PostM2_NK); D(x2,FDCT_PostM2_NK); D(y3,FDCT_PostM2_NK); D(y4,FDCT_PostM2_NK);
      cout << "\n     ";
      D(x5,FDCT_PostM2_NK); D(y6,FDCT_PostM2_NK); D(x7,FDCT_PostM2_NK); D(y8,FDCT_PostM2_NK);
      cout << endl;
#endif

      x3 = y3 + y4;
      x4 = y4 - y3;
      x6 = y6 + y8;
      x8 = y8 - y6;

      y5 = x5 + x8;
      y6 = x6 + x7;
      y7 = x6 - x7;
      y8 = x8 - x5;

      out.m_coeff[v][0] = (short)(x1 * d_fact[v][0] + 0.499999);
      out.m_coeff[v][4] = (short)(x2 * d_fact[v][4] + 0.499999);
      out.m_coeff[v][2] = (short)(x3 * d_fact[v][2] + 0.499999);
      out.m_coeff[v][6] = (short)(x4 * d_fact[v][6] + 0.499999);
      out.m_coeff[v][5] = (short)(y5 * d_fact[v][5] + 0.499999);
      out.m_coeff[v][1] = (short)(y6 * d_fact[v][1] + 0.499999);
      out.m_coeff[v][7] = (short)(y7 * d_fact[v][7] + 0.499999);
      out.m_coeff[v][3] = (short)(y8 * d_fact[v][3] + 0.499999);
    }
}









IDCT2_AraiAguiNakajima::IDCT2_AraiAguiNakajima()
{
  RemoveOutputScalefactors();
}

void IDCT2_AraiAguiNakajima::RemoveOutputScalefactors()
{
  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = InputScaleFactor[u]*InputScaleFactor[v];
      }
}

void IDCT2_AraiAguiNakajima::SetOutputScalefactors(const DCTFloat* fact[8])
{
  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = InputScaleFactor[u]*InputScaleFactor[v] * fact[v][u];
      }
}


void IDCT2_AraiAguiNakajima::PerformIDCT2(const short* in[8], uint8* out[8],bool add)
{
  DCTFloat fin[8][8];
  DCTFloat tmp[8][8];

  DCTFloat a0,a1,a2,a3,a4,a5,a6,a7;
  DCTFloat       b2,   b4,b5,b6   ;
  DCTFloat n0,n1,n2,n3            ;
  DCTFloat          m3,m4,m5,m6,m7;
  DCTFloat tmp1,tmp2,tmp3,tmp4;


  // Zeilentransformationen

  {for (int u=0;u<8;u++)
    {
      {for (int v=0;v<8;v++)
        fin[v][u] = in[v][u] * d_fact[v][u];}

      a0 = fin[0][u];
      a1 = fin[4][u];
      a2 = fin[2][u]-fin[6][u];
      a3 = fin[2][u]+fin[6][u];
      a4 = fin[5][u]-fin[3][u];
      tmp1 = fin[1][u]+fin[7][u];
      tmp2 = fin[3][u]+fin[5][u];
      a5 = tmp1 - tmp2;
      a6 = fin[1][u]-fin[7][u];
      a7 = tmp1 + tmp2;

      tmp4 = C6*(a4+a6);
      b2 = a2 * C4;
      b4 = Q*a4-tmp4;
      b5 = a5*C4;
      b6 = R*a6-tmp4;

      tmp3 = b6-a7;
      n0 = tmp3-b5;
      n1 = a0-a1;
      n2 = b2-a3;
      n3 = a0+a1;

      m3 = n1+n2;
      m4 = n3+a3;
      m5 = n1-n2;
      m6 = n3-a3;
      m7 = b4-n0;

      tmp[0][u] = m4+a7;
      tmp[1][u] = m3+tmp3;
      tmp[2][u] = m5-n0;
      tmp[3][u] = m6-m7;
      tmp[4][u] = m6+m7;
      tmp[5][u] = m5+n0;
      tmp[6][u] = m3-tmp3;
      tmp[7][u] = m4-a7;
    }}


  // Spaltentransformationen

  {for (int v=0;v<8;v++)
    {
      a0 = tmp[v][0];
      a1 = tmp[v][4];
      a2 = tmp[v][2]-tmp[v][6];
      a3 = tmp[v][2]+tmp[v][6];
      a4 = tmp[v][5]-tmp[v][3];
      tmp1 = tmp[v][1]+tmp[v][7];
      tmp2 = tmp[v][3]+tmp[v][5];
      a5 = tmp1 - tmp2;
      a6 = tmp[v][1]-tmp[v][7];
      a7 = tmp1 + tmp2;

      tmp4 = C6*(a4+a6);
      b2 = a2 * C4;
      b4 = Q*a4-tmp4;
      b5 = a5*C4;
      b6 = R*a6-tmp4;

      tmp3 = b6-a7;
      n0 = tmp3-b5;
      n1 = a0-a1;
      n2 = b2-a3;
      n3 = a0+a1;

      m3 = n1+n2;
      m4 = n3+a3;
      m5 = n1-n2;
      m6 = n3-a3;
      m7 = b4-n0;

      tmp[v][0] = m4+a7;
      tmp[v][1] = m3+tmp3;
      tmp[v][2] = m5-n0;
      tmp[v][3] = m6-m7;
      tmp[v][4] = m6+m7;
      tmp[v][5] = m5+n0;
      tmp[v][6] = m3-tmp3;
      tmp[v][7] = m4-a7;
    }}

  if (add)
    {
      for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
          {
            short val;
            if (tmp[v][u]<0.0) val = (short)(tmp[v][u]-0.5);
            else               val = (short)(tmp[v][u]+0.5);
      
            val += out[v][u];
            if (val<0) val=0; else if (val>255) val=255;
            out[v][u] = val;
          }
    }
  else
    {
      for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
          {
            short val;
            if (tmp[v][u]<0.0) val = (short)(tmp[v][u]-0.5);
            else               val = (short)(tmp[v][u]+0.5);
            if (val<0) val=0; else if (val>255) val=255;
            out[v][u] = val;
          }
    }
}
