/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "math/dct2_arai_int.hh"
#include <math.h>
#include <assert.h>
#include <iostream.h>
#include <iomanip.h>

// static int c1,c2;

#define SC(x,u,v) { D(x); ibuf[u][v] = ((x) >> (FDCTScale-FDCT_STAGE1_NK)); }

#define D(x,s) { cout << setbase(16) << setfill('0') << setw(16) << (x) << setbase(10) << " "; \
  ibuf[c1][c2]=x; c2++; if (c2==8) { c1++; c2=0; } cout << setfill(' '); }

//#define DESCALE(x,f)   ((x<0) ? ((x - ((1LL<<(f-1)))  ) >> (f)) : ((x + ((1LL<<(f-1)))  ) >> (f)))
#define DESCALE(x,f)    ((x + ((1LL<<(f)) -1 )  ) >> (f))

//#define DESCALE(x,f)      (x >> (f))
//#define DESCALE(x,f)   ((x<0) ? ((x - (1LL<<(f))/2  ) >> (f)) : ((x + (1LL<<(f))/2  ) >> (f)))
//#define DESCALE(x,f)    ((x + ((1LL<<(f))/2 )  ) >> (f))

#define IDESCALE(x,f)    ((x + ((1LL<<(f-1)) )  ) >> (f))
#define IDESCALENeg(x,f) ((x - ((1LL<<(f-1)) )  ) >> (f))

/* Eine 32x32 -> 64bit Multiplikation durchfuehren */
#define MUL32(a,b) (((long long)(a))*((long)(b)))

const long long FDCT_M1ScaleFact = (1LL<<FDCT_M1Scale);
const long long FDCT_M2ScaleFact = (1LL<<FDCT_M2Scale);
const long long FDCT_M3ScaleFact = (1LL<<FDCT_M3Scale);

const long long IDCT_M3ScaleFact = (1LL<<IDCT_M3Scale);
const long long IDCT_M2ScaleFact = (1LL<<IDCT_M2Scale);
const long long IDCT_M1ScaleFact = (1LL<<IDCT_M1Scale);

#ifndef PI
#define PI 3.1415926535
#endif

typedef long long DCTInt2;

static bool IsInitialized = false;

static DCTInt2  A1;
static DCTInt2 MA2;
static DCTInt2  A3;
static DCTInt2  A4;
static DCTInt2  A5;

static DCTInt2  A1b;
static DCTInt2 MA2b;
static DCTInt2  A3b;
static DCTInt2  A4b;
static DCTInt2  A5b;

static IDCTInt C2;
static IDCTInt C4;
static IDCTInt C6;
static IDCTInt Q;
static IDCTInt R;

static IDCTInt C2b;
static IDCTInt C4b;
static IDCTInt C6b;
static IDCTInt Qb;
static IDCTInt Rb;

static double OutputScaleFactor[8];  // fuer FDCT
static double InputScaleFactor[8];   // fuer IDCT


static void Init()
{
  if (IsInitialized)
    return;

  /* Beachte: Jeder Wert wird exakt einmal zugewiesen und nicht veraendert, dadurch
     ist der Code Multithread-sicher. */

  // FDCT

  A1 = (DCTInt2)(sqrt(0.5)*FDCT_M1ScaleFact+0.5);
  MA2 = (DCTInt2)((-(cos(PI/8.0) - cos(3.0*PI/8.0)))*FDCT_M1ScaleFact-0.5);
  A3 = (DCTInt2)(sqrt(0.5)*FDCT_M1ScaleFact+0.5);
  A4 = (DCTInt2)((cos(PI/8.0) + cos(3.0*PI/8.0))*FDCT_M1ScaleFact+0.5);
  A5 = (DCTInt2)(cos(3.0*PI/8.0)*FDCT_M1ScaleFact+0.5);

  A1b = (DCTInt2)(sqrt(0.5)*FDCT_M2ScaleFact+0.5);
  MA2b = (DCTInt2)((-(cos(PI/8.0) - cos(3.0*PI/8.0)))*FDCT_M2ScaleFact-0.5);
  A3b = (DCTInt2)(sqrt(0.5)*FDCT_M2ScaleFact+0.5);
  A4b = (DCTInt2)((cos(PI/8.0) + cos(3.0*PI/8.0))*FDCT_M2ScaleFact+0.5);
  A5b = (DCTInt2)(cos(3.0*PI/8.0)*FDCT_M2ScaleFact+0.5);

  OutputScaleFactor[0] = sqrt(0.5)/2.0;
  {for (int i=1;i<8;i++) OutputScaleFactor[i] = 0.25/cos(PI*(double)i/16.0);}


  // IDCT

  C2  = (DCTInt2)(2.0*cos(PI/8.0)*IDCT_M2ScaleFact+0.5);
  C4  = (DCTInt2)(sqrt(2.0)*IDCT_M2ScaleFact+0.5);
  C6  = (DCTInt2)(2.0*cos(3.0*PI/8.0)*IDCT_M2ScaleFact+0.5);
  Q   = (DCTInt2)((-2.0*cos(PI/8.0) + 2.0*cos(3.0*PI/8.0))*IDCT_M2ScaleFact-0.5);
  R   = (DCTInt2)(( 2.0*cos(PI/8.0) + 2.0*cos(3.0*PI/8.0))*IDCT_M2ScaleFact+0.5);
  /*
    cout << "C2 " << 2.0*cos(PI/8.0) << endl;
    cout << "C4 " << sqrt(2.0) << endl;
    cout << "C6 " << 2.0*cos(3.0*PI/8.0) << endl;
    cout << "Q  " << (-2.0*cos(PI/8.0) + 2.0*cos(3.0*PI/8.0)) << endl;
    cout << "R  " << ( 2.0*cos(PI/8.0) + 2.0*cos(3.0*PI/8.0)) << endl;
  */
  C2b = (DCTInt2)(2.0*cos(PI/8.0)*IDCT_M1ScaleFact+0.5);
  C4b = (DCTInt2)(sqrt(2.0)*IDCT_M1ScaleFact+0.5);
  C6b = (DCTInt2)(2.0*cos(3.0*PI/8.0)*IDCT_M1ScaleFact+0.5);
  Qb  = (DCTInt2)((-2.0*cos(PI/8.0) + 2.0*cos(3.0*PI/8.0))*IDCT_M1ScaleFact-0.5);
  Rb  = (DCTInt2)(( 2.0*cos(PI/8.0) + 2.0*cos(3.0*PI/8.0))*IDCT_M1ScaleFact+0.5);
  
  InputScaleFactor[0] = 0.5*0.5*sqrt(2);
  {for (int i=1;i<8;i++) InputScaleFactor[i] = 0.5*cos(PI*((double)i)/16.0);}

  IsInitialized=true;
}


static struct DummyInit
{
  DummyInit() { Init(); }
} dummyinit;



DCT2_AraiAguiNakajima_Int::DCT2_AraiAguiNakajima_Int()
{
  RemoveOutputScalefactors();
}

void DCT2_AraiAguiNakajima_Int::RemoveOutputScalefactors()
{
  Init();

  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = (DCTInt2)(OutputScaleFactor[u]*OutputScaleFactor[v]*FDCT_M3ScaleFact+0.5);
      }
}

void DCT2_AraiAguiNakajima_Int::SetOutputScalefactors(const DCTFloat* fact[8])
{
  Init();

  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = ((DCTInt2)(OutputScaleFactor[u]*OutputScaleFactor[v] *
                                  fact[v][u]*FDCT_M3ScaleFact));
      }
}

void DCT2_AraiAguiNakajima_Int::PerformFDCT2(const short* in[8] /* [y][x] */, DCTBlock& out)
{
  FDCTInt tmp[8][8];

  FDCTInt x1,x2,x3,x4,x5,x6,x7,x8;
  FDCTInt y1,y2,y3,y4,y5,y6,y7,y8;
  long long a5;

  // Zeilentransformationen

  for (int u=0;u<8;u++)
    {
      x1 = (FDCTInt)(in[0][u] + in[7][u]);
      y8 = (FDCTInt)(in[0][u] - in[7][u]);
      x2 = (FDCTInt)(in[1][u] + in[6][u]);
      x7 = (FDCTInt)(in[1][u] - in[6][u]);
      x3 = (FDCTInt)(in[2][u] + in[5][u]);
      x6 = (FDCTInt)(in[2][u] - in[5][u]);
      x4 = (FDCTInt)(in[3][u] + in[4][u]);
      x5 = (FDCTInt)(in[3][u] - in[4][u]);

      y1 =   x1 + x4;
      y2 =   x2 + x3;
      y3 =   x2 - x3;
      y4 =   x1 - x4;
      y5 = - x5 - x6;
      y6 =   x6 + x7;
      y7 =   x7 + y8;

      y8 <<= FDCT_M1Scale - FDCT_PostM1_Shift;

      x1 = (y1 + y2) << (FDCT_M1Scale - FDCT_PostM1_Shift);
      x2 = (y1 - y2) << (FDCT_M1Scale - FDCT_PostM1_Shift);
      x3 = y3 + y4;

      y4 <<= FDCT_M1Scale - FDCT_PostM1_Shift;
      
      a5 = MUL32(A5,(y5+y7));

      y3 = DESCALE(MUL32( A1 , x3 )     , FDCT_PostM1_Shift);
      x5 = DESCALE(MUL32(MA2 , y5 ) -a5 , FDCT_PostM1_Shift);
      y6 = DESCALE(MUL32( A3 , y6 )     , FDCT_PostM1_Shift);
      x7 = DESCALE(MUL32( A4 , y7 ) -a5 , FDCT_PostM1_Shift);

#if SHOW_M1
      cout << "IM1: ";
      D(x1,FDCT_PostM1_NK); D(x2,FDCT_PostM1_NK); D(y3,FDCT_PostM1_NK); D(y4,FDCT_PostM1_NK);
      cout << "\n     ";
      D(x5,FDCT_PostM1_NK); D(y6,FDCT_PostM1_NK); D(x7,FDCT_PostM1_NK); D(y8,FDCT_PostM1_NK);
      cout << endl;
#endif

      x3 = y4 + y3;
      x4 = y4 - y3;
      x6 = y8 + y6;
      x8 = y8 - y6;
      
      tmp[0][u] = x1;
      tmp[4][u] = x2;
      tmp[2][u] = x3;
      tmp[6][u] = x4;
      tmp[5][u] = x5 + x8;
      tmp[1][u] = x6 + x7;
      tmp[7][u] = x6 - x7;
      tmp[3][u] = x8 - x5;
    }

  // Spaltentransformationen

  for (int v=0;v<8;v++)
    {
      x1 = tmp[v][0] + tmp[v][7];
      x2 = tmp[v][1] + tmp[v][6];
      x3 = tmp[v][2] + tmp[v][5];
      x4 = tmp[v][3] + tmp[v][4];
      x5 = tmp[v][3] - tmp[v][4];
      x6 = tmp[v][2] - tmp[v][5];
      x7 = tmp[v][1] - tmp[v][6];
      y8 = tmp[v][0] - tmp[v][7];

      y1 =   x1 + x4;
      y2 =   x2 + x3;
      y3 =   x2 - x3;
      y4 =   x1 - x4;
      y5 = - x5 - x6;
      y6 =   x6 + x7;
      y7 =   x7 + y8;

      y8 = DESCALE(y8 , -(FDCT_M2Scale - FDCT_PostM2_Shift) );

      x1 = DESCALE(y1 + y2 , -(FDCT_M2Scale - FDCT_PostM2_Shift));
      x2 = DESCALE(y1 - y2 , -(FDCT_M2Scale - FDCT_PostM2_Shift));
      x3 = y3 + y4;

      y4 = DESCALE(y4 , -(FDCT_M2Scale - FDCT_PostM2_Shift));
      
      a5 = MUL32(A5b , (y5+y7));

      y3 = DESCALE(MUL32( A1 , x3 )    , FDCT_PostM2_Shift);
      x5 = DESCALE(MUL32(MA2 , y5 ) -a5, FDCT_PostM2_Shift);
      y6 = DESCALE(MUL32( A3 , y6 )    , FDCT_PostM2_Shift);
      x7 = DESCALE(MUL32( A4 , y7 ) -a5, FDCT_PostM2_Shift);

#if SHOW_M2
      cout << "IM2: ";
      D(x1,FDCT_PostM1_NK); D(x2,FDCT_PostM1_NK); D(y3,FDCT_PostM1_NK); D(y4,FDCT_PostM1_NK);
      cout << "\n     ";
      D(x5,FDCT_PostM1_NK); D(y6,FDCT_PostM1_NK); D(x7,FDCT_PostM1_NK); D(y8,FDCT_PostM1_NK);
      cout << endl;
#endif

      x3 = y3 + y4;
      x4 = y4 - y3;
      x6 = y6 + y8;
      x8 = y8 - y6;

      y5 = x5 + x8;
      y6 = x6 + x7;
      y7 = x6 - x7;
      y8 = x8 - x5;

      out.m_coeff[v][0] = (short)DESCALE(MUL32(d_fact[v][0] , x1) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][4] = (short)DESCALE(MUL32(d_fact[v][4] , x2) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][2] = (short)DESCALE(MUL32(d_fact[v][2] , x3) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][6] = (short)DESCALE(MUL32(d_fact[v][6] , x4) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][5] = (short)DESCALE(MUL32(d_fact[v][5] , y5) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][1] = (short)DESCALE(MUL32(d_fact[v][1] , y6) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][7] = (short)DESCALE(MUL32(d_fact[v][7] , y7) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
      out.m_coeff[v][3] = (short)DESCALE(MUL32(d_fact[v][3] , y8) , FDCT_PostM3_Shift+FDCT_PostM3_NK);
    }
}








IDCT2_AraiAguiNakajima_Int::IDCT2_AraiAguiNakajima_Int()
{
  RemoveOutputScalefactors();
}

void IDCT2_AraiAguiNakajima_Int::RemoveOutputScalefactors()
{
  Init();

  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = (DCTInt2)(InputScaleFactor[u]*InputScaleFactor[v]*IDCT_M3ScaleFact+0.5);
      }
}

void IDCT2_AraiAguiNakajima_Int::SetOutputScalefactors(const DCTFloat* fact[8])
{
  Init();

  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = (DCTInt2)(InputScaleFactor[u]*InputScaleFactor[v] * fact[v][u] *
                                 IDCT_M3ScaleFact+0.5);
      }
}


void IDCT2_AraiAguiNakajima_Int::PerformIDCT2(const short* in[8], uint8* out[8],bool add)
{
#if 0
  for (int u=0;u<8;u++,cout << endl)
    for (int v=0;v<8;v++)
      cout << "I" << in[u][v] << ' ';
#endif

  IDCTInt fin[8][8];
  IDCTInt tmp[8][8];

  IDCTInt a0,a1,a2,a3,a4,a5,a6,a7;
  IDCTInt       b2,   b4,b5,b6   ;
  IDCTInt n0,n1,n2,n3            ;
  IDCTInt          m3,m4,m5,m6,m7;
   DCTInt2 tmp1,tmp2,tmp3,tmp4;


  // Zeilentransformationen

  {for (int u=0;u<8;u++)
    {
      if (!in[0][u] && !in[1][u] && !in[2][u] && !in[3][u] &&
          !in[4][u] && !in[5][u] && !in[6][u] && !in[7][u])
        {
          tmp[0][u] = tmp[1][u] = tmp[2][u] = tmp[3][u] = 
          tmp[4][u] = tmp[5][u] = tmp[6][u] = tmp[7][u] = 0;
        }
      else
        {
          {
            for (int v=0;v<8;v++)
              {
                //cout << d_fact[v][u] << "< ";
                fin[v][u] = IDESCALE(MUL32(d_fact[v][u] , in[v][u]) , IDCT_PostM3_Shift);
                //cout << fin[v][u] << " . ";
              }
            //cout << endl;
          }

          a0 = fin[0][u];
          a1 = fin[4][u];
          a2 = fin[2][u]-fin[6][u];
          a3 = fin[2][u]+fin[6][u];
          a4 = fin[5][u]-fin[3][u];
          tmp1 = fin[1][u]+fin[7][u];
          tmp2 = fin[3][u]+fin[5][u];
          a5 = tmp1 - tmp2;
          a6 = fin[1][u]-fin[7][u];
          a7 = tmp1 + tmp2;
          
          tmp4 = MUL32(C6,(a4+a6));
          b2 = IDESCALE(MUL32(C4 , a2)       , IDCT_PostM2_Shift);
          b4 = IDESCALE(MUL32(Q  , a4) -tmp4 , IDCT_PostM2_Shift);
          b5 = IDESCALE(MUL32(C4 , a5)       , IDCT_PostM2_Shift);
          b6 = IDESCALE(MUL32(R  , a6) -tmp4 , IDCT_PostM2_Shift);

          a0 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          a1 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          a3 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          a7 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          
          tmp3 = b6-a7;
          n0 = tmp3-b5;
          n1 = a0-a1;
          n2 = b2-a3;
          n3 = a0+a1;
          
          m3 = n1+n2;
          m4 = n3+a3;
          m5 = n1-n2;
          m6 = n3-a3;
          m7 = b4-n0;
          
          tmp[0][u] = m4+a7;
          tmp[1][u] = m3+tmp3;
          tmp[2][u] = m5-n0;
          tmp[3][u] = m6-m7;
          tmp[4][u] = m6+m7;
          tmp[5][u] = m5+n0;
          tmp[6][u] = m3-tmp3;
          tmp[7][u] = m4-a7;
        }
    }}

#if 0
  for (int u=0;u<8;u++,cout << endl)
    for (int v=0;v<8;v++)
      cout << "T" << tmp[u][v] << ' ';
#endif


  // Spaltentransformationen

  {for (int v=0;v<8;v++)
    {
      a0 = tmp[v][0];
      a1 = tmp[v][4];
      a2 = tmp[v][2]-tmp[v][6];
      a3 = tmp[v][2]+tmp[v][6];
      a4 = tmp[v][5]-tmp[v][3];
      tmp1 = tmp[v][1]+tmp[v][7];
      tmp2 = tmp[v][3]+tmp[v][5];
      a5 = tmp1 - tmp2;
      a6 = tmp[v][1]-tmp[v][7];
      a7 = tmp1 + tmp2;

      tmp4 = MUL32(C6,(a4+a6));
      b2 = IDESCALE(MUL32(C4 , a2)       , IDCT_PostM1_Shift);
      b4 = IDESCALE(MUL32(Q  , a4) -tmp4 , IDCT_PostM1_Shift);
      b5 = IDESCALE(MUL32(C4 , a5)       , IDCT_PostM1_Shift);
      b6 = IDESCALE(MUL32(R  , a6) -tmp4 , IDCT_PostM1_Shift);

      a0 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);
      a1 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);
      a3 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);
      a7 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);

      tmp3 = b6-a7;
      n0 = tmp3-b5;
      n1 = a0-a1;
      n2 = b2-a3;
      n3 = a0+a1;

      m3 = n1+n2;
      m4 = n3+a3;
      m5 = n1-n2;
      m6 = n3-a3;
      m7 = b4-n0;

      tmp[v][0] = m4+a7;
      tmp[v][1] = m3+tmp3;
      tmp[v][2] = m5-n0;
      tmp[v][3] = m6-m7;
      tmp[v][4] = m6+m7;
      tmp[v][5] = m5+n0;
      tmp[v][6] = m3-tmp3;
      tmp[v][7] = m4-a7;
    }}

  if (add)
    {
      {for (int v=0;v<8;v++)
	for (int u=0;u<8;u++)
          {
            short val;
            if (tmp[v][u]<0.0) val = IDESCALE   (tmp[v][u] , IDCT_PostM1_NK);
            else               val = IDESCALENeg(tmp[v][u] , IDCT_PostM1_NK);

            val += out[v][u];
            if (val<0) val=0; else if (val>255) val=255;
            out[v][u] = val;
          }
      }
    }
  else
    {
      {
        for (int v=0;v<8;v++ /*,cout << endl*/ )
          for (int u=0;u<8;u++)
            if (tmp[v][u]<=0.0)
              {
                out[v][u]=0;
                //cout << "d0 ";
              }
            else
              {
                short t = IDESCALE(tmp[v][u] , IDCT_PostM1_NK);
                if (t>255) t=255;
                out[v][u]=t;
                //cout << "d" << t << ' ';
              }
      }
    }
}



void IDCT2_AraiAguiNakajima_Int::PerformIDCT2(const short* in[8], short* out[8],bool add)
{
  IDCTInt fin[8][8];
  IDCTInt tmp[8][8];

  IDCTInt a0,a1,a2,a3,a4,a5,a6,a7;
  IDCTInt       b2,   b4,b5,b6   ;
  IDCTInt n0,n1,n2,n3            ;
  IDCTInt          m3,m4,m5,m6,m7;
   DCTInt2 tmp1,tmp2,tmp3,tmp4;


  // Zeilentransformationen

  {for (int u=0;u<8;u++)
    {
      if (!in[0][u] && !in[1][u] && !in[2][u] && !in[3][u] &&
          !in[4][u] && !in[5][u] && !in[6][u] && !in[7][u])
        {
          tmp[0][u] = tmp[1][u] = tmp[2][u] = tmp[3][u] = 
          tmp[4][u] = tmp[5][u] = tmp[6][u] = tmp[7][u] = 0;
        }
      else
        {
          {for (int v=0;v<8;v++)
            fin[v][u] = IDESCALE(MUL32(d_fact[v][u] , in[v][u]) , IDCT_PostM3_Shift);}

          a0 = fin[0][u];
          a1 = fin[4][u];
          a2 = fin[2][u]-fin[6][u];
          a3 = fin[2][u]+fin[6][u];
          a4 = fin[5][u]-fin[3][u];
          tmp1 = fin[1][u]+fin[7][u];
          tmp2 = fin[3][u]+fin[5][u];
          a5 = tmp1 - tmp2;
          a6 = fin[1][u]-fin[7][u];
          a7 = tmp1 + tmp2;
          
          tmp4 = MUL32(C6,(a4+a6));
          b2 = IDESCALE(MUL32(C4 , a2)       , IDCT_PostM2_Shift);
          b4 = IDESCALE(MUL32(Q  , a4) -tmp4 , IDCT_PostM2_Shift);
          b5 = IDESCALE(MUL32(C4 , a5)       , IDCT_PostM2_Shift);
          b6 = IDESCALE(MUL32(R  , a6) -tmp4 , IDCT_PostM2_Shift);

          a0 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          a1 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          a3 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          a7 >>= -(IDCT_M2Scale - IDCT_PostM2_Shift);
          
          tmp3 = b6-a7;
          n0 = tmp3-b5;
          n1 = a0-a1;
          n2 = b2-a3;
          n3 = a0+a1;
          
          m3 = n1+n2;
          m4 = n3+a3;
          m5 = n1-n2;
          m6 = n3-a3;
          m7 = b4-n0;
          
          tmp[0][u] = m4+a7;
          tmp[1][u] = m3+tmp3;
          tmp[2][u] = m5-n0;
          tmp[3][u] = m6-m7;
          tmp[4][u] = m6+m7;
          tmp[5][u] = m5+n0;
          tmp[6][u] = m3-tmp3;
          tmp[7][u] = m4-a7;
        }
    }}


  // Spaltentransformationen

  {for (int v=0;v<8;v++)
    {
      a0 = tmp[v][0];
      a1 = tmp[v][4];
      a2 = tmp[v][2]-tmp[v][6];
      a3 = tmp[v][2]+tmp[v][6];
      a4 = tmp[v][5]-tmp[v][3];
      tmp1 = tmp[v][1]+tmp[v][7];
      tmp2 = tmp[v][3]+tmp[v][5];
      a5 = tmp1 - tmp2;
      a6 = tmp[v][1]-tmp[v][7];
      a7 = tmp1 + tmp2;

      tmp4 = MUL32(C6,(a4+a6));
      b2 = IDESCALE(MUL32(C4 , a2)       , IDCT_PostM1_Shift);
      b4 = IDESCALE(MUL32(Q  , a4) -tmp4 , IDCT_PostM1_Shift);
      b5 = IDESCALE(MUL32(C4 , a5)       , IDCT_PostM1_Shift);
      b6 = IDESCALE(MUL32(R  , a6) -tmp4 , IDCT_PostM1_Shift);

      a0 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);
      a1 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);
      a3 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);
      a7 >>= -(IDCT_M1Scale - IDCT_PostM1_Shift);

      tmp3 = b6-a7;
      n0 = tmp3-b5;
      n1 = a0-a1;
      n2 = b2-a3;
      n3 = a0+a1;

      m3 = n1+n2;
      m4 = n3+a3;
      m5 = n1-n2;
      m6 = n3-a3;
      m7 = b4-n0;

      tmp[v][0] = m4+a7;
      tmp[v][1] = m3+tmp3;
      tmp[v][2] = m5-n0;
      tmp[v][3] = m6-m7;
      tmp[v][4] = m6+m7;
      tmp[v][5] = m5+n0;
      tmp[v][6] = m3-tmp3;
      tmp[v][7] = m4-a7;
    }}

    {
      {for (int v=0;v<8;v++ /*,cout << endl*/ )
	for (int u=0;u<8;u++)
	  {
            short t = IDESCALE(tmp[v][u] , IDCT_PostM1_NK);
            out[v][u]=t;

            //cout << "D" << t << ' ';
          }
      }
    }
}


