/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "math/dct2_theoretic.hh"
#include <math.h>
#include <assert.h>

#ifndef PI
#define PI 3.1415926535
#endif

static DCTFloat costab[8][8]; // Tabelle aller benoetigten Cosinus-Werte
static DCTFloat C0; // 1/sqrt(2)

static class InitCosTab  // Dummy-Klasse zur Initialisierung der Cosinuswerte-Tabelle
{
public:
  InitCosTab()
  {
    for (int x=0;x<8;x++)
      for (int u=0;u<8;u++)
	costab[x][u] = cos((2*((DCTFloat)x)+1)*((DCTFloat)u)*PI/16.0);

    C0 = sqrt(0.5);
  }
} initcostab;


DCT2_Theoretic::DCT2_Theoretic()
{
  RemoveOutputScalefactors();
}

void DCT2_Theoretic::RemoveOutputScalefactors()
{
  d_scale_output=false;
}

void DCT2_Theoretic::SetOutputScalefactors(const DCTFloat* fact[8])
{
  for (int u=0;u<8;u++)
    for (int v=0;v<8;v++)
      {
	d_fact[v][u] = fact[v][u];
      }

  d_scale_output=true;
}


void DCT2_Theoretic::PerformIDCT2(const short* in[8], short* out[8],bool add)
{
  DCTFloat tmp[8][8];

  // Zeilentransformationen

  for (int v=0;v<8;v++)
    for (int x=0;x<8;x++)
      {
	DCTFloat sum=0.0;
	
	for (int u=0;u<8;u++)
	  {
	    DCTFloat val = in[v][u] * costab[x][u];
	    if (u==0) val *= C0;
	    sum += val;
	  }
	
	tmp[v][x] = sum;
      }

  // Spaltentransformationen

  for (int u=0;u<8;u++)
    for (int x=0;x<8;x++)
      {
	DCTFloat sum=0.0;
	
	for (int v=0;v<8;v++)
	  {
	    DCTFloat val = tmp[v][u] * costab[x][v];
	    if (v==0) val *= C0;
	    sum += val;
	  }

	sum *= 0.25;

	if (d_scale_output) sum *= d_fact[x][u];

	if (add)
	  {
	    if (sum<0.0) out[x][u] += (short)(sum-0.5);
	    else         out[x][u] += (short)(sum+0.5);
	  }
	else
	  {
	    if (sum<0.0) out[x][u] = (short)(sum-0.5);
	    else         out[x][u] = (short)(sum+0.5);
	  }
      }
}
