/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef MPEG_SYNTAXELEMS_HH
#define MPEG_SYNTAXELEMS_HH

#include "types.hh"

struct SequenceHeaderData
{
  uint14 m_width,m_height;
  uint4  m_aspect_ratio_information;
  uint16 m_framerate100;
  uint30 m_bit_rate;
  uint18 m_vbv_buffer_size;
  bool   m_constrained_parameters_flag;
  
  bool   m_load_intra_quantiser_matrix;
  uint8  m_intra_matrix[64];
  
  bool   m_load_nonintra_quantiser_matrix;
  uint8  m_nonintra_matrix[64];
  
  // MPEG2
  
  uint3  m_profile;
  uint4  m_level;
  bool   m_progressive_sequence;
  uint2  m_chroma_format;
  bool   m_low_delay;
};


struct GOPHeaderData
{
  GOPHeaderData();
  
  void AdvanceOnePicture(int fps);
  
  bool  m_time_code_drop_frame_flag;
  uint5 m_time_code_hours;
  uint6 m_time_code_minutes;
  uint6 m_time_code_seconds;
  uint6 m_time_code_pictures;
  bool  m_closed_gop;
  bool  m_broken_link;
};


struct PictureHeaderData
{
  uint10 m_temporal_reference;
  uint3  m_picture_coding_type;
  uint16 m_vbv_delay;
  bool   m_full_pel_forward_vector;
#define m_forward_f_code  m_f_code[0][0]
  bool   m_full_pel_backward_vector;
#define m_backward_f_code m_f_code[1][0]
  
  // MPEG2
  
  uint4  m_f_code[2 /* Zeitrichtung 0=forw/1=back */][2 /* Raumrichtung 0=horiz/1=vert */];
  uint8  m_intra_dc_precision;  /* 8-11 */
  uint2  m_picture_structure;
  bool   m_top_field_first;
  bool   m_frame_pred_frame_dct;
  bool   m_concealment_motion_vectors;
  uint1  m_q_scale_type;
  uint1  m_intra_vlc_format;
  bool   m_alternate_scan;
  bool   m_repeat_first_field;
  bool   m_progressive_frame;
  bool   m_composite_display_flag;
  uint1  m_v_axis;
  uint3  m_field_sequence;
  bool   m_sub_carrier;
  uint7  m_burst_amplitude;
  uint8  m_sub_carrier_phase;
};

struct SliceHeaderData
{
  uint10 m_slice_vertical_position;
  uint7  m_priority_breakpoint;
  uint5  m_quantiser_scale_code;
  bool   m_CodeIntraSlice;
  bool   m_intra_slice;
};

#endif
