/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "mpeg/utility.hh"
#include "mpeg/constants.hh"
#include <stdlib.h>
// #include <cstddef>

static struct
{
  uint16 framerate100;

  uint4 frame_rate_code;
  uint2 extnum; // 1-4
  uint5 extden; // 1-32
} frameratetable[] =
  {
    { 2397, FRAMERATE_23_976,1, 1 },
    { 2400, FRAMERATE_24    ,1, 1 },
    { 2500, FRAMERATE_25    ,1, 1 },
    { 2997, FRAMERATE_29_97 ,1, 1 },
    { 3000, FRAMERATE_30    ,1, 1 },
    { 5000, FRAMERATE_50    ,1, 1 },
    { 5994, FRAMERATE_59_94 ,1, 1 },
    { 6000, FRAMERATE_60    ,1, 1 },
    {  600, FRAMERATE_60    ,1,10 },
    {  500, FRAMERATE_50    ,1,10 },
    {  400, FRAMERATE_50    ,4,50 },
    {  300, FRAMERATE_50    ,3,50 },
    {  200, FRAMERATE_50    ,2,50 },
    {  100, FRAMERATE_50    ,1,50 },
    { 1000, FRAMERATE_50    ,1, 5 },
    { 1200, FRAMERATE_60    ,1, 5 },
    { 1500, FRAMERATE_30    ,1, 2 },
    { 2000, FRAMERATE_30    ,2, 3 },
    { 9000, FRAMERATE_30    ,3, 1 },
    { 1250, FRAMERATE_25    ,1, 2 },
    {  833, FRAMERATE_25    ,1, 3 },
    {  625, FRAMERATE_25    ,1, 4 },
    {  500, FRAMERATE_25    ,1, 5 }
  };

void CodeFrameRate(int framerate100,bool mayuseextension,
		   uint4& frame_rate_code, uint2& extnum, uint5& extden)
{
  for (size_t i=0;i<sizeof(frameratetable)/sizeof(frameratetable[0]);i++)
    {
      if (framerate100 == frameratetable[i].framerate100)
        {
          frame_rate_code = frameratetable[i].frame_rate_code;
          extnum          = frameratetable[i].extnum-1;
          extden          = frameratetable[i].extden-1;
          
          if (!mayuseextension && (extnum!=0 || extden!=0))
            continue;
          
          return;
        }
    }

  cerr << "WARNING: UNKNOWN FRAMERATE: " << framerate100 << endl;
  CodeFrameRate(2500,false,frame_rate_code,extnum,extden);
  // throw "Could not code specified frame rate\n";
}

uint8 CodeProfileLevel(uint3 profile,uint4 level)
{
  return (profile<<4) | level;
}

