/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "mpeg/vlctabs.hh"
#include "mpeg/constants.hh"
#include <stdlib.h>

struct DirectVLCTableEntry
{
  BitArray bits; int nBits;
};


static DirectVLCTableEntry vlc_mbaddrinc_tab[] =
{
  {0,0}, // Dummy-Eintrag

  /* Tabelle geklaut aus MSSG-Encoder */
  {0x01,1},  {0x03,3},  {0x02,3},  {0x03,4},
  {0x02,4},  {0x03,5},  {0x02,5},  {0x07,7},
  {0x06,7},  {0x0b,8},  {0x0a,8},  {0x09,8},
  {0x08,8},  {0x07,8},  {0x06,8},  {0x17,10},
  {0x16,10}, {0x15,10}, {0x14,10}, {0x13,10},
  {0x12,10}, {0x23,11}, {0x22,11}, {0x21,11},
  {0x20,11}, {0x1f,11}, {0x1e,11}, {0x1d,11},
  {0x1c,11}, {0x1b,11}, {0x1a,11}, {0x19,11},
  {0x18,11}
};


void WriteMacroblockAddressIncrement(BitBuffer& bs,int incr)
{
  while (incr>33)
    { bs.WriteBits(8,11); incr-=33; }

  assert(incr>0 && incr<=33);

  bs.WriteBits(vlc_mbaddrinc_tab[incr].bits,
	       vlc_mbaddrinc_tab[incr].nBits);
}


// =========================================================================


static DirectVLCTableEntry vlc_mbtype_tab[4][32]=
{
  { /* Dummy */
    {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}
  },

  /* Tabelle geklaut aus MSSG-Encoder */

  /* I */
  {
    {0,0}, {1,1}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {1,2}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}
  },
  /* P */
  {
    {0,0}, {3,5}, {1,2}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {1,3}, {0,0}, {1,1}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {1,6}, {1,5}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0},
    {0,0}, {0,0}, {2,5}, {0,0}, {0,0}, {0,0}, {0,0}, {0,0}
  },
  /* B */
  {
    {0,0}, {3,5}, {0,0}, {0,0}, {2,3}, {0,0}, {3,3}, {0,0},
    {2,4}, {0,0}, {3,4}, {0,0}, {2,2}, {0,0}, {3,2}, {0,0},
    {0,0}, {1,6}, {0,0}, {0,0}, {0,0}, {0,0}, {2,6}, {0,0},
    {0,0}, {0,0}, {3,6}, {0,0}, {0,0}, {0,0}, {2,5}, {0,0}
  }
};

#include <stdio.h>

void WriteMacroblockModes(BitBuffer& bs,uint8 packedmodes,uint3 pictype,uint2 scalabilitytype)
{
  if (packedmodes<0 || packedmodes>31)
    {
      fprintf(stderr,"pm: %d\n",packedmodes);
      assert(0);
    }

  if (vlc_mbtype_tab[pictype][packedmodes].bits == 0 &&
      vlc_mbtype_tab[pictype][packedmodes].nBits == 0)
    {
      fprintf(stderr,"INVALID MB-TYPE: %d\n",packedmodes);
      assert(0);
    }

  bs.WriteBits(vlc_mbtype_tab[pictype][packedmodes].bits,
	       vlc_mbtype_tab[pictype][packedmodes].nBits);
}


// =========================================================================


static DirectVLCTableEntry vlc_cbp_tab[64]=
{
  /* Tabelle geklaut aus MSSG-Encoder */

  {0x01,9}, {0x0b,5}, {0x09,5}, {0x0d,6}, 
  {0x0d,4}, {0x17,7}, {0x13,7}, {0x1f,8}, 
  {0x0c,4}, {0x16,7}, {0x12,7}, {0x1e,8}, 
  {0x13,5}, {0x1b,8}, {0x17,8}, {0x13,8}, 
  {0x0b,4}, {0x15,7}, {0x11,7}, {0x1d,8}, 
  {0x11,5}, {0x19,8}, {0x15,8}, {0x11,8}, 
  {0x0f,6}, {0x0f,8}, {0x0d,8}, {0x03,9}, 
  {0x0f,5}, {0x0b,8}, {0x07,8}, {0x07,9}, 
  {0x0a,4}, {0x14,7}, {0x10,7}, {0x1c,8}, 
  {0x0e,6}, {0x0e,8}, {0x0c,8}, {0x02,9}, 
  {0x10,5}, {0x18,8}, {0x14,8}, {0x10,8}, 
  {0x0e,5}, {0x0a,8}, {0x06,8}, {0x06,9}, 
  {0x12,5}, {0x1a,8}, {0x16,8}, {0x12,8}, 
  {0x0d,5}, {0x09,8}, {0x05,8}, {0x05,9}, 
  {0x0c,5}, {0x08,8}, {0x04,8}, {0x04,9},
  {0x07,3}, {0x0a,5}, {0x08,5}, {0x0c,6}
};

void WriteCBP(BitBuffer& bs,uint8 cbp1,uint8 cbp_ext,uint2 chromaformat)
{
  bs.WriteBits(vlc_cbp_tab[cbp1].bits , vlc_cbp_tab[cbp1].nBits);

  if (chromaformat==CHROMA_420)
    { assert(cbp1 != 0); }

       if (chromaformat==CHROMA_422) bs.WriteBits(cbp_ext,2,4);
  else if (chromaformat==CHROMA_444) bs.WriteBits(cbp_ext,6,0);
}


// =========================================================================

static DirectVLCTableEntry vlc_DClum_tab[12]=
{
  /* Tabelle geklaut aus MSSG-Encoder */
  {0x0004,3}, {0x0000,2}, {0x0001,2}, {0x0005,3}, {0x0006,3}, {0x000e,4},
  {0x001e,5}, {0x003e,6}, {0x007e,7}, {0x00fe,8}, {0x01fe,9}, {0x01ff,9}
};

static DirectVLCTableEntry vlc_DCchrom_tab[12]=
{
  /* Tabelle geklaut aus MSSG-Encoder */
  {0x0000,2}, {0x0001,2}, {0x0002,2}, {0x0006,3}, {0x000e,4}, {0x001e,5},
  {0x003e,6}, {0x007e,7}, {0x00fe,8}, {0x01fe,9}, {0x03fe,10},{0x03ff,10}
};


void WriteDCSizeLumi  (BitBuffer& bs,int size)
{
  assert(size >= 0);
  assert(size <= 11);

  bs.WriteBits(vlc_DClum_tab[size].bits,
	       vlc_DClum_tab[size].nBits);
}

void WriteDCSizeChroma(BitBuffer& bs,int size)
{
  assert(size >= 0);
  assert(size <= 11);

  bs.WriteBits(vlc_DCchrom_tab[size].bits,
	       vlc_DCchrom_tab[size].nBits);
}

// =========================================================================

static struct DirectVLCTableEntry dct_code_B14a[2][40]=
{
 /* run = 0, level = 1...40 */
 {
  {0x03, 2}, {0x04, 4}, {0x05, 5}, {0x06, 7},
  {0x26, 8}, {0x21, 8}, {0x0a,10}, {0x1d,12},
  {0x18,12}, {0x13,12}, {0x10,12}, {0x1a,13},
  {0x19,13}, {0x18,13}, {0x17,13}, {0x1f,14},
  {0x1e,14}, {0x1d,14}, {0x1c,14}, {0x1b,14},
  {0x1a,14}, {0x19,14}, {0x18,14}, {0x17,14},
  {0x16,14}, {0x15,14}, {0x14,14}, {0x13,14},
  {0x12,14}, {0x11,14}, {0x10,14}, {0x18,15},
  {0x17,15}, {0x16,15}, {0x15,15}, {0x14,15},
  {0x13,15}, {0x12,15}, {0x11,15}, {0x10,15}
 },
 /* run = 1, level = 1...18 */
 {
  {0x03, 3}, {0x06, 6}, {0x25, 8}, {0x0c,10},
  {0x1b,12}, {0x16,13}, {0x15,13}, {0x1f,15},
  {0x1e,15}, {0x1d,15}, {0x1c,15}, {0x1b,15},
  {0x1a,15}, {0x19,15}, {0x13,16}, {0x12,16},
  {0x11,16}, {0x10,16}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}
 }
};

static struct DirectVLCTableEntry dct_code_B14b[30][5]=
{
  /* run = 2...31, level = 1...5 */
  {{0x05, 4}, {0x04, 7}, {0x0b,10}, {0x14,12}, {0x14,13}},
  {{0x07, 5}, {0x24, 8}, {0x1c,12}, {0x13,13}, {0x00, 0}},
  {{0x06, 5}, {0x0f,10}, {0x12,12}, {0x00, 0}, {0x00, 0}},
  {{0x07, 6}, {0x09,10}, {0x12,13}, {0x00, 0}, {0x00, 0}},
  {{0x05, 6}, {0x1e,12}, {0x14,16}, {0x00, 0}, {0x00, 0}},
  {{0x04, 6}, {0x15,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x07, 7}, {0x11,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x05, 7}, {0x11,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x27, 8}, {0x10,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x23, 8}, {0x1a,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x22, 8}, {0x19,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x20, 8}, {0x18,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x0e,10}, {0x17,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x0d,10}, {0x16,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x08,10}, {0x15,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1a,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x19,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x17,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x16,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1e,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1d,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1c,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1b,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1e,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1d,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1c,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1b,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}}
};


static struct DirectVLCTableEntry dct_code_B15a[2][40]=
{
 /* run = 0, level = 1...40 */
 {
  {0x02, 2}, {0x06, 3}, {0x07, 4}, {0x1c, 5},
  {0x1d, 5}, {0x05, 6}, {0x04, 6}, {0x7b, 7},
  {0x7c, 7}, {0x23, 8}, {0x22, 8}, {0xfa, 8},
  {0xfb, 8}, {0xfe, 8}, {0xff, 8}, {0x1f,14},
  {0x1e,14}, {0x1d,14}, {0x1c,14}, {0x1b,14},
  {0x1a,14}, {0x19,14}, {0x18,14}, {0x17,14},
  {0x16,14}, {0x15,14}, {0x14,14}, {0x13,14},
  {0x12,14}, {0x11,14}, {0x10,14}, {0x18,15},
  {0x17,15}, {0x16,15}, {0x15,15}, {0x14,15},
  {0x13,15}, {0x12,15}, {0x11,15}, {0x10,15}
 },
 /* run = 1, level = 1...18 */
 {
  {0x02, 3}, {0x06, 5}, {0x79, 7}, {0x27, 8},
  {0x20, 8}, {0x16,13}, {0x15,13}, {0x1f,15},
  {0x1e,15}, {0x1d,15}, {0x1c,15}, {0x1b,15},
  {0x1a,15}, {0x19,15}, {0x13,16}, {0x12,16},
  {0x11,16}, {0x10,16}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0},
  {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}
 }
};

static struct DirectVLCTableEntry dct_code_B15b[30][5]=
{
  /* run = 2...31, level = 1...5 */
  {{0x05, 5}, {0x07, 7}, {0xfc, 8}, {0x0c,10}, {0x14,13}},
  {{0x07, 5}, {0x26, 8}, {0x1c,12}, {0x13,13}, {0x00, 0}},
  {{0x06, 6}, {0xfd, 8}, {0x12,12}, {0x00, 0}, {0x00, 0}},
  {{0x07, 6}, {0x04, 9}, {0x12,13}, {0x00, 0}, {0x00, 0}},
  {{0x06, 7}, {0x1e,12}, {0x14,16}, {0x00, 0}, {0x00, 0}},
  {{0x04, 7}, {0x15,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x05, 7}, {0x11,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x78, 7}, {0x11,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x7a, 7}, {0x10,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x21, 8}, {0x1a,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x25, 8}, {0x19,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x24, 8}, {0x18,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x05, 9}, {0x17,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x07, 9}, {0x16,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x0d,10}, {0x15,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1a,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x19,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x17,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x16,12}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1e,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1d,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1c,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1b,13}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1f,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1e,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1d,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1c,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}},
  {{0x1b,16}, {0x00, 0}, {0x00, 0}, {0x00, 0}, {0x00, 0}}
};

void CodeRunValue(BitBuffer& bs,int run,int value, bool first, bool UseTabB15, bool GenerateMPEG1)
{
  assert(run>=0);
  assert(run<=63);
  assert(value>=-2047);
  assert(value<= 2047);
  assert(value!=0);

  int absvalue = abs(value);

  bool InTab = false;
  bool TabA = false;   // Initialisierung eigentlich nicht notwendig, beruhigt aber den Compiler.

  BitArray bits=0;  // Initialialisierung auch hier nicht noetig.
  int nBits = 0;

  if (first && !UseTabB15 && run==0 && absvalue==1)
    {
      if (value>0) bs.WriteBits(2,2);
      else         bs.WriteBits(3,2);
    }
  else
    {
      if (run<=1)
	{ if (absvalue<=40) { InTab = true; TabA=true;  } }
      else if (run<=31)
	{ if (absvalue<=5)  { InTab = true; TabA=false; } }

      if (InTab)
	{
	  if (UseTabB15)
	    {
	      if (TabA)
		{
		  bits  = dct_code_B15a[run][absvalue-1].bits;
		  nBits = dct_code_B15a[run][absvalue-1].nBits;
		}
	      else
		{
		  bits  = dct_code_B15b[run-2][absvalue-1].bits;
		  nBits = dct_code_B15b[run-2][absvalue-1].nBits;
		}
	    }
	  else
	    {
	      if (TabA)
		{
		  bits  = dct_code_B14a[run][absvalue-1].bits;
		  nBits = dct_code_B14a[run][absvalue-1].nBits;
		}
	      else
		{
		  bits  = dct_code_B14b[run-2][absvalue-1].bits;
		  nBits = dct_code_B14b[run-2][absvalue-1].nBits;
		}
	    }
	}

      if (nBits>0)
	{
	  nBits++; bits <<= 1;
	  if (value<0) bits |= 1;
	  bs.WriteBits(bits,nBits);
	}
      else
	{
	  /* Escape-Sequenz codieren */

	  if (GenerateMPEG1)
	    {
	      bs.WriteBits(1,6);
	      bs.WriteBits(run,6);
	      if (absvalue <= 127)
		bs.WriteBitsMasked(value,8);
	      else
		{
		  if (value>0)
                    {
                      assert(value>=128);
                      assert(value<=255);
                      bs.WriteBitsMasked(value,16);
                    }
		  else
                    {
                      assert(value>=-255);
                      assert(value<=-128);
                      bs.WriteBits(0x80,8);
                      bs.WriteBitsMasked(value,8);
                    }
		}
	    }
	  else
	    {
	      bs.WriteBits(1,6);
	      bs.WriteBits(run,6);
	      bs.WriteBitsMasked(value,12);
	    }
	}
    }
}

int  BitsForRunValue(int run,int value, bool UseTabB15, bool GenerateMPEG1)
{
  assert(run>=0);
  assert(run<=63);
  assert(value>=-2047);
  assert(value<= 2047);
  assert(value!=0);

  int absvalue = abs(value);

  bool InTab = false;
  bool TabA = false;   // Initialisierung eigentlich nicht notwendig, beruhigt aber den Compiler.

  if (run<=1)
    { if (absvalue<=40) { InTab = true; TabA=true;  } }
  else if (run<=31)
    { if (absvalue<=5)  { InTab = true; TabA=false; } }

  if (InTab)
    {
      if (UseTabB15)
	{
	  if (TabA) return dct_code_B15a[run][absvalue-1].nBits+1;
	  else      return dct_code_B15b[run-2][absvalue-1].nBits+1;
	}
      else
	{
	  if (TabA) return dct_code_B14a[run][absvalue-1].nBits+1;
	  else      return dct_code_B14b[run-2][absvalue-1].nBits+1;
	}
    }


  /* Escape-Sequenz codieren */
      
  if (GenerateMPEG1)
    {
      if (absvalue <= 127)
	return 20;
      else
	return 28;
    }
  else
    return 24;
}



// =========================================================================

static struct DirectVLCTableEntry motionvectab[17]=
{
  /* Tabelle geklaut aus MSSG-Encoder */
  {0x01,1},  {0x01,2},  {0x01,3},  {0x01,4},
  {0x03,6},  {0x05,7},  {0x04,7},  {0x03,7},
  {0x0b,9},  {0x0a,9},  {0x09,9},  {0x11,10},
  {0x10,10}, {0x0f,10}, {0x0e,10}, {0x0d,10},
  {0x0c,10}
};

#include <iostream.h>

void CodeMotionCode(BitBuffer& bs,int diff,int f,int rsize)
{
  if (diff==0)
    bs.WriteBits(1,1);
  else
    {
      int absdiffS1;
      int mcode_part;
      int residual;

      absdiffS1  = (diff<0 ? -diff : diff) -1;
      mcode_part = absdiffS1 / f +1;
      residual   = absdiffS1 % f;

      bs.WriteBits(motionvectab[mcode_part].bits, motionvectab[mcode_part].nBits);
      bs.WriteBits(diff>0 ? 0 : 1 ,1);
      if (rsize) bs.WriteBits(residual,rsize);
    }
}
