/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef OPTIONS_HH
#define OPTIONS_HH

#include "types.hh"


struct Options
{
  // Multiprozessor-Parameter

  uint32 nThreads;


  // Misc. Optionen

  enum ProgressDispl { ProgressDisplay_None  =0, ProgressDisplay_Minimal=1,
                       ProgressDisplay_Normal=2, ProgressDisplay_Maximal=3 } ProgressDisplay;


  // Eingabe

  enum { PPM, Targa, MJPEG_AVI } InputFormat;
  uint32 FirstFrameToLoad;
  uint32 LastFrameToLoad;
  
  struct
  {
    bool SmallPic; // Reduziert 2-Field, volle Aufloesung zu 1-Frame halbe Aufloesung
    bool BypassNTDriverBug;
    bool SaveAudio;
  } mjpeg_avi;

  struct
  {
    bool Interlaced;
  } ppm;

  // Bitstream-Format

  bool CreateMPEG2;
  bool CreateElementaryStream;


  // Codierungsoptionen / Video

  enum { FramePattern, SceneChange }         FrameTypeAlgo;

  struct
  {
    int GOPSize;    // N
    int PDistance;  // M
  } framepattern;

  struct
  {
    int Lookahead;
    int PDistance;
  } scenechange;

  enum { FullSearch, TSS, NTSS, OTS, ODFS, NullMV }  MVSearchAlgo;
  enum { FixedQScale, VariableQuantisation } QScaleAlgo;
  enum { MBType_CountBits, MBType_Variance } MBTypeAlgo;

  int  HSearchRange,VSearchRange; // in Pixel     (auf Frameebene?)

  struct
  {
    int qscale_I;
    int qscale_P;
    int qscale_B;
  } qscale_const;


  bool   Enable_CBR_RateControl;   // Ob CBR-Rate-Control benutzt werden soll.
  bool   Enable_JND;
  bool   Enable_LuminanceMasking;
  bool   Enable_ActivityMasking;
  bool   Enable_MotionMasking;
  uint32 Bitrate;                  // Ausgabe Bitrate it bits/sec


  // MPEG2-spezifische Codierungsoptionen

  struct
  {
    bool  CodeIntraSliceFlag;
    bool  ExponentialQScale;   // TODO: Auf Bildebene ?
    bool  AlternateScan;       // TODO: Auf Bildebene ?
    uint1 IntraVLCFormat;      // TODO: Auf Bildebene ?
    uint8 DC_Precision;        // TODO: Auf Bildebene ?
  } mpeg2;


  // Debug-Optionen

  bool Visualize;
  bool DisablePredictionCorrection;
  bool Show_DerivVsBits;
};

void InitOptions();
extern Options options;

#endif
