/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "output/bitbuffer.hh"

#define SHOWOUTPUT 0

const int BitsPerEntry  = sizeof(BitArray)*8;
//const int BytesPerEntry = sizeof(BitArray);
bool BitBuffer::IsLittleEndian;

static class EndianTest
{
public:
  EndianTest()
  {
    uint32 test32 = 0x12345678;
    const char* ptr = (char*)&test32;

         if (*ptr==0x12) BitBuffer::IsLittleEndian=false;
    else if (*ptr==0x78) BitBuffer::IsLittleEndian=true;
    else { assert(0); }
  }
} dummy_endiantest;


BitBuffer::BitBuffer()
{
  d_buffer = new uint8[BufferLength];
  d_buffersize = BufferLength;
  d_bufferidx  = 0;
  d_freebits   = BitsPerEntry;
  d_tmpbuf     = 0;
}

BitBuffer::~BitBuffer()
{
  delete[] d_buffer;
}

void BitBuffer::WriteBitsMasked(BitArray bits,int nBits)
{
  static const uint32 rmask[33] = {
    0x00000000,
    MASK1,  MASK2,  MASK3,  MASK4,   MASK5,  MASK6,  MASK7,  MASK8,
    MASK9,  MASK10, MASK11, MASK12,  MASK13, MASK14, MASK15, MASK16,
    MASK17, MASK18, MASK19, MASK20,  MASK21, MASK22, MASK23, MASK24,
    MASK25, MASK26, MASK27, MASK28,  MASK29, MASK30, MASK31, MASK32 };

  bits &= rmask[nBits];
  WriteBits(bits,nBits);
}


void BitBuffer::WriteBits(BitArray bits,int nBits,int rightoffset)
{
  bits >>= rightoffset;

  WriteBits(bits,nBits);
}

#include <iostream.h>

void BitBuffer::WriteBits(BitArray bits,int nBits)
{
#if 0
  for (int i=nBits-1;i>=0;i--)
    if (bits & (1<<i)) cerr << '1'; else cerr << '0';
  cerr << endl;
#endif

  /*
    +------+----------------------+
    |XXXXXX|0000000000000000000000|
    +------+----------------------+
                d_freebits

    +----+------------------------+
    |0000|XXXXXXXXXXXXXXXXXXXXXXXX|
    +----+------------------------+
                   nBits         
   */

  if (d_freebits>=nBits)
    {
      // Neue Bits passen komplett in das aktuelle Feld hinein.

      BitArray aligned = bits;

      int nShiftLeft = d_freebits - nBits;

      if (nShiftLeft>0)
	{
	  aligned <<= nShiftLeft;
	  d_tmpbuf |= aligned;
	  d_freebits = nShiftLeft;
	}
      else
	{
	  d_tmpbuf |= aligned;

          TmpArrayToBuffer();

          d_tmpbuf   = 0;
	  d_freebits = BitsPerEntry;
	}
    }
  else
    {
      // Neue Bits passen nicht mehr komplett in das aktuelle Feld.
      // Sie werden folglich in zwei Teile geteilt.

      BitArray rightpart = bits;   // Der Teil, der noch in das aktuelle Feld passt.
      BitArray leftpart  = bits;   // Der Rest, der in ein neues Feld kommt.

      int nShiftRight = nBits - d_freebits;
      if (nShiftRight != 0) rightpart >>= nShiftRight;

      d_tmpbuf |= rightpart;

      TmpArrayToBuffer();

      int nShiftLeft = BitsPerEntry-nShiftRight;
      leftpart <<= nShiftLeft;
      d_tmpbuf = leftpart;

      d_freebits = nShiftLeft;
    }
}

void BitBuffer::FillByte()
{
  d_freebits -= d_freebits%8;

  if (d_freebits==0)
    {
      TmpArrayToBuffer();
      d_tmpbuf = 0;
      d_freebits = BitsPerEntry;
    }
}


void BitBuffer::Flush()
{
  FillByte();

  if (d_freebits == BitsPerEntry)
    return;


  if (d_bufferidx == d_buffersize)
    {
      uint8* newbuf = new uint8[d_buffersize*2];
      for (int i=0;i<d_buffersize;i++)
        newbuf[i]=d_buffer[i];
      
      d_buffersize *= 2;

      delete[] d_buffer;
      d_buffer = newbuf;
    }

  int bytesinbuffer = (BitsPerEntry-d_freebits+7)/8;

  if (IsLittleEndian)
    {
      unsigned char* p = (unsigned char*)(&d_tmpbuf);
      for (int j=BytesPerEntry-1;j>=BytesPerEntry-bytesinbuffer;j--)
        d_buffer[d_bufferidx++] = p[j];
    }
  else
    {
      unsigned char* p = (unsigned char*)(&d_tmpbuf);
      for (int j=0;j<bytesinbuffer;j++)
        d_buffer[d_bufferidx++] = p[j];
    }
}
