/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#ifndef UTILITY_THREAD_HH
#define UTILITY_THREAD_HH

#ifdef VCPP
#include "win/thread.hh"
#else

#include <pthread.h>
// #include "utility/idlewait.hh"


class Thread
{
public:
  Thread() { d_stopthread=false; }

  void Start(void* (*funct)(void*),void* arg)
  {
    d_didstop=false;
#ifdef HPUX
    pthread_attr_t attr;
    pthread_attr_create(&attr);
    pthread_attr_setsched(&attr,SCHED_RR);
    pthread_create(&d_thread,attr,funct,arg);
#else
    pthread_create(&d_thread,NULL,funct,arg);
#endif
  }
  void Stop (bool waitforjoin=true)
  {
    d_stopthread=true;

    if (waitforjoin)
      while (!d_didstop)    //  ----- Das muesste doch auch ein Join() tun ?   Mal ausprobieren.
	{ /* idlewait(100000); */ }
  }

  bool ShallIStop() const { return d_stopthread; }
  void ImStopping() { d_didstop=true; }

  void Join() { pthread_join(d_thread,NULL); }

private:
  pthread_t d_thread;
  bool d_stopthread;
  volatile bool d_didstop;
};


class Mutex
{
public:
  Mutex()
  {
#ifdef HPUX
    pthread_mutex_init(&d_mutex,pthread_mutexattr_default);
#else
    pthread_mutex_init(&d_mutex,NULL);
#endif
  }
  ~Mutex() { pthread_mutex_destroy(&d_mutex); }

  void Lock() { pthread_mutex_lock(&d_mutex); }
  void Unlock() { pthread_mutex_unlock(&d_mutex); }

private:
  pthread_mutex_t d_mutex;
};

#endif
#endif
