/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "writeppm.hh"
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>

#ifndef SHOWBIGARRAY
#define SHOWBIGARRAY 1
#endif


void OpenStream(class ofstream& ostr,const char* nametemplate,int framenr)
{
  char buf[1000];
  sprintf(buf,nametemplate,framenr);
  ostr.open((char*)buf);
}


Image<Pixel>* CreateVisualizationImage(const Image<Pixel>& img,bool centerpoints)
{
  Image<Pixel>* vimg = new Image<Pixel>();
  vimg->CreateImage(RGB_Undefined,img.AskWidth(),img.AskHeight());

  Pixel** r = vimg->AskBitmap(Bitmap_Red  ).AskFrame();
  Pixel** g = vimg->AskBitmap(Bitmap_Green).AskFrame();
  Pixel** b = vimg->AskBitmap(Bitmap_Blue ).AskFrame();

  const Pixel*const* lum = img.AskBitmap_const(Bitmap_Y).AskFrame_const();
  assert(lum);

  {for (int y=0;y<img.AskHeight();y++)
    for (int x=0;x<img.AskWidth();x++)
      {
	r[y][x] = g[y][x] = b[y][x]  = lum[y][x];
      }
  }

  if (centerpoints)
    {for (int y=9;y<img.AskHeight();y+=16)
      for (int x=9;x<img.AskWidth();x+=16)
        {
          g[y-1][x] = g[y][x-1] = g[y-2][x-1] = g[y-1][x-2] = 255;
        }
    }

  return vimg;
}

inline int abs(int x)  { if (x<0) return -x; else return x; }
inline int sign(int x) { if (x<0) return -1; else if (x>0) return 1; else return 0; }

Image<Pixel>* CreateDifferenceImage(const Image<DiffPixel>& img)
{
  Image<Pixel>* vimg = new Image<Pixel>();
  vimg->CreateImage(RGB_Undefined,img.AskWidth(),img.AskHeight());

  Pixel** r = vimg->AskBitmap(Bitmap_Red  ).AskFrame();
  Pixel** g = vimg->AskBitmap(Bitmap_Green).AskFrame();
  Pixel** b = vimg->AskBitmap(Bitmap_Blue ).AskFrame();

  const DiffPixel*const* lum = img.AskBitmap_const(Bitmap_Y).AskFrame_const();
  assert(lum);

  for (int y=0;y<img.AskHeight();y++)
    for (int x=0;x<img.AskWidth();x++)
      {
	r[y][x] = g[y][x] = b[y][x]  = (Pixel)abs(lum[y][x]);
      }

  return vimg;
}

static void DrawVec(Pixel** p,int x0,int y0, int vx,int vy)
{
  if (abs(vx)>abs(vy))
    {
      for (int x=0;x!=vx;x+=sign(vx))
	{ p[vy*x/vx+y0][x0+x] = 255; }
    }
  else
    {
      for (int y=0;y!=vy;y+=sign(vy))
	{ p[y0+y][x0+vx*y/vy] = 255; }
    }
}

void FillMotionVectors(Image<Pixel>& img,const ImageData& codedimg)
{
  Pixel** r = img.AskBitmap(Bitmap_Red  ).AskFrame();
  
  for (int my=0;my<codedimg.mblks.AskHeight();my++)
    for (int mx=0;mx<codedimg.mblks.AskWidth();mx++)
      {
	const Macroblock& mb = codedimg.mblks.Ask_const(my,mx);

	if (mb.m_HasMotionForward)
          DrawVec(r,mx*16+8,my*16+8,
                  mb.m_vector[0][0].m_habs/2,
                  mb.m_vector[0][0].m_vabs/2);

	if (mb.m_HasMotionBackward)
          DrawVec(r,mx*16+8,my*16+8,
                  mb.m_vector[0][1].m_habs/2,
                  mb.m_vector[0][1].m_vabs/2);
      }
}

void FillBlockTypes(Image<Pixel>& img,const ImageData& codedimg)
{
  Pixel** r = img.AskBitmap(Bitmap_Red  ).AskFrame();
  Pixel** g = img.AskBitmap(Bitmap_Green).AskFrame();
  
  for (int my=0;my<codedimg.mblks.AskHeight();my++)
    for (int mx=0;mx<codedimg.mblks.AskWidth();mx++)
      {
	const Macroblock& mb = codedimg.mblks.Ask_const(my,mx);

	Pixel** p;
	int step;
	if (mb.m_IsIntra) { p=g; step=1; } else { p=r; step=3; }

	for (int x=0;x<16;x+=step)
	  {
	    p[my*16]   [mx*16+x]=255;
	    p[my*16+15][mx*16+x]=255;
	  }
	for (int y=0;y<16;y+=step)
	  {
	    p[my*16+y][mx*16   ]=255;
	    p[my*16+y][mx*16+15]=255;
	  }
      }
}

void FillBlockClass(Image<Pixel>& img,const uint8 cls[100][200])
{
  Pixel** r = img.AskBitmap(Bitmap_Red  ).AskFrame();
  Pixel** g = img.AskBitmap(Bitmap_Green).AskFrame();
  Pixel** b = img.AskBitmap(Bitmap_Blue ).AskFrame();
  
  for (int my=0;my<img.AskHeight()/16;my++)
    for (int mx=0;mx<img.AskWidth()/16;mx++)
      {
	Pixel** p = 0;
	if (cls[my][mx]==0) p = r;
	if (cls[my][mx]==1) p = g;
	if (cls[my][mx]==2) p = b;
	assert(p);

	for (int x=0;x<16;x++)
	  {
	    p[my*16]   [mx*16+x]=255;
	    p[my*16+15][mx*16+x]=255;
	  }
	for (int y=0;y<16;y++)
	  {
	    p[my*16+y][mx*16   ]=255;
	    p[my*16+y][mx*16+15]=255;
	  }
      }
}

void FillBlockClass(Image<Pixel>& img,const Array2<uint8>& cls,bool mb)
{
  Pixel** r = img.AskBitmap(Bitmap_Red  ).AskFrame();
  Pixel** g = img.AskBitmap(Bitmap_Green).AskFrame();
  Pixel** b = img.AskBitmap(Bitmap_Blue ).AskFrame();

  /*
  cout << "R: " << r << endl;
  cout << "G: " << g << endl;
  cout << "B: " << b << endl;
  */

  int width  = img.AskWidth();
  int height = img.AskHeight();
  int fact;

  if (mb) { width/=16; height/=16; fact=16; }
  else    { width/= 8; height/= 8; fact= 8; }
  
  for (int my=0;my<height;my++)
    for (int mx=0;mx<width;mx++)
      {
        uint8 val = cls.Ask_const(my,mx);
        for (int i=0x01;i<=0x04;i<<=1)
          {
            Pixel** p = 0;
            uint8 pen;
            if (val & i) pen=255; else pen=0;
            if (i == BLK_RED  ) p = r;
            if (i == BLK_GREEN) p = g;
            if (i == BLK_BLUE ) p = b;
            assert(p);

            if (val & BLK_DOTS)
              {
                p[my*fact       ][mx*fact       ]=pen;
                p[my*fact+fact-1][mx*fact       ]=pen;
                p[my*fact+fact-1][mx*fact+fact-1]=pen;
                p[my*fact       ][mx*fact+fact-1]=pen;

                p[my*fact+1     ][mx*fact       ]=pen;
                p[my*fact+fact-2][mx*fact       ]=pen;
                p[my*fact+fact-2][mx*fact+fact-1]=pen;
                p[my*fact+1     ][mx*fact+fact-1]=pen;

                p[my*fact       ][mx*fact+1     ]=pen;
                p[my*fact+fact-1][mx*fact+1     ]=pen;
                p[my*fact+fact-1][mx*fact+fact-2]=pen;
                p[my*fact       ][mx*fact+fact-2]=pen;
              }

            if (val & BLK_RECT)
              {
                for (int x=0;x<fact;x++)
                  {
                    p[my*fact]       [mx*fact+x]=pen;
                    p[my*fact+fact-1][mx*fact+x]=pen;
                  }
                for (int y=0;y<fact;y++)
                  {
                    p[my*fact+y][mx*fact       ]=pen;
                    p[my*fact+y][mx*fact+fact-1]=pen;
                  }
              }

            if (val & BLK_LINE2)
              {
                for (int x=0;x<fact;x++)
                  {
                    p[my*fact+x][mx*fact+x]=pen;
                  }
              }

            if (val & BLK_LINE)
              {
                for (int x=0;x<fact;x++)
                  {
                    p[my*fact+fact-1-x][mx*fact+x]=pen;
                  }
              }
          }
      }
}

void WriteImageAsPPM(const Image<Pixel>& img,class ostream& ostr)
{
  ostr << "P6\n" << img.AskWidth() << ' ' << img.AskHeight() << "\n255\n";

  const Pixel*const * r = img.AskBitmap_const(Bitmap_Red  ).AskFrame_const();
  const Pixel*const * g = img.AskBitmap_const(Bitmap_Green).AskFrame_const();
  const Pixel*const * b = img.AskBitmap_const(Bitmap_Blue ).AskFrame_const();
  assert(r);
  assert(g);
  assert(b);

  for (int y=0;y<img.AskHeight();y++)
    for (int x=0;x<img.AskWidth();x++)
      {
	ostr << ((char)r[y][x])
	     << ((char)g[y][x])
	     << ((char)b[y][x]);
      }
}


void WriteArrayAsPPM(const Array2<int32>& img,const int32& minval,const int32& maxval,
                     class ofstream& ostr,int blksize)
{
  ostr << "P5\n"
#if SHOWBIGARRAY
       << (blksize*img.AskWidth()) << ' ' << (blksize*img.AskHeight())
#else
       << img.AskWidth() << ' ' << img.AskHeight()
#endif
       << "\n255\n";


  for (int y=0;y<img.AskHeight();y++)
#if SHOWBIGARRAY
    for (int p=0;p<blksize;p++)
#endif
      for (int x=0;x<img.AskWidth();x++)
#if SHOWBIGARRAY
        for (int q=0;q<blksize;q++)
#endif
          {
            int32 val = img.Ask_const(y,x);
            if (val>maxval) { val=maxval; /*cerr<<"WriteArrayAsPPM: Overflow\n"; */ }
            if (val<minval) { val=minval; /*cerr<<"WriteArrayAsPPM: Underflow\n";*/ }

            if ((((x*blksize+q)%16)==0 || ((x*blksize+q)%16)==15) &&
                (((y*blksize+p)%16)==0 || ((y*blksize+p)%16)==15))
              ostr << (char)(255-255*(val-minval)/(maxval-minval));
            else
              ostr << (char)(255*(val-minval)/(maxval-minval));
          }
}

/*
template void WriteArrayAsPPM(const Array2<uint32>& img,const uint32&,const uint32& maxval,class ostream&);
template void WriteArrayAsPPM(const Array2<int32>&  img,const  int32&,const  int32& maxval,class ostream&);
template void WriteArrayAsPPM(const Array2<double>& img,const double&,const double& maxval,class ostream&);
*/

#if 0
void WriteBitmap(const char* templ,int nr,const Bitmap<Pixel>& bm)
{
  char buf[500];
  sprintf(buf,templ,nr);
  fostream
}
#endif

#include "image/array2.cc"
template class Array2<uint8>;
